/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: catalib.c
 * Author 	: Paolo Montegriffo
 * Created on	: July, 2000	
 * Language	: ANSI C
 * Description	: library functions for CataPack
 * Version      : 2.0 - October 6, 2000
 * Version      : 2.1.3 - September 7, 2001
 * Version      : 2.1.9 - September 26, 2002 : added 'isspace' check on 
 *                        tokenize function
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include <ctype.h>

#include <math.h>
#include <regex.h>
#include <time.h>

#include "catalib.h"

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
 * Function  : tokenize
 * In        : line to be tokenized
 * Out       : array with line elements, number of tokens in the line
 * Purpouse  : Tokenize the input line, returning a char array with each 
 *             line element
 *---------------------------------------------------------------------------*/

char **tokenize (char *line, char *del, int *ntok)
{
   char  **tokens ;
   char    line2[2048] ;
   char   *s=line2 ;
   char   *token ;
   int     i ;

   /* If delimiter is space, substitute any other isspace char  with it */
   if (isspace (*del)) {
      char *ptr = line ;
      while (*ptr) {
         if (isspace (*ptr)) *ptr = *del ;
         ptr++ ;
      }
   }
   /* Counts number of elements */
   (void) strcpy (line2, line) ;
   i = 0 ;
   while ((token = strtok (s, del)) != NULL) {
      i++ ;
      s = (char*) NULL ;
   }
   *ntok = i ;

   /* Tokenize the line, allocating memory for output array */

   tokens = (char**) calloc ((size_t)(*ntok), sizeof(char*)) ;
   (void) strcpy (line2, line) ;
   s = line2 ;
   i = 0 ;
   while ((token = strtok(s,del)) != NULL) {
      strip_newline (token) ;
      tokens[i] = (char *) calloc (strlen (token)+(size_t)1, sizeof(char)) ;
      (void) strcpy (tokens[i], token) ;
      i++ ;
      s = (char*) NULL ;
   }
   
   return (tokens) ;
}

/*-----------------------------------------------------------------------------
 * Function  : free_*array
 *
 * Purpouse  : free memory allocated by an array of pointers
 *---------------------------------------------------------------------------*/

void free_list_names (char **names, int n)
{
   int     i ;
   if (names != (char**)NULL) {
      for (i = 0 ; i<n ; i++) if (names[i] != (char*)NULL) free (names[i]) ;
      free (names) ;
   }
   names = (char**)NULL ;
   return ;
}

char **free_carray (char **array, int n)
{
   int     i ;
   if (array != (char**)NULL) {
      for (i = 0 ; i < n ; i++) if (array[i] != (char*)NULL) free (array[i]) ;
      free (array) ;
   }
   return (char**)NULL ;
}

int **free_iarray (int **array, int n)
{
   int     i ;
   if (array != (int**)NULL) {
      for (i = 0 ; i < n ; i++) if (array[i] != (int*)NULL) free (array[i]) ;
      free (array) ;
   }
   return (int**)NULL ;
}

float **free_farray (float **array, int n)
{
   int     i ;
   if (array != (float**)NULL) {
      for (i = 0 ; i < n ; i++) if (array[i] != (float*)NULL) free (array[i]) ;
      free (array) ;
   }
   return (float**)NULL ;
}

double **free_darray (double **array, int n)
{
   int     i ;
   if (array != (double**)NULL) {
      for (i = 0 ; i < n ; i++) if (array[i] != (double*)NULL) free (array[i]) ;
      free (array) ;
   }
   return (double**)NULL ;
}

/*----------------------------------------------------------------------------
 * Function	:	strip_blanks()
 * In 		:	character string
 * Out 		:	void
 * Job		:	strip blanks ' ' off a character string
 * Notice	:	the input string is modified
 *--------------------------------------------------------------------------*/

void strip_blanks (char *expr)
{
   char  *tmp ;
   int    i, j ;

   tmp = (char*) calloc ((size_t)1+strlen (expr), sizeof(char)) ;
   j = 0 ;
   for (i = 0 ; i <= (int) strlen (expr) ; i++) {
      if (expr[i] != ' ') {
	 tmp[j] = expr[i] ;
	 j++ ;
      }	
   }
   (void) strcpy (expr, tmp) ;
   free (tmp) ;
}

/*----------------------------------------------------------------------------
 * Function	:	strip_newline()
 * In 		:	character string
 * Out 		:	void
 * Job		:	strip newline off a character string
 * Notice	:	the input string is modified
 *--------------------------------------------------------------------------*/

void strip_newlineold0(char *expr)
{
   char *ptr = expr ;
   while (*ptr) {
      if (isspace(*ptr)) *ptr = '\0' ;
      ptr++ ;
   }
}
void strip_newlineold(char *expr)
{
   char *ptr1 = expr ;
   char *ptr2 = expr ;
   while (*ptr1) {
      if (*ptr1 != '\n') 
         *ptr2++ = *ptr1++ ;
      else 
         *ptr2 = *++ptr1 ;
   }
   *ptr2 = '\0' ;
}

void strip_newline (char *expr)
{
   char * tmp ;
   int    i, j ;

   tmp = (char*) calloc ((size_t)1+strlen (expr), sizeof(char)) ;
   j = 0 ;
   for (i = 0 ; i <= (int) strlen (expr) ; i++) {
      if (expr[i] != '\n') {
         tmp[j] = expr[i] ;
         j++ ;
      }	
   }
   (void) strcpy (expr, tmp) ;
   free (tmp) ;
}

/*-----------------------------------------------------------------------------
 * Function  : Is_In
 *
 * Purpouse  : checks if number n is in the binary coded flag ns
 *---------------------------------------------------------------------------*/

int Is_In (unsigned int ns, unsigned int n)
{
   
   return bin_check_n(ns,(int)n);
   
   
   
   if (n >= 32 || n < 0)
      return 0 ;
   int is_in = 0 ;
   unsigned int i ;
   unsigned int j = n ;

   while (pow ((double)2.0, (double)j) <= ns) j++ ;
   while (j > n) {
      i = (unsigned int) pow ((double)2.0, (double)(--j)) ;
      if (i <= ns) ns -= i ;
   }
   if (ns >= (unsigned int) pow ((double)2.0, (double)(--n))) is_in = 1 ;
   return is_in ;
}

/*-----------------------------------------------------------------------------
 * Function  : bin_switch_on / bin_switch_off
 *
 * Purpouse  : switch on / off number nn in the binary coded flag ns
 *---------------------------------------------------------------------------*/

unsigned int 
bin_switch_on (unsigned int ns, int nn)
{
   if (nn >= 32 || nn < 0)
      return ns ;
   return (ns | (01 << nn)) ;
}

unsigned int 
bin_switch_off (unsigned int ns, int nn)
{
   if (nn >= 32 || nn < 0)
      return ns ;
   return (ns & ~(01 << nn)) ;
}

/*-----------------------------------------------------------------------------
 * Function  : bin_check_n
 *
 * Purpouse  : checks if number nn is in the binary coded flag ns
 *---------------------------------------------------------------------------*/

int 
bin_check_n (unsigned int ns, int nn)
{
   if (nn >= 32 || nn < 0)
      return 0 ;
   /*cp_message(1,0,"ns %u ; nn %d",ns,nn);*/
   ns >>= nn ;
   /*cp_message(0,0,"....  ns %u ; ns&01 %d %u\n",ns,(ns&01), (ns&01));*/
   return (ns & 01) ;
}

/*-----------------------------------------------------------------------------
 * Function  : bin_get_count
 *
 * Purpouse  : get the number of fields switched on in the binary coded flag ns
 *---------------------------------------------------------------------------*/

int 
bin_get_count (unsigned int ns)
{
   int  count ;
   for (count = 0; ns != 0; ns >>= 1)
      if (ns & 01)
         count++ ;
   return count ;
}

/*-----------------------------------------------------------------------------
 * Function  : bin_get_string
 *
 * Purpouse  : gives the binary representation of binary coded flag ns
 *---------------------------------------------------------------------------*/

char *
bin_get_string (char *user_buf, unsigned int ns)
{
   static char static_buf[33];
   char  *buf ;
   int    n ;

   if (user_buf != NULL)
      buf = user_buf ;
   else
      buf = static_buf ;

   for (n = 0; n < 32; n++)
      *(buf+n) = '0' ;
   *(buf+32) = '\0' ;

   for (n = 0; ns != 0; ns >>= 1) 
      *(buf + n++) |= (ns & 01) ;

   return buf ;
}

/*-----------------------------------------------------------------------------
 * Function  : old versions
 *
 * Purpouse  : 
 *---------------------------------------------------------------------------*/

int 
bin_check_nold(int ns, int nn)
{
   double dns = (double) ns ;
   int    n ;
   while (dns > 0) {
      n = (int) floor ( log(dns) / log(2.0) ) ;
      dns -= pow ((double)2.0, (double)n) ;
      if (n == nn)
         return 1 ;
   }
   return 0 ;
}

int 
bin_get_countold(int ns)
{
   double dns = (double) ns ;
   int    n, count = 0 ;
   while (dns > 0) {
      n = (int) floor ( log(dns) / log(2.0) ) ;
      dns -= pow ((double)2.0, (double)n) ;
      count++ ;
   }
   return count ;
}

char *
bin_get_stringold (char *user_buf, int ns)
{
   static char static_buf[33];
   char  *buf ;
   double dns = (double) ns ;
   int    n ;

   if (user_buf != NULL)
      buf = user_buf ;
   else
      buf = static_buf ;

   for (n = 0; n < 32; n++)
      *(buf+n) = '0' ;
   *(buf+32) = '\0' ;

   while (dns > 0) {
      n = (int) floor ( log(dns) / log(2.0) ) ;
      dns -= pow ((double)2.0, (double)n) ;
      *(buf+n) = '1' ;
   }
   return buf ;
}

/*-----------------------------------------------------------------------------
 * Function  : getNumbers
 *
 * Purpouse  : get list of numbers in expr like 1,3-5,7,...
 *---------------------------------------------------------------------------*/

int *getNumbers (char *expr, int *n)
{
   int    *ret = (int*)NULL ;
   char  **toks ;
   int     i, j, k, k1, k2, nn ;

   toks = tokenize (expr,",",&j) ;
   for (nn = i = 0; i < j; i++) {
      if (index (toks[i],'-') == (char*)NULL) {
         (void) sscanf (toks[i], "%d", &k1) ;
         ret = (int*) realloc (ret, (size_t)++nn*sizeof(int)) ;
         ret[nn-1] = k1 ;
      }
      else {
         (void) sscanf (toks[i], "%d-%d", &k1, &k2) ;
         if (k2 < k1) {
            free (ret) ;
            (void)free_carray (toks, j) ;
            return (int*)NULL ;
         }
         for (k = k1; k <= k2; k++) {
            ret = (int*) realloc (ret, (size_t)++nn*sizeof(int)) ;
            ret[nn-1] = k ;
         }
      }
   }
   (void) free_carray (toks, j) ;
   *n = nn ;

   return ret ;
}

char **getFormat (char *expr, int *nf)
{
   char  **ret = (char **)NULL ;
   int     i, sz, beg, n, sz1 ;
   char   *cur ;
   char   *truncexp ;
   char    form[] = "^%('|-|[+]| |#|0)*[0-9]*[.]?[0-9]*[h|l]?(d|i|u|o|x|X|f|e|E|g|G|c|C|s|S|p|n|%)$" ;
   boolean matched ;
   regex_t re_form ;

   *nf = 0 ;

   /* allocate all work strings	*/
   sz = (int) strlen (expr) ;
   if ((ret = (char**) calloc ((size_t)(sz), sizeof(char*))) == (char**)NULL) {
      return (char**)NULL ;
   }
   if ((cur = (char*) calloc ((size_t)(sz+1), sizeof(char))) == (char*)NULL) {
      free (ret) ;
      return (char**)NULL ;
   }
   if ((truncexp = (char*) calloc ((size_t)(sz+1), sizeof(char))) == (char*)NULL) {
      free (ret) ;
      free (cur) ;
      return (char**)NULL ;
   }

   /* Compile rule for format expression */
   if (regcomp (&re_form, &form[0], REG_EXTENDED|REG_NOSUB) != 0) {
      free (ret) ;
      free (cur) ;
      free (truncexp) ;
      return (char**)NULL ;
   }

   /*
    * The pattern identification is done in the following way:
    * try to match the entire string. if it matches, store it
    * as an identified token, else remove the rightmost character and
    * try again to match it. 
    */

   beg = n = sz1 = 0 ;
   while (beg < sz) {
      (void) memset ((char*)truncexp, (int)0, (size_t)sz) ;
      (void) memcpy ((char*)truncexp, (char*)expr+beg, (size_t)(sz-beg+1)) ;
      matched = FALSE ;
      for (i = (int) strlen (truncexp) ; i > 0 ; i--) {
         (void) memset ((char*)cur, (int)0, (size_t)sz) ;
         (void) strncpy ((char*)cur, (char*)truncexp, (size_t)i) ;

            /* Try to match a format expression */
         if (regexec (&re_form, cur, (size_t) 0, NULL, 0) == 0) {

            /* One match found */
            if (sz1 != 0) {
               n++ ;
               sz1 = 0 ;
            }
            ret[n] = (char*) calloc (strlen(cur), sizeof(char)) ;
            (void) strcpy (ret[n], cur) ;
            n++ ;
            ret[n] = (char*)NULL ;
            beg += (int) strlen (cur) ;
            matched = TRUE ;
            break ;
         }
      }
      if (matched == FALSE) {
         if (strcmp (cur, "%") == 0) {
            for (i = 0; i < n; i++) free (ret[i]) ;
            free (ret) ;
            ret = (char **)NULL ;
            goto exit ;
         }
         if (sz1 == 0) ret[n] = (char*)NULL ;
         sz1++ ;
         ret[n] = (char*) realloc (ret[n], (size_t)sz1*sizeof(char)) ;
         if (sz1 > 1) 
            (void) sprintf (ret[n], "%s%s", ret[n], cur) ;
         else 
            (void) sprintf (ret[n], "%s", cur) ;
         beg += (int) strlen (cur) ;
      }
   }
   if (sz1 != 0) n++ ;

   ret = (char**) realloc (ret, (size_t)n*sizeof(char*)) ;
   *nf = n ;

   exit:

   /* free regexp rules	*/
   regfree (&re_form) ;
   free (cur) ;
   free (truncexp) ;

   return (ret) ;
}

int is_integer (char *expr)
{
   int      ret = 0 ;
   char     number[] = "^([+-]?[0-9]+)$" ;
   regex_t  re_numb ;
   /* Compile rules for integer and real numbers matching */
   if (regcomp (&re_numb, &number[0], REG_EXTENDED|REG_NOSUB) != 0)
      return 2 ;
   /* Uses RegExpr matching rules for data type */
   if (regexec (&re_numb, expr, (size_t) 0, NULL, 0) == 0) 
      ret = 1 ;
   /* Free regexp rules	*/
   regfree (&re_numb) ;
   
   return ret ;
}

int
ndecimals (char *expr)
{
   char     *ptr1, *ptr2 ;
   if ((ptr1 = index (expr,'.')) == (char*)NULL ) 
      return 0 ;
   if ((ptr2 = index (expr,'E')) != (char*)NULL ) 
      return (int)(strlen (ptr1) - strlen(ptr2)) - 1 ;
   if ((ptr2 = index (expr,'e')) != (char*)NULL ) 
      return (int)(strlen (ptr1) - strlen (ptr2)) - 1 ;
   return (int) strlen (ptr1) - 1 ;
}

double 
my_ang_res_pos (double a)
{
   while (a < 0.0) a += TWOPI ;
   while (a >= TWOPI) a -= TWOPI ;
   return a ;
}
double 
my_ang_res_sym (double a)
{
   while (a <= -M_PI) a += TWOPI ;
   while (a > M_PI) a -= TWOPI ;
   return a ;
}

/* Generate a string containing the date and time and return a pointer
   to it.  If user_buf is not NULL, the string is stored in that
   buffer (21 bytes).  If strict is FALSE, the 'T' between the date
   and the time is replaced by a space, which makes the format a bit
   more readable (IMHO). */
char *
iso_8601_date_format (char *user_buf, int strict)
{
  static char static_buf[21] ;
  char *buf ;
  time_t clock ;
  struct tm *now ;

  if (user_buf != NULL)
    buf = user_buf ;
  else
    buf = static_buf ;
  clock = time (NULL) ;
  now = gmtime (&clock) ;
  /* date format derived from ISO 8601:1988 */
  sprintf (buf, "%04d-%02d-%02d%c%02d:%02d:%02d%c",
	   now->tm_year + 1900, now->tm_mon + 1, now->tm_mday,
	   (strict ? 'T' : ' '),
	   now->tm_hour, now->tm_min, now->tm_sec,
	   (strict ? 'Z' : '\000')) ;
  return buf ;
}
