/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include "cp_messages.h"

#include <glib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "cpk_env.h"

#ifdef G_OS_WIN32
#define STRICT
#include <windows.h>            /* For GetModuleFileName */
#include <io.h>
#ifndef S_IWUSR
# define S_IWUSR _S_IWRITE
#endif
#ifndef S_IWGRP
#define S_IWGRP (_S_IWRITE>>3)
#define S_IWOTH (_S_IWRITE>>6)
#endif
#ifndef S_ISDIR
# define __S_ISTYPE(mode, mask) (((mode) & _S_IFMT) == (mask))
# define S_ISDIR(mode)  __S_ISTYPE((mode), _S_IFDIR)
#endif
#define uid_t gint
#define gid_t gint
#define geteuid() 0
#define getegid() 0
#endif

#ifdef __EMX__
extern const char *__XOS2RedirRoot(const char *);
#endif

/**
 * cpk_data_directory:
 *
 * Returns the top directory for CataPack data. If the environment variable 
 * CPK_DATADIR exists, that is used.  It should be an absolute pathname.
 * Otherwise, on Unix the compile-time defined directory is used.  On
 * Win32, the installation directory as deduced from the executable's
 * name is used.
 *
 * The returned string is allocated just once, and should *NOT* be
 * freed with g_free().
 *
 * Returns: The top directory for CataPack data.
 **/

gchar *
cpk_data_directory (void)
{
  static gchar *cpk_data_dir = NULL;
         gchar *env_cpk_data_dir = NULL;
  
  if (cpk_data_dir != NULL)
    return cpk_data_dir;

  env_cpk_data_dir = g_getenv ("CPK_DATADIR");

   if (NULL != env_cpk_data_dir) {
      if (!g_path_is_absolute (env_cpk_data_dir))
        cp_error ("CPK_DATADIR environment variable should be an absolute path.");
#ifndef __EMX__
      cpk_data_dir = g_strdup (env_cpk_data_dir);
#else      
      cpk_data_dir = g_strdup (__XOS2RedirRoot(env_cpk_data_dir));
#endif      
    }
    else {
#ifndef G_OS_WIN32
#ifndef __EMX__
      cpk_data_dir = DATADIR;
#else
      cpk_data_dir = g_strdup(__XOS2RedirRoot(DATADIR));
#endif
#else
      /* Figure it out from the executable name */
      gchar filename[MAX_PATH];
      gchar *sep1, *sep2;

      if (GetModuleFileName (NULL, filename, sizeof (filename)) == 0)
        cp_error ("GetModuleFilename failed\n");
      
      /* If the executable file name is of the format
       * <foobar>\bin\cpk.exe of <foobar>\plug-ins\filter.exe, * use
       * <foobar>. Otherwise, use the directory where the executable
       * is.
       */

      sep1 = strrchr (filename, G_DIR_SEPARATOR);

      *sep1 = '\0';

      sep2 = strrchr (filename, G_DIR_SEPARATOR);

      if (sep2 != NULL) {
         if (g_strcasecmp (sep2 + 1, "bin") == 0 || 
             g_strcasecmp (sep2 + 1, "plug-ins") == 0)
            *sep2 = '\0';
      }

      cpk_data_dir = g_strdup (filename);
#endif
    }
  return cpk_data_dir;
}
      

/**
 * cpk_gtkrc:
 *
 * Returns the name of the CataPack's application-specific gtkrc file.
 *
 * The returned string is allocated just once, and should *NOT* be
 * freed with g_free().
 *
 * Returns: The name of the CataPack's application-specific gtkrc file.
 **/ 
gchar*
cpk_gtkrc (void)
{
   static gchar *cpk_gtkrc_filename = NULL;

   if (cpk_gtkrc_filename != NULL)
      return cpk_gtkrc_filename ;

   cpk_gtkrc_filename = g_strconcat (cpk_sysconf_directory (),
				     G_DIR_SEPARATOR_S,
				     "gtkrc",
				     NULL);
   return cpk_gtkrc_filename ;
}

/**
 * cpk_sysconf_directory:
 *
 * Returns the top directory for CataPack config files. If the environment
 * variable CPK_SYSCONFDIR exists, that is used.  It should be an
 * absolute pathname.  Otherwise, on Unix the compile-time defined
 * directory is used.  On Win32, the installation directory as deduced
 * from the executable's name is used.
 *
 * The returned string is allocated just once, and should *NOT* be
 * freed with g_free().
 *
 * Returns: The top directory for CataPack config files.
 **/
gchar*
cpk_sysconf_directory (void)
{
   static gchar *cpk_sysconf_dir = NULL ;
   gchar *env_cpk_sysconf_dir = NULL ;
  
   if (cpk_sysconf_dir != NULL)
      return cpk_sysconf_dir ;

   env_cpk_sysconf_dir = g_getenv ("CPK_SYSCONFDIR") ;

   if (NULL != env_cpk_sysconf_dir) {
      if (!g_path_is_absolute (env_cpk_sysconf_dir))
         cp_error ("CPK_SYSCONFDIR environment variable should be an absolute path.") ;
#ifndef __EMX__
      cpk_sysconf_dir = g_strdup (env_cpk_sysconf_dir) ;
#else      
      cpk_sysconf_dir = g_strdup (__XOS2RedirRoot(env_cpk_sysconf_dir)) ;
#endif      
   }
   else {
#ifndef G_OS_WIN32
#ifndef __EMX__
      cpk_sysconf_dir = SYSCONFDIR ;
#else
      cpk_sysconf_dir = g_strdup(__XOS2RedirRoot(SYSCONFDIR)) ;
#endif
#else
      /* Figure it out from the executable name */
      gchar filename[MAX_PATH] ;
      gchar *sep1, *sep2 ;

      if (GetModuleFileName (NULL, filename, sizeof (filename)) == 0)
         cp_error ("GetModuleFilename failed\n") ;
      
      /* If the executable file name is of the format
       * <foobar>\bin\cpk.exe of <foobar>\plug-ins\filter.exe, * use
       * <foobar>. Otherwise, use the directory where the executable
       * is.
       */

      sep1 = strrchr (filename, G_DIR_SEPARATOR) ;

      *sep1 = '\0' ;

      sep2 = strrchr (filename, G_DIR_SEPARATOR) ;

      if (sep2 != NULL) {
         if (g_strcasecmp (sep2 + 1, "bin") == 0 || 
	     g_strcasecmp (sep2 + 1, "plug-ins") == 0)
            *sep2 = '\0';
      }

      cpk_sysconf_dir = g_strdup (filename) ;
#endif
   }
   return cpk_sysconf_dir ;
}

