/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <math.h>
#include <gtk/gtk.h>

#include "cpk_frame.h"
#include "cpk_plot.h"
#include "cpk_gtk_util.h"

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/


/* Local data */

static GtkContainerClass *parent_class1 = NULL;
static GtkContainerClass *parent_class2 = NULL;
static guint              frame_signals[FRAME_LAST_SIGNAL] = { 0 };
static guint              frame_set_signals[FRAME_SET_LAST_SIGNAL] = { 0 };

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static void  cpk_frame_class_init      (CpkFrameClass     *klass) ;
static void  cpk_frame_init            (CpkFrame          *frame) ;
static gint
cpk_frame_expose (GtkWidget *widget, GdkEventExpose *event) ;
static gint
cpk_frame_configure (GtkWidget *widget, GdkEventConfigure *event) ;

static gint  cpk_frame_button_press    (GtkWidget         *widget, 
                                        GdkEventButton    *event) ;
static gint  cpk_frame_button_release  (GtkWidget         *widget, 
                                        GdkEventButton    *event) ;
static gint  cpk_frame_enter_notify    (GtkWidget         *widget, 
                                        GdkEventCrossing  *event) ;
static gint  cpk_frame_leave_notify    (GtkWidget         *widget, 
                                        GdkEventCrossing  *event) ;
static gint  cpk_frame_motion_notify   (GtkWidget         *widget,  
                                        GdkEventMotion    *event) ;
static guint cpk_frame_n_active        (CpkFrame          *frame) ;
static void  cpk_frame_focus           (CpkFrame          *frame) ;
static void  cpk_frame_select          (CpkFrame          *frame) ;


static void  cpk_frame_set_class_init  (CpkFrameSetClass  *klass) ;
static void  cpk_frame_set_init        (CpkFrameSet       *frame) ;

static void  cpk_frame_reparent_box    (GtkWidget         *new_parent, 
                                        CpkFrameSet       *fs) ;
static void  cpk_frame_reparent_table  (CpkFrameSet       *fs) ;
static void  cpk_frame_set_added       (GtkWidget         *w,
                                        CpkFrameSet       *fs) ;
static void  cpk_frame_set_removed     (GtkWidget         *w,
                                        CpkFrameSet       *fs) ;
static void  cpk_frame_set_new_focus   (GtkWidget         *w,
                                        CpkFrameSet       *fs) ;

static void  cpk_frame_set_send_expose (CpkFrameSet *fs) ;

static gint  cpk_frame_set_pan_configure (GtkWidget         *widget, 
                                          GdkEventConfigure *event, 
                                          CpkFrameSet       *fs) ;

static gint  cpk_frame_set_pan_expose    (GtkWidget         *widget, 
                                          GdkEventExpose    *event, 
                                          CpkFrameSet       *fs) ;

static gint  cpk_frame_set_mgf_configure (GtkWidget         *widget, 
                                          GdkEventConfigure *event, 
                                          CpkFrameSet       *fs) ;

static gint  cpk_frame_set_mgf_expose    (GtkWidget         *widget, 
                                          GdkEventExpose    *event, 
                                          CpkFrameSet       *fs) ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

GtkType
cpk_frame_get_type (void)
{
  static GtkType frame_type = 0;

  if (!frame_type)
    {
      static const GtkTypeInfo frame_info =
      {
        "CpkFrame",
        sizeof (CpkFrame),
        sizeof (CpkFrameClass),
        (GtkClassInitFunc) cpk_frame_class_init,
        (GtkObjectInitFunc) cpk_frame_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      frame_type = gtk_type_unique (GTK_TYPE_FRAME, &frame_info);
    }
  return frame_type;
}

static void
cpk_frame_class_init (CpkFrameClass *class)
{
   GtkObjectClass    *object_class;
   GtkWidgetClass    *widget_class;
   GtkContainerClass *container_class;

   object_class = (GtkObjectClass*) class;
   widget_class = (GtkWidgetClass*) class;
   container_class = (GtkContainerClass*) class;

   parent_class1 = gtk_type_class (GTK_TYPE_FRAME);

   frame_signals[FRAME_ADDED] =
         gtk_signal_new ("frame-added",
         GTK_RUN_FIRST,
         object_class->type,
         GTK_SIGNAL_OFFSET (CpkFrameClass, frame_added),
         gtk_marshal_NONE__NONE,
	 GTK_TYPE_NONE, 0);
   frame_signals[FRAME_REMOVED] =
         gtk_signal_new ("frame-removed",
         GTK_RUN_FIRST,
         object_class->type,
         GTK_SIGNAL_OFFSET (CpkFrameClass, frame_removed),
         gtk_marshal_NONE__NONE,
	 GTK_TYPE_NONE, 0);
   frame_signals[FRAME_FOCUS] =
         gtk_signal_new ("frame-focus",
         GTK_RUN_FIRST,
         object_class->type,
         GTK_SIGNAL_OFFSET (CpkFrameClass, frame_focus),
         gtk_marshal_NONE__NONE,
	 GTK_TYPE_NONE, 0);
   frame_signals[FRAME_SELECT] =
         gtk_signal_new ("frame-select",
         GTK_RUN_FIRST,
         object_class->type,
         GTK_SIGNAL_OFFSET (CpkFrameClass, frame_select),
         gtk_marshal_NONE__NONE,
	 GTK_TYPE_NONE, 0);
   gtk_object_class_add_signals (object_class, frame_signals, 
         FRAME_LAST_SIGNAL);

   class->frame_focus = NULL ;
}

static void
cpk_frame_init (CpkFrame *frame)
{
   frame->area = NULL ;
   frame->plot  = NULL ;
   frame->hrule  = NULL ;
   frame->vrule  = NULL ;
   frame->show  = TRUE ;
   frame->prev = NULL ;
   frame->next  = NULL ;
}

static gint
cpk_frame_expose (GtkWidget *widget, GdkEventExpose *event)
{
   CpkFrame *frame ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_FRAME (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
   CPKDEBUG5 ("cpk_frame_expose", 4) ;

   frame = CPK_FRAME (widget) ;
   gtk_widget_event (frame->plot, (GdkEvent*) event) ;

   return FALSE ;
}


static gint
cpk_frame_configure (GtkWidget *widget, GdkEventConfigure *event)
{
   CpkFrame *frame ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_FRAME (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
   CPKDEBUG5 ("cpk_frame_configure", 4) ;

   frame = CPK_FRAME (widget) ;
   gtk_widget_event (frame->plot, (GdkEvent*) event) ;

   return FALSE ;
}


static gint
cpk_frame_enter_notify (GtkWidget *widget, GdkEventCrossing *event)
{
   CpkFrame *frame ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_FRAME (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
   CPKDEBUG5 ("cpk_frame_enter_notify", 6) ;

   frame = CPK_FRAME (widget) ;
   cpk_frame_focus (frame) ;
   gtk_widget_event (frame->plot, (GdkEvent*) event) ;

   return FALSE ;
}

static gint
cpk_frame_leave_notify (GtkWidget *widget, GdkEventCrossing *event)
{
   CpkFrame *frame ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_FRAME (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
   CPKDEBUG5 ("cpk_frame_leave_notify", 6) ;

   frame = CPK_FRAME (widget) ;
   gtk_widget_event (frame->plot, (GdkEvent*) event) ;

   return FALSE ;
}

static gint
cpk_frame_motion_notify (GtkWidget *widget, GdkEventMotion *event)
{
   CpkFrame *frame ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_FRAME (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
   CPKDEBUG5 ("cpk_frame_motion_notify", 7) ;

   frame = CPK_FRAME (widget) ;
   gtk_widget_event (frame->plot, (GdkEvent*) event) ;

   return FALSE ;
}

static gint
cpk_frame_button_press (GtkWidget *widget, GdkEventButton *event)
{
   CpkFrame *frame ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_FRAME (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
   CPKDEBUG5 ("cpk_frame_button_press", 6) ;

   if (!GTK_WIDGET_HAS_FOCUS (widget))
      gtk_widget_grab_focus (widget) ;
   frame = CPK_FRAME (widget) ;
   cpk_frame_select (frame) ;
   gtk_widget_event (frame->plot, (GdkEvent*) event) ;

   return FALSE ;
}

static gint
cpk_frame_button_release (GtkWidget *widget, GdkEventButton *event)
{
   CpkFrame *frame ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_FRAME (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
   CPKDEBUG5 ("cpk_frame_button_release", 6) ;

   frame = CPK_FRAME (widget) ;
   gtk_widget_event (frame->plot, (GdkEvent*) event) ;

   return FALSE ;
}

static guint
cpk_frame_n_active (CpkFrame *frame)
{
   guint nf = 0 ;

   g_return_val_if_fail (CPK_IS_FRAME (frame), 0) ;
   CPKDEBUG5 ("cpk_frame_n_active", 6) ;

   while (frame) {
      if (frame->show)
         nf++ ;
      frame = frame->next ;
   }
   printf("nf %d\n",nf);
   return nf ;
}

static void
cpk_frame_focus (CpkFrame *frame)
{
   CPKDEBUG5 ("cpk_frame_focus", 6) ;
   gtk_signal_emit_by_name (GTK_OBJECT (frame), "frame-focus") ;
}

static void
cpk_frame_select (CpkFrame *frame)
{
   CPKDEBUG5 ("cpk_frame_select", 5) ;
   gtk_signal_emit_by_name (GTK_OBJECT (frame), "frame-select") ;
}

/*----------------------------------------------------------------------------
 *							CpkFrameSet functions
 *--------------------------------------------------------------------------*/

GtkType
cpk_frame_set_get_type (void)
{
  static GtkType frame_set_type = 0;

  if (!frame_set_type)
    {
      static const GtkTypeInfo frame_set_info =
      {
        "CpkFrameSet",
        sizeof (CpkFrameSet),
        sizeof (CpkFrameSetClass),
        (GtkClassInitFunc) cpk_frame_set_class_init,
        (GtkObjectInitFunc) cpk_frame_set_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      frame_set_type = gtk_type_unique (GTK_TYPE_HBOX, &frame_set_info);
    }
  return frame_set_type;
}

static void
cpk_frame_set_class_init (CpkFrameSetClass *class)
{
   GtkObjectClass    *object_class;
   GtkWidgetClass    *widget_class;
   GtkContainerClass *container_class;

   object_class = (GtkObjectClass*) class;
   widget_class = (GtkWidgetClass*) class;
   container_class = (GtkContainerClass*) class;

   parent_class2 = gtk_type_class (GTK_TYPE_HBOX);

   frame_set_signals[FRAME_SET_FIRST_SIGNAL] = 
         gtk_signal_new ("cazcaz",
         GTK_RUN_FIRST,
         object_class->type,
         GTK_SIGNAL_OFFSET (CpkFrameSetClass, cazcaz),
         gtk_marshal_NONE__NONE,
         GTK_TYPE_NONE, 0);
   gtk_object_class_add_signals (object_class, frame_set_signals, 
         FRAME_SET_LAST_SIGNAL);

   /* widget_class->button_press_event = cpk_frame_set_button_press ;
   widget_class->button_release_event = cpk_frame_set_button_release ;
   widget_class->enter_notify_event = cpk_frame_set_enter_notify ;
   widget_class->leave_notify_event = cpk_frame_set_leave_notify ;
   widget_class->motion_notify_event = cpk_frame_set_motion_notify ; */
}

static void
cpk_frame_set_init (CpkFrameSet *frame_set)
{
   frame_set->paner_box = NULL ;
   frame_set->paner = NULL ;
   frame_set->magnifier_box = NULL ;
   frame_set->magnifier = NULL ;
   frame_set->mgf_factor = 0 ;
   frame_set->mgf_timer = 0 ;
   frame_set->frame_box = NULL ;
   frame_set->frame_hbox = NULL ;
   frame_set->frame_vbox = NULL ;
   frame_set->frame_tbox = NULL ;
   frame_set->nframes = 0 ;
   frame_set->mode = 0 ;
   frame_set->tile_mode = 0 ;
   frame_set->last_tile_mode = 0 ;
   frame_set->cursor = 0 ;
   frame_set->p_action = 0 ;
   frame_set->frames = NULL ;
   frame_set->gc = NULL ;
   frame_set->focused = NULL ;
   frame_set->reference = NULL ;
   frame_set->current = NULL ;
   frame_set->old_current = NULL ;
}

static void
cpk_frame_set_added (GtkWidget *w, CpkFrameSet *fs)
{
   CpkFrame *last = cpk_frame_last (fs->frames) ;
   CPKDEBUG5 ("cpk_frame_set_added", 5) ;

   fs->nframes++ ;
   gtk_signal_connect (GTK_OBJECT (last), "frame-added",
         GTK_SIGNAL_FUNC (cpk_frame_set_added), GTK_OBJECT (fs)) ;
   gtk_signal_connect (GTK_OBJECT (last), "frame-removed",
         GTK_SIGNAL_FUNC (cpk_frame_set_removed), GTK_OBJECT (fs)) ;
   gtk_signal_connect (GTK_OBJECT (last), "frame-focus", 
         GTK_SIGNAL_FUNC (cpk_frame_set_new_focus), GTK_OBJECT (fs)) ;
   gtk_signal_connect (GTK_OBJECT (last), "frame-select", 
         GTK_SIGNAL_FUNC (cpk_frame_set_select), GTK_OBJECT (fs)) ;

   cpk_frame_set_reparent (fs) ;
}

static void
cpk_frame_set_removed (GtkWidget *w, CpkFrameSet *fs)
{
   CPKDEBUG5 ("cpk_frame_set_removed", 5) ;
   fs->nframes-- ;
   cpk_frame_set_reparent (fs) ;
}

static void
cpk_frame_set_new_focus (GtkWidget *w, CpkFrameSet *fs)
{
   CPKDEBUG5 ("cpk_frame_set_new_focus", 6) ;
   fs->focused = CPK_FRAME (w) ;
   gtk_widget_draw (GTK_WIDGET (fs->magnifier), NULL) ;
}

static void
cpk_frame_reparent_box (GtkWidget *new_parent, CpkFrameSet *fs)
{
   GtkWidget *old_parent = NULL ;
   CpkFrame  *frame = fs->frames ;
   CPKDEBUG5 ("cpk_frame_reparent_box", 6) ;

   if (!new_parent)
      return ;

   while (frame) {
      old_parent = GTK_WIDGET (frame)->parent ;
      if (old_parent) {
         gtk_widget_ref (GTK_WIDGET (frame)) ;
         gtk_container_remove (GTK_CONTAINER (old_parent), GTK_WIDGET (frame)) ;
      }
      gtk_container_add (GTK_CONTAINER (new_parent), GTK_WIDGET (frame)) ;
      if (old_parent) 
         gtk_widget_unref (GTK_WIDGET (frame)) ;

      /* Show/hide frames depending on current frame mode */
      switch  (fs->mode) {

         case FRAME_SINGLE :
            if (frame == fs->current)
               gtk_widget_show_all (GTK_WIDGET (frame)) ;
            else
               gtk_widget_hide (GTK_WIDGET (frame)) ;
            break ;

         case FRAME_PAIR :
            if (frame == fs->reference || frame == fs->current)
               gtk_widget_show (GTK_WIDGET (frame)) ;
            else
               gtk_widget_hide (GTK_WIDGET (frame)) ;
            gtk_widget_hide (frame->hrule) ;
            gtk_widget_hide (frame->vrule) ;
            break ;

         case FRAME_TILE :
            gtk_widget_show (GTK_WIDGET (frame)) ;
            gtk_widget_hide (frame->hrule) ;
            gtk_widget_hide (frame->vrule) ;
            break ;

      }
      frame = frame->next ;
   }
   gtk_widget_show (new_parent) ;
}

static void
cpk_frame_reparent_table (CpkFrameSet *fs)
{
   GtkWidget *old_parent ;
   CpkFrame  *frame = fs->frames ;
   guint      nc, nr, nf, nrow, ncol ;
   CPKDEBUG5 ("cpk_frame_reparent_table", 5) ;

   nf = cpk_frame_n_active (fs->frames) ;
   ncol = ceil (sqrt ((gdouble)nf)) ;
   nrow = ceil (nf / (gdouble)ncol) ;

   /*** FIXME: table should be resized before while() cicle ? */
   /*   If we do it here the table is not correctly resized when number 
    *   of columns is decreasing */
   nc = nr = 0 ;
   while (frame) {
      old_parent = GTK_WIDGET (frame)->parent ;
      if (old_parent) {
         gtk_widget_ref (GTK_WIDGET (frame)) ;
         gtk_container_remove (GTK_CONTAINER (old_parent), GTK_WIDGET (frame)) ;
      }
      gtk_table_attach (GTK_TABLE (fs->frame_tbox), 
            GTK_WIDGET (frame), nc, nc+1, nr, nr+1,
            GTK_SHRINK | GTK_EXPAND | GTK_FILL, 
            GTK_SHRINK | GTK_EXPAND | GTK_FILL, 0, 0) ;
      if (frame->show) {
         gtk_widget_show_all (GTK_WIDGET (frame)) ;
         gtk_widget_hide (frame->vrule) ;
         gtk_widget_hide (frame->hrule) ;
         if (nc < ncol-1)
            nc++ ;
         else {
            nc = 0 ;
            nr++ ;
         }
      }
      if (old_parent) 
         gtk_widget_unref (GTK_WIDGET (frame)) ;
      frame = frame->next ;
   }
   gtk_table_resize (GTK_TABLE (fs->frame_tbox), nrow, ncol) ;
   gtk_widget_show (fs->frame_tbox) ;
}

static void
cpk_frame_set_send_expose (CpkFrameSet *fs)
{
  GtkWidget *widget;
  GdkEventExpose event;

  widget = GTK_WIDGET (fs);

  event.type = GDK_EXPOSE;
  event.window = widget->window;
  event.send_event = TRUE;
  event.area.x = widget->allocation.x;
  event.area.y = widget->allocation.y;
  event.area.width = widget->allocation.width;
  event.area.height = widget->allocation.height;
  event.count  = 0 ;
  
  gtk_widget_event (widget, (GdkEvent*) &event);
}

static gint
cpk_frame_set_pan_configure (GtkWidget         *widget, 
                             GdkEventConfigure *event, 
                             CpkFrameSet       *fs)
{
   return cpk_plot_pan_configure (widget, event, CPK_PLOT (fs->current->plot)) ;
}

static gint
cpk_frame_set_pan_expose (GtkWidget         *widget, 
                          GdkEventExpose    *event, 
                          CpkFrameSet       *fs)
{
   return cpk_plot_pan_expose (widget, event, CPK_PLOT (fs->current->plot)) ;
}

static gint
cpk_frame_set_mgf_configure (GtkWidget         *widget, 
                             GdkEventConfigure *event, 
                             CpkFrameSet       *fs)
{
   return cpk_plot_mgf_configure (widget, event, CPK_PLOT (fs->focused->plot)) ;
}

static gint
cpk_frame_set_mgf_expose (GtkWidget         *widget, 
                          GdkEventExpose    *event, 
                          CpkFrameSet       *fs)
{
   return cpk_plot_mgf_expose (widget, event, CPK_PLOT (fs->focused->plot)) ;
}



/*----------------------------------------------------------------------------
 *							Public interface
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 *							CpkFrame functions
 *--------------------------------------------------------------------------*/
static gint cc = 0 ;

GtkWidget *
cpk_frame_new (void)
{
   GtkWidget   *frame, *table ;
   gchar buf[10];
   CPKDEBUG5 ("cpk_frame_new", 5) ;

   sprintf (buf, "%d", ++cc) ;

   frame =  GTK_WIDGET (gtk_type_new (TYPE_CPK_FRAME)) ;

   gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN) ;
   gtk_container_set_border_width (GTK_CONTAINER (frame), 0) ;

   table = gtk_table_new (3, 2, FALSE) ;
   gtk_container_add (GTK_CONTAINER (frame), table) ;

   CPK_FRAME (frame)->area = gtk_drawing_area_new () ;
   CPK_FRAME (frame)->plot = cpk_plot_new (CPK_FRAME (frame)->area, buf) ;
   gtk_table_attach (GTK_TABLE (table), CPK_FRAME (frame)->area, 1, 2, 0, 1, 
         GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0) ;

   CPK_FRAME (frame)->hrule = gtk_hruler_new () ;
   gtk_ruler_set_metric (GTK_RULER (CPK_FRAME (frame)->hrule), GTK_PIXELS) ;
gtk_ruler_set_range( GTK_RULER(CPK_FRAME (frame)->hrule), 0, 0.6, 0.5, 100);
   gtk_table_attach (GTK_TABLE (table), CPK_FRAME (frame)->hrule, 1, 2, 1, 2, 
         GTK_EXPAND | GTK_SHRINK | GTK_FILL, GTK_FILL, 0, 0) ;
   CPK_FRAME (frame)->vrule = gtk_vruler_new () ;
   gtk_ruler_set_metric (GTK_RULER (CPK_FRAME (frame)->vrule), GTK_PIXELS) ;
gtk_ruler_set_range( GTK_RULER(CPK_FRAME (frame)->vrule), 100, 0, 0.5, 0.4);
   gtk_table_attach (GTK_TABLE (table), CPK_FRAME (frame)->vrule, 0, 1, 0, 1,
         GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0) ;

   CPK_PLOT_SET_FLAGS (CPK_FRAME (frame)->plot, CPK_LABEL) ;

   gtk_signal_connect_object (GTK_OBJECT (CPK_FRAME (frame)->area), 
         "expose_event", GTK_SIGNAL_FUNC (cpk_frame_expose), 
         GTK_OBJECT (CPK_FRAME (frame))) ;
   gtk_signal_connect_object (GTK_OBJECT (CPK_FRAME (frame)->area), 
         "configure_event", GTK_SIGNAL_FUNC (cpk_frame_configure), 
         GTK_OBJECT (CPK_FRAME (frame))) ;

   gtk_signal_connect_object (GTK_OBJECT (CPK_FRAME (frame)->area), 
         "enter_notify_event", GTK_SIGNAL_FUNC (cpk_frame_enter_notify), 
         GTK_OBJECT (CPK_FRAME (frame))) ;
   gtk_signal_connect_object (GTK_OBJECT (CPK_FRAME (frame)->area), 
         "leave_notify_event", GTK_SIGNAL_FUNC (cpk_frame_leave_notify), 
         GTK_OBJECT (CPK_FRAME (frame))) ;
   gtk_signal_connect_object (GTK_OBJECT (CPK_FRAME (frame)->area), 
         "button_press_event", GTK_SIGNAL_FUNC (cpk_frame_button_press), 
         GTK_OBJECT (CPK_FRAME (frame))) ;
   gtk_signal_connect_object (GTK_OBJECT (CPK_FRAME (frame)->area), 
         "button_release_event", GTK_SIGNAL_FUNC (cpk_frame_button_release), 
         GTK_OBJECT (CPK_FRAME (frame))) ;
   gtk_signal_connect_object (GTK_OBJECT (CPK_FRAME (frame)->area), 
         "motion_notify_event", GTK_SIGNAL_FUNC (cpk_frame_motion_notify), 
         GTK_OBJECT (CPK_FRAME (frame))) ;

   GTK_WIDGET_SET_FLAGS (CPK_FRAME (frame)->area, GTK_CAN_FOCUS) ;

   gtk_widget_set_events (CPK_FRAME (frame)->area, GDK_EXPOSURE_MASK
         | GDK_ENTER_NOTIFY_MASK
         | GDK_LEAVE_NOTIFY_MASK
         | GDK_BUTTON_PRESS_MASK
         | GDK_BUTTON_RELEASE_MASK
         | GDK_POINTER_MOTION_MASK
         | GDK_POINTER_MOTION_HINT_MASK) ;

   /* gtk_container_add (GTK_CONTAINER (frame), CPK_FRAME (frame)->area) ; */
   gtk_widget_show_all (frame) ;

   return frame ;
}

void
cpk_frame_destroy (CpkFrame *frame)
{
   CPKDEBUG5 ("cpk_frame_destroy", 5);
   gtk_widget_destroy (frame->area) ;
   gtk_widget_destroy (GTK_WIDGET (frame)) ;
}

CpkFrame *
cpk_frame_first (CpkFrame *frame)
{
   g_return_val_if_fail (CPK_IS_FRAME (frame), FALSE) ;
   CPKDEBUG5 ("cpk_frame_first", 5);

   while (frame->prev)
      frame = frame->prev ;
   return frame ;
}

CpkFrame *
cpk_frame_prev (CpkFrame *frame)
{
   g_return_val_if_fail (CPK_IS_FRAME (frame), FALSE) ;
   CPKDEBUG5 ("cpk_frame_prev", 5) ;

   if (frame->prev)
      return frame->prev ;
   else
      return cpk_frame_last (frame) ;
}

CpkFrame *
cpk_frame_next (CpkFrame *frame)
{
   g_return_val_if_fail (CPK_IS_FRAME (frame), FALSE) ;
   CPKDEBUG5 ("cpk_frame_next", 5) ;

   if (frame->next)
      return frame->next ;
   else
      return cpk_frame_first (frame) ;
}

CpkFrame *
cpk_frame_last (CpkFrame *frame)
{
   g_return_val_if_fail (CPK_IS_FRAME (frame), FALSE) ;
   CPKDEBUG5 ("cpk_frame_last", 5) ;

   while (frame->next)
      frame = frame->next ;
   return frame ;
}

CpkFrame *
cpk_frame_add (CpkFrame *f)
{
   CpkFrame *frame ;
   g_return_val_if_fail (CPK_IS_FRAME (f), NULL) ;
   CPKDEBUG5 ("cpk_frame_add", 5) ;

   frame = cpk_frame_last (f) ;
   frame->next =  CPK_FRAME (cpk_frame_new ()) ;
   (frame->next)->prev =  frame ;
   gtk_signal_emit_by_name (GTK_OBJECT (f), "frame-added") ;
   return frame->next ;
}

CpkFrame *
cpk_frame_remove (CpkFrame *f)
{
   CpkFrame *frame = f;
   CpkFrame *prev = frame->prev ;
   CpkFrame *next = frame->next ;
   g_return_val_if_fail (CPK_IS_FRAME (f), NULL);
   CPKDEBUG5 ("cpk_frame_remove", 5) ;

   f = cpk_frame_first (frame) ;

   if (f == frame)
      f = next ;

   cpk_frame_destroy (frame) ;

   if (prev)
      prev->next = next ;
   if (next)
      next->prev = prev ;
   if (f)
      gtk_signal_emit_by_name (GTK_OBJECT (f), "frame-removed") ;

   if (prev)
      return prev ;
   else
      return next ;
}

/*----------------------------------------------------------------------------
 *							CpkFrameSet functions
 *--------------------------------------------------------------------------*/

GtkWidget *
cpk_frame_set_new (void)
{
   GtkWidget   *frame_set ;
   CpkFrameSet *fs ;

   frame_set =  GTK_WIDGET (gtk_type_new (TYPE_CPK_FRAME_SET)) ;
   fs = CPK_FRAME_SET (frame_set) ;

   fs->paner_box = gtk_frame_new (NULL) ;
   gtk_frame_set_shadow_type (GTK_FRAME (fs->paner_box), GTK_SHADOW_IN) ;
   gtk_widget_set_usize (fs->paner_box, 120, 120) ;
   gtk_container_set_border_width (GTK_CONTAINER (fs->paner_box), 0) ;
   fs->paner = gtk_drawing_area_new () ;
   gtk_container_add (GTK_CONTAINER (fs->paner_box), fs->paner) ;   
   gtk_widget_show_all (fs->paner_box) ;

   fs->magnifier_box = gtk_frame_new (NULL) ;
   gtk_frame_set_shadow_type (GTK_FRAME (fs->magnifier_box), GTK_SHADOW_IN) ;
   gtk_widget_set_usize (fs->magnifier_box, 120, 120) ;
   gtk_container_set_border_width (GTK_CONTAINER (fs->magnifier_box), 0) ;
   fs->magnifier = gtk_drawing_area_new () ;
   gtk_container_add (GTK_CONTAINER (fs->magnifier_box), fs->magnifier) ;   
   gtk_widget_show_all (fs->magnifier_box) ;
   fs->mgf_factor = 4 ;

   gtk_signal_connect (GTK_OBJECT (fs->paner), "expose_event", 
         GTK_SIGNAL_FUNC (cpk_frame_set_pan_expose), GTK_OBJECT (fs)) ;
   gtk_signal_connect (GTK_OBJECT (fs->paner), "configure_event", 
         GTK_SIGNAL_FUNC (cpk_frame_set_pan_configure), GTK_OBJECT (fs)) ;
   gtk_signal_connect (GTK_OBJECT (fs->magnifier), "expose_event", 
         GTK_SIGNAL_FUNC (cpk_frame_set_mgf_expose), GTK_OBJECT (fs)) ;
   gtk_signal_connect (GTK_OBJECT (fs->magnifier), "configure_event", 
         GTK_SIGNAL_FUNC (cpk_frame_set_mgf_configure), GTK_OBJECT (fs)) ;

   fs->nframes = 0 ;

   fs->frame_hbox = gtk_hbox_new (TRUE, 2) ;
   fs->frame_vbox = gtk_vbox_new (TRUE, 2) ;
   fs->frame_tbox = gtk_table_new (1, 1, TRUE) ;
   gtk_table_set_row_spacings (GTK_TABLE (fs->frame_tbox), 2) ;
   gtk_table_set_col_spacings (GTK_TABLE (fs->frame_tbox), 2) ;
   gtk_container_add (GTK_CONTAINER (fs), fs->frame_hbox) ;
   gtk_container_add (GTK_CONTAINER (fs), fs->frame_vbox) ;
   gtk_container_add (GTK_CONTAINER (fs), fs->frame_tbox) ;
   fs->mode = FRAME_TILE ;
   fs->tile_mode = TILE_GRID ;
   fs->last_tile_mode = TILE_COLUMNS ;

   /* Commentare se si vuole iniziare senza frames */
   cpk_frame_set_select (GTK_WIDGET (cpk_frame_set_initialize (fs)), fs) ;

   return frame_set ;
}

CpkFrame *
cpk_frame_set_initialize (CpkFrameSet *fs)
{
   fs->frames = fs->reference = fs->current = fs->old_current = fs->focused = 
         CPK_FRAME (cpk_frame_new ()) ;
   cpk_frame_set_added (NULL, fs) ;
   return fs->frames ;
}

void
cpk_frame_set_reparent (CpkFrameSet *fs)
{
   CPKDEBUG5 ("cpk_frame_set_reparent", 5) ;
   gtk_widget_hide (fs->frame_hbox) ;
   gtk_widget_hide (fs->frame_vbox) ;
   gtk_widget_hide (fs->frame_tbox) ;
   if (!fs->frames)
      return ;

   switch  (fs->mode) {
      case FRAME_TILE :
         switch (fs->tile_mode) {
            case TILE_GRID : 
               cpk_frame_reparent_table (fs) ;
               break ;
            case TILE_COLUMNS : 
               cpk_frame_reparent_box (fs->frame_hbox, fs) ;
               break ;
            case TILE_ROW :
               cpk_frame_reparent_box (fs->frame_vbox, fs) ;
               break ;
            default :
               ;
         }
         break ;
      case FRAME_BLINK : case FRAME_COINC :
         break ;
      default :
         if (fs->last_tile_mode == TILE_COLUMNS)
            cpk_frame_reparent_box (fs->frame_hbox, fs) ;
         else
            cpk_frame_reparent_box (fs->frame_vbox, fs) ;
      break ;
   }
   cpk_frame_set_cursor (fs) ;
   /* This is needed to resize the frames container immediately */
   while (gtk_events_pending ())
      gtk_main_iteration () ;
}

void
cpk_frame_set_select (GtkWidget *w, CpkFrameSet *fs)
{
   CPKDEBUG5 ("cpk_frame_set_select", 5) ;
   if (fs->current && fs->current != CPK_FRAME (w)) {
      CPK_PLOT_UNSET_FLAGS (CPK_FRAME (fs->current)->plot, CPK_BORDER) ;
      gtk_widget_draw (GTK_WIDGET (fs->current), NULL) ;
   }
   if (CPK_PLOT_BORDER (CPK_FRAME (w)->plot))
      return ;
   /* if (fs->current == CPK_FRAME (w))
      return ; */
   fs->current = CPK_FRAME (w) ;
   CPK_PLOT_SET_FLAGS (CPK_FRAME (w)->plot, CPK_BORDER) ;
   gtk_widget_draw (GTK_WIDGET (fs->current), NULL) ;
   gtk_widget_draw (GTK_WIDGET (fs->paner), NULL) ;
   if (fs->mode != FRAME_TILE)
      cpk_frame_set_reparent (fs) ;
}

void
cpk_frame_set_cursor (CpkFrameSet *fs)
{
   CpkFrame *frame ;
   for (frame = fs->frames; frame; frame = frame->next) {
      if (frame->area->window)
         change_win_cursor (frame->area->window, fs->cursor, CURSOR_MODIFIER_NONE) ;
      CPK_PLOT (frame->plot)->p_action = fs->p_action ;
   }
}
