/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include "catalib.h"

#include <math.h>

#include <gtk/gtk.h>

#include "cpk_gtk_util.h"
#include "../pixmaps/quest_duck.xpm"
#include "../pixmaps/warn_duck.xpm"

void lut_red_blue       (float f, float *r, float *g, float *b) ;
void lut_red_blue_black (float f, float *r, float *g, float *b) ;

/*----------------------------------------------------------------------------
 * 							Private types
 *--------------------------------------------------------------------------*/

struct sel1_tag {
      GtkWidget *selector ;
      GtkWidget *entry ;
} ;

struct sel2_tag {
      GtkWidget *selector ;
      gchar     *buf ;
} ;

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static void get_name            (struct sel1_tag *s) ; 

static void file_ok_sel         (struct sel2_tag *s) ; 

static void cpk_entry_changed   (GtkWidget       *entry, 
                                 uvalue          *uval) ; 

static void cpk_entry_focus_out (GtkWidget       *entry, 
                                 GdkEvent        *event, 
                                 gpointer         data) ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

void
set_font(gchar *font_str)
{
   GdkFont       *font;
   GtkStyle      *style;
   font = gdk_font_load (font_str) ;
   if (font) {
      style = gtk_style_new () ;
      gdk_font_unref (style->font) ;
      style->font = font ;
      gdk_font_ref (style->font) ;
      gtk_widget_push_style (style) ;
      gtk_style_unref (style) ;
   }
}

GdkColor *
get_color (struct color_tag col)
{
   GdkColor *color ;
   color = (GdkColor *) malloc (sizeof(GdkColor));
   /* red, green, and blue are passed values, indicating the RGB triple
    * of the color we want to draw. Note that the values of the RGB components
    * within the GdkColor are taken from 0 to 65535, not 0 to 255.
    */
   color->red   = col.r * (65535/255) ;
   color->green = col.g * (65535/255) ;
   color->blue  = col.b * (65535/255) ;
   color->pixel = gdk_rgb_xpixel_from_rgb ((col.r << 16) | (col.g << 8) | col.b) ;
   /* color->pixel = (gulong)(col.r*65536 + col.g*256 + col.b) ; */
   /* However, the pixel valule is only truly valid on 24-bit (TrueColor)
    * displays. Therefore, this call is required so that GDK and X can
    * give us the closest color available in the colormap
    */
   return color ;
}

void
set_color (GtkWidget *widget, GdkGC  *gc, Color col)
{
   GdkColor *color ;

   color = get_color (col) ;
   gdk_color_alloc (gtk_widget_get_colormap (widget), color) ;

   /* set the foreground to our color */
   gdk_gc_set_foreground (gc, color) ;
   gdk_colormap_free_colors (gtk_widget_get_colormap (widget), color, 1) ;

   free (color) ;
}

void
set_color_sv (GtkWidget *widget, GdkGC  *gc, Color col, 
              gfloat sat, gfloat val)
{
   GdkColor *color ;
   gfloat   max ;

   col.r *= val ;
   col.g *= val ;
   col.b *= val ;
   max = MAX (col.r, MAX (col.g, col.b) ) ;
   col.r += (1.0-sat) * (max-col.r) ;
   col.g += (1.0-sat) * (max-col.g) ;
   col.b += (1.0-sat) * (max-col.b) ;

   color = get_color (col) ;
   /* gdk_color_alloc (gdk_rgb_get_cmap (), color) ; */
   gdk_color_alloc (gtk_widget_get_colormap (widget), color) ;

   /* set the foreground to our color */
   gdk_gc_set_foreground (gc, color) ;
   gdk_colormap_free_colors (gtk_widget_get_colormap (widget), color, 1) ;

   free (color) ;
}

void
cpk_set_fg_color (GtkWidget *widget, GdkGC  *gc, const gchar *spec)
{
   GdkColor     color ;
   GdkColormap *colormap;
   gint         i ;

   colormap = gtk_widget_get_colormap (widget) ;
   gdk_color_parse (spec, &color) ;
   color.pixel = gdk_rgb_xpixel_from_rgb ((color.red << 16) | (color.green << 8) | color.blue);
   gdk_colormap_alloc_color (colormap, &color, FALSE, TRUE) ;

   /* set the foreground to our color */
   gdk_gc_set_foreground (gc, &color) ;

   gdk_colormap_free_colors (colormap, &color, 1) ;

/*    printf ("color %s : %d %d %d   %ld ---  %ld; cmap size : %d\n", spec, 
      color.red,color.green,color.blue, color.pixel,
      gdk_rgb_xpixel_from_rgb ((color.red << 16) | (color.green << 8) | color.blue)
         ,colormap->size);
printf("Colormap size : %d\n",colormap->size) ;
   for(i=0; i<colormap->size; i++) printf("%d   %d %d %d   %ld\n", i, 
      colormap->colors->red,colormap->colors->green,colormap->colors->blue,
      colormap->colors->pixel); */
}

void
cpk_set_bg_color (GtkWidget *widget, GdkGC  *gc, const gchar *spec)
{
   GdkColor color ;

   gdk_color_parse (spec, &color) ;
   color.pixel = gdk_rgb_xpixel_from_rgb ((color.red << 16) | (color.green << 8) | color.blue);
   gdk_colormap_alloc_color (gtk_widget_get_colormap (widget), &color, 
         FALSE, TRUE) ;

   /* set the background to our color */
   gdk_gc_set_background (gc, &color) ;
   gdk_colormap_free_colors (gtk_widget_get_colormap (widget), &color, 1) ;
}

void
cpk_set_ns_color (GtkWidget *widget, GdkGC  *gc, gfloat f)
{
   GdkColor     color ;
   GdkColormap *colormap;
   gint         i ;
   gfloat       r, g, b ;
   gint         red, green, blue ;

   /* /usr/lib/X11/rgb.txt */

   lut_red_blue_black (f, &r, &g, &b) ;
   red   = 255 * r ;
   green = 255 * g ;
   blue  = 255 * b ;

   /* red, green, and blue are passed values, indicating the RGB triple
    * of the color we want to draw. Note that the values of the RGB components
    * within the GdkColor are taken from 0 to 65535, not 0 to 255.
    */
   color.red   = (gint) (r * 65535) ;
   color.green = (gint) (g * 65535) ;
   color.blue  = (gint) (b * 65535) ;
   /* color.pixel = (gulong)(r*65536 + g*256 + b) ; */
   color.pixel = gdk_rgb_xpixel_from_rgb ((red << 16) | (green << 8) | blue) ;

   /* However, the pixel valule is only truly valid on 24-bit (TrueColor)
    * displays. Therefore, this call is required so that GDK and X can
    * give us the closest color available in the colormap
    */

   /* colormap = gdk_rgb_get_cmap () ; */
   colormap = gtk_widget_get_colormap (widget) ;
   gdk_colormap_alloc_color (colormap, &color, FALSE, TRUE) ;

   /* set the foreground to our color */
   gdk_gc_set_foreground (gc, &color) ;
   gdk_colormap_free_colors (colormap, &color, 1) ;
}

void
lut_red_blue (float f, float *r, float *g, float *b)
{
   *r = ( (f <= 0.5) ? 1.0 : 2.0*(1.0-f) ) ;
   *g = 0 ;
   *b = ( (f <= 0.5) ? 2.0*f : 1.0 ) ;
}

void
lut_red_blue_black (float f, float *r, float *g, float *b)
{
   f *= 2.0 ;
   if (f <= 0.5) {
      *r = 1.0 ;
      *g = 0.0 ;
      *b = 2.0*f ;
      return ;
   }
   if (f <= 1.0) {
      *r = 2.0*(1.0-f) ;
      *g = 0.0 ;
      *b = 1.0 ;
      return ;
   }
   *r = 0 ;
   *g = 0 ;
   *b = 2.0-f ;
}

/* Draw a circle on the pixmap */
void
draw_circle (GdkPixmap *pixmap, GdkGC  *gc, 
             gdouble x, gdouble y, gdouble r, gint nv)
{
  GdkPoint     points[256] ;
  gdouble      a ;
  gint         i, n = (nv<256) ? nv : 256 ;

  for (i = 0; i < n; i++) {
     a = i*360.0/(gdouble)n ;
     points[i].x = (gint16)(x + cos (DEGTORAD (a))*r) ;
     points[i].y = (gint16)(y + sin (DEGTORAD (a))*r) ;
  }
  
  gdk_draw_polygon (pixmap, gc, FALSE, points, n) ;
}

void
cpk_plot_symb (GdkPixmap *pixmap, GdkGC  *gc, cpk_symb type,
               gdouble x, gdouble y, gdouble r_in, gdouble r_out, gint nv)
{
  GdkPoint     points[256] ;
  gint         n = (nv < 256) ? nv : 256 ;

   switch (type) {
      case CPK_CIRCLE :
         cpk_circle (points, x, y, r_out, r_in, n) ;
         break ;
      case CPK_PLUS :
         cpk_poly_plus (points, (gint)x, (gint)y, (gint)r_in, (gint)r_out) ;
         n = 9 ;
         break ;
      case CPK_CROSS :
         cpk_poly_cross (points, (gint)x, (gint)y, (gint)r_in, (gint)r_out) ;
         n = 9 ;
         break ;
      default :
         break ;
   }
   gdk_draw_polygon (pixmap, gc, FALSE, points, n) ;
}

void
cpk_plot_seg (GdkPixmap *pixmap, GdkGC  *gc, 
              gdouble x0, gdouble y0, gdouble x1, gdouble y1)
{
   gdk_draw_line (pixmap, gc, x0, y0, x1, y1) ;
}

void
cpk_circle (GdkPoint points[], gdouble x, gdouble y, gdouble r, gdouble a0, gint nv)
{
  gdouble      a ;
  gint         i, n = (nv < 256) ? nv : 256 ;

  for (i = 0; i < n; i++) {
     a = i * 360.0 / (gdouble)n + a0 ;
     points[i].x = (gint)(x + cos (DEGTORAD (a))*r) ;
     points[i].y = (gint)(y + sin (DEGTORAD (a))*r) ;
  }
}

void
cpk_poly_plus (GdkPoint points[], gint x, gint y, gint ws, gint wl)
{   
   points[0].x = x - wl ;
   points[0].y = y ;
   points[1].x = x - ws ;
   points[1].y = y + ws ;

   points[2].x = x ;
   points[2].y = y + wl ;
   points[3].x = x + ws ;
   points[3].y = y + ws ;

   points[4].x = x + wl ;
   points[4].y = y ;
   points[5].x = x + ws ;
   points[5].y = y - ws ;

   points[6].x = x ;
   points[6].y = y - wl ;
   points[7].x = x - ws ;
   points[7].y = y - ws ;
   points[8].x = points[0].x ;
   points[8].y = points[0].y ;
}

void
cpk_poly_cross (GdkPoint points[], gint x, gint y, gint ws, gint wl)
{   
   points[0].x = x - wl ;
   points[0].y = y - wl;
   points[1].x = x - ws ;
   points[1].y = y ;

   points[2].x = x - wl ;
   points[2].y = y + wl ;
   points[3].x = x ;
   points[3].y = y + ws ;

   points[4].x = x + wl ;
   points[4].y = y + wl ;
   points[5].x = x + ws ;
   points[5].y = y ;

   points[6].x = x + wl ;
   points[6].y = y - wl ;
   points[7].x = x ;
   points[7].y = y - ws ;
   points[8].x = points[0].x ;
   points[8].y = points[0].y ;
}

GtkWidget *
cpk_warning (GtkWidget *window, gchar *message)
{
   GtkWidget *dialog, *box, *label, *button, *duck ;

   /* Create a new window */
   dialog = gtk_dialog_new () ;
   gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window)) ;
   gtk_window_set_title (GTK_WINDOW (dialog), "*** Warning ***") ;
   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER) ;

   box = gtk_hbox_new (FALSE, 0) ;
   gtk_container_set_border_width (GTK_CONTAINER (box), 30) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), box,
                       TRUE, TRUE, 5) ;

   duck = cpk_xpm_box_new (window, warn_duck, message) ;
   gtk_box_pack_start (GTK_BOX (box), duck, FALSE, FALSE, 0) ;
   /* label = gtk_label_new (message) ;
   gtk_box_pack_start (GTK_BOX (box), label, FALSE, FALSE, 0) ; */

   button = gtk_button_new_with_label ("Dismissis") ;
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer)dialog) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, FALSE, FALSE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_grab_default (button) ;

   gtk_widget_show_all (dialog) ;

   return dialog ;
}

void
entry_changed (GtkWidget *entry, gchar *value)
{
   gchar    buffer[255] ;
   gchar   *text ;
   gint     n ;

   if (entry == NULL) 
      return ;
   if (!GTK_IS_ENTRY (entry)) 
      return ;
   text = gtk_entry_get_text (GTK_ENTRY (entry)) ;
   if (text == NULL) 
      return ;
   if (index (text,'e') || index (text,'E'))
      sprintf (buffer, "%.*e", ndecimals (text), atof (text)) ;
   else
      sprintf (buffer, "%.*f", ndecimals (text), atof (text)) ;
   gtk_entry_set_text (GTK_ENTRY (entry), buffer) ;
   text = gtk_entry_get_text (GTK_ENTRY (entry)) ;
   if (value) 
      strcpy (value, text) ;
}

void
entry_focus_out (GtkWidget *entry, GdkEvent *event, gpointer data)
{
   entry_changed (entry, data) ;
}

static void
confirm_close (GtkWidget *window, gpointer data)
{
   gtk_main_quit () ;
   gtk_widget_destroy (window) ;
}

static void
confirm_yes (GtkWidget *widget, gint *status)
{
   *status = 0 ;
}

static void
confirm_no (GtkWidget *widget, gint *status)
{
   *status = 1 ;
}

extern gint 
cpk_confirm (GtkWidget *window, gchar *title, gchar *msg)
{
   GtkWidget *dialog, *box, *label, *button, *duck ;
   static gint status = 1 ;

   dialog = gtk_dialog_new () ;
   gtk_window_set_transient_for (GTK_WINDOW (dialog), 
         GTK_WINDOW (window)) ;
   gtk_window_set_title (GTK_WINDOW (dialog), title) ;
   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE) ;
   gtk_signal_connect (GTK_OBJECT (dialog), "destroy", 
         GTK_SIGNAL_FUNC (confirm_close), NULL) ;
   box = gtk_vbox_new (FALSE, 0) ;
   gtk_container_set_border_width (GTK_CONTAINER(box), 8) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dialog)->vbox), box,
                       TRUE, TRUE, 0) ;
   duck = cpk_xpm_box_new (window, quest_duck, msg) ;
   gtk_box_pack_start (GTK_BOX (box), duck, FALSE, FALSE, 0) ;
   /* label = gtk_label_new (msg) ;
   gtk_container_add (GTK_CONTAINER (box), label) ; */
   
   button = gtk_button_new_with_label ("Continue") ;
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (confirm_yes), &status) ;
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) dialog) ;
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (gtk_main_quit), NULL) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;

   button = gtk_button_new_with_label ("Dismissis") ;
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (confirm_no), &status) ;
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) dialog) ;
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (gtk_main_quit), NULL) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_grab_default (button) ;

   gtk_widget_show_all (dialog) ;
   /* Control */
   gtk_main () ;
   return status ;
}

GtkWidget *
make_menu_item (gchar *name, GtkSignalFunc callback, gpointer data)
{
    GtkWidget *item ;
  
    item = gtk_menu_item_new_with_label (name) ;
    gtk_signal_connect (GTK_OBJECT (item), "activate",
                        callback, data) ;
    gtk_widget_show (item) ;
    return (item) ;
}

GtkWidget *
cpk_xpm_box_new (GtkWidget *parent, gchar **xpm_data, const gchar *text)
{
   GtkWidget *box, *bpixmapwid ;
   GdkPixmap *bpixmap ;
   GdkBitmap *mask ;
   GtkStyle  *style ;
  
   box = gtk_hbox_new (FALSE, 0) ;
   gtk_container_set_border_width (GTK_CONTAINER(box), 0) ;
   style = gtk_widget_get_style (parent) ;
  
   bpixmap = gdk_pixmap_colormap_create_from_xpm_d (parent->window, 
         gtk_widget_get_colormap (parent), 
         &mask, 
         &style->bg[GTK_STATE_NORMAL], 
         (gchar **)xpm_data) ;
  
   bpixmapwid = gtk_pixmap_new (bpixmap, mask) ;

   gdk_pixmap_unref (bpixmap) ;
   gdk_bitmap_unref (mask) ;

   if (text) {
      GtkWidget *abox, *hbox, *label ;

      abox = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
      gtk_container_add (GTK_CONTAINER (box), abox);
      gtk_widget_show (abox);

      hbox = gtk_hbox_new (FALSE, 0);
      gtk_container_add (GTK_CONTAINER (abox), hbox);
      gtk_widget_show (hbox);

      gtk_box_pack_start (GTK_BOX (hbox), bpixmapwid, FALSE, FALSE, 4);
      gtk_widget_show (bpixmapwid) ;

      label = gtk_label_new (text);
      gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 4);
      gtk_widget_show (label);
   }
   else {
      gtk_container_add (GTK_CONTAINER(box), bpixmapwid) ;
      gtk_widget_show (bpixmapwid) ;
   }

   return box ;
}

static void
set_units_sec (GtkWidget *item, uvalue *uval)
{
   gchar      buffer[256] ;
   gdouble    val ;
   if (uval->tunits == 3600.0)
      return ;
   if (uval->entry) {
      val = atof (gtk_entry_get_text (GTK_ENTRY (uval->entry))) ;
      sprintf (buffer, "%g", val * 3600.0) ;
      gtk_entry_set_text (GTK_ENTRY (uval->entry), buffer) ;
   }
   uval->tunits = 3600.0 ;
}

static void
set_units_deg (GtkWidget *item, uvalue *uval)
{
   gchar      buffer[256] ;
   gdouble    val ;
   if (uval->tunits == 1.0)
      return ;
   if (uval->entry) {
      val = atof (gtk_entry_get_text (GTK_ENTRY (uval->entry))) ;
      sprintf (buffer, "%g", val / 3600.0) ;
      gtk_entry_set_text (GTK_ENTRY (uval->entry), buffer) ;
   }
   uval->tunits = 1.0 ;
}

GtkWidget *
cpk_get_units (uvalue *uval, gboolean spheric)
{
   GtkWidget   *option, *menu, *item ;
   option = gtk_option_menu_new() ;
   menu = gtk_menu_new () ;
   if (spheric) {
      item = make_menu_item ("[arcsec]", GTK_SIGNAL_FUNC (set_units_sec), uval) ;
      gtk_menu_append (GTK_MENU (menu), item) ;
      item = make_menu_item ("[deg]",    GTK_SIGNAL_FUNC (set_units_deg), uval) ;
      gtk_menu_append (GTK_MENU (menu), item) ;
      if (uval->tunits == 3600.0)
         gtk_menu_set_active (GTK_MENU (menu), 0) ;
      else
         gtk_menu_set_active (GTK_MENU (menu), 1) ;
   }
   else {
      item = make_menu_item ("[pixel]", NULL, NULL) ;
      gtk_menu_append (GTK_MENU (menu), item) ;
   }
   gtk_option_menu_set_menu (GTK_OPTION_MENU (option), menu) ;

   return option ;
}

void
cpk_entry_with_units (uvalue *uval, gdouble value, gboolean spheric)
{
   GtkWidget  *entry, *units ;
   gchar       buffer[256] ;
   gdouble     s ;

   s = (spheric) ? tunits : 1 ;

   uval->entry = entry = gtk_entry_new () ;
   uval->value = value ;
   uval->tunits = s ;
   
   sprintf (buffer, "%g", value * s) ;
   gtk_entry_set_text (GTK_ENTRY (entry), buffer) ;

   gtk_signal_connect (GTK_OBJECT (entry), "activate",
         GTK_SIGNAL_FUNC (entry_changed), NULL) ;
   gtk_signal_connect (GTK_OBJECT (entry), "focus-out-event",
         GTK_SIGNAL_FUNC (entry_focus_out), NULL) ;

   gtk_signal_connect (GTK_OBJECT (entry), "activate",
         GTK_SIGNAL_FUNC (cpk_entry_changed), uval) ;
   gtk_signal_connect (GTK_OBJECT (entry), "focus-out-event",
         GTK_SIGNAL_FUNC (cpk_entry_focus_out), uval) ;

   uval->units = cpk_get_units (uval, spheric) ;
}

static void
cpk_entry_changed (GtkWidget *entry, uvalue *uval)
{
   uval->value = atof (gtk_entry_get_text (GTK_ENTRY (uval->entry))) / 
                 uval->tunits ;
}

static void
cpk_entry_focus_out (GtkWidget *entry, GdkEvent *event, gpointer data)
{
   cpk_entry_changed (entry, data) ;
}

static void 
get_name (struct sel1_tag *s) 
{
   gtk_entry_set_text (GTK_ENTRY (s->entry),
         gtk_file_selection_get_filename (GTK_FILE_SELECTION (s->selector))) ;
   gtk_editable_set_position (GTK_EDITABLE (s->entry), -1) ;
}

void 
cpk_get_file_entry (GtkWidget *entry) 
{
   static struct sel1_tag s ;

   if (GTK_IS_FILE_SELECTION (s.selector)) 
      return ;
   g_return_if_fail (GTK_IS_ENTRY (entry));

   s.selector = gtk_file_selection_new ("Please select a file name.");
   s.entry    = entry ;

   gtk_file_selection_set_filename (GTK_FILE_SELECTION (s.selector),
         gtk_entry_get_text (GTK_ENTRY (s.entry))) ;
   gtk_signal_connect_object (GTK_OBJECT (
         GTK_FILE_SELECTION (s.selector)->ok_button), 
         "clicked", GTK_SIGNAL_FUNC (get_name), (gpointer)&s) ;
   gtk_signal_connect_object (GTK_OBJECT (
         GTK_FILE_SELECTION (s.selector)->ok_button), 
         "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), 
         (gpointer) s.selector) ;
   gtk_signal_connect_object (GTK_OBJECT (
         GTK_FILE_SELECTION (s.selector)->cancel_button), 
         "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), 
         (gpointer) s.selector) ;
   gtk_widget_show (s.selector) ;
   gtk_window_set_transient_for (GTK_WINDOW (s.selector), 
         GTK_WINDOW (gtk_widget_get_toplevel (entry))) ;
}

static void 
destroy (GtkWidget *widget, gpointer   data)
{
    gtk_main_quit () ;
}

static void 
file_ok_sel (struct sel2_tag *s)
{
   strcpy (s->buf, gtk_file_selection_get_filename (
         GTK_FILE_SELECTION (s->selector))) ;
}

static void 
file_canc_sel (struct sel2_tag *s)
{
   s->buf = NULL ;
}

gchar * 
cpk_get_file_name (GtkWidget *widget, char *user_buf) 
{
   struct sel2_tag s ;
   static char static_buf[MAXPATHLEN] ;

   if (user_buf) 
      s.buf = user_buf ;
   else
      s.buf = static_buf ;

   if (!strcmp (s.buf, ""))
      strcpy (s.buf, "cxc.dat") ;
   s.selector = gtk_file_selection_new ("Please select a file name.") ;
   gtk_window_set_position (GTK_WINDOW (s.selector), GTK_WIN_POS_MOUSE) ;
   if (widget)
      if (GTK_IS_WIDGET (widget))
         gtk_window_set_transient_for (GTK_WINDOW (s.selector), 
            GTK_WINDOW (gtk_widget_get_toplevel (widget))) ;
   gtk_signal_connect_object (GTK_OBJECT (s.selector), "destroy",
                      (GtkSignalFunc) destroy, NULL) ;

   gtk_file_selection_set_filename (GTK_FILE_SELECTION (s.selector), s.buf) ;
   gtk_signal_connect_object (GTK_OBJECT (
         GTK_FILE_SELECTION (s.selector)->ok_button), 
         "clicked", GTK_SIGNAL_FUNC (file_ok_sel), (gpointer)&s) ;
   gtk_signal_connect_object (GTK_OBJECT (
         GTK_FILE_SELECTION (s.selector)->ok_button), 
         "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) s.selector) ;
   gtk_signal_connect_object (GTK_OBJECT (
         GTK_FILE_SELECTION (s.selector)->cancel_button), 
         "clicked", GTK_SIGNAL_FUNC (file_canc_sel), (gpointer)&s) ;
   gtk_signal_connect_object (GTK_OBJECT (
         GTK_FILE_SELECTION (s.selector)->cancel_button), 
         "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) s.selector) ;
   gtk_widget_show (s.selector) ;

   gtk_main () ;

   return s.buf ;
}

GtkWidget *
cpk_icon_button (GtkWidget *parent, gchar **xpm_data, 
                 const gchar *text,  const gchar *tip)
{
   GtkWidget *button, *box ;
   GtkTooltips *ttip ;

   button = gtk_button_new () ;
   box = cpk_xpm_box_new (parent, xpm_data, text) ;
   gtk_container_add (GTK_CONTAINER (button), box) ;
   if (tip) {
      ttip = gtk_tooltips_new () ;
      gtk_tooltips_set_tip (GTK_TOOLTIPS (ttip), button, tip, NULL) ;
   }
   return button ;
}

GtkWidget *
cpk_console_new (GtkWidget *window)
{
   GtkWidget     *box, *table, *text, *vscrollbar ;
   GdkColormap   *cmap;
   GdkColor       color;
   GdkFont       *fixed_font;

   box = gtk_vbox_new (FALSE, 2);
   gtk_container_set_border_width (GTK_CONTAINER (box), 2);
  
   table = gtk_table_new (1, 2, FALSE);
   gtk_table_set_row_spacing (GTK_TABLE (table), 0, 2);
   gtk_table_set_col_spacing (GTK_TABLE (table), 0, 2);
   gtk_box_pack_start (GTK_BOX (box), table, TRUE, TRUE, 0);

   /* Create the GtkText widget */
   /* display->console =  */ text = gtk_text_new (NULL, NULL);
   gtk_text_set_editable (GTK_TEXT (text), FALSE);
   gtk_table_attach (GTK_TABLE (table), text, 0, 1, 0, 1,
		     GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		     GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);

   /* Add a vertical scrollbar to the GtkText widget */
   vscrollbar = gtk_vscrollbar_new (GTK_TEXT (text)->vadj);
   gtk_table_attach (GTK_TABLE (table), vscrollbar, 1, 2, 0, 1,
		     GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);

   /* Get the system color map and allocate the color red */
   cmap = gtk_widget_get_colormap (window) ;

   color.red = 0xffff;
   color.green = 0;
   color.blue = 0;
   if (!gdk_color_alloc (cmap, &color)) {
     g_error ("couldn't allocate color");
   }

   /* Load a fixed font */
   fixed_font = gdk_font_load ("-misc-fixed-medium-r-*-*-*-140-*-*-*-*-*-*");

   /* Realizing a widget creates a window for it,
    * ready for us to insert some text */
   /* gtk_widget_realize (text) ; */

   /* Freeze the text widget, ready for multiple updates */
   gtk_text_freeze (GTK_TEXT (text)) ;
   /* Thaw the text widget, allowing the updates to become visible */  
   gtk_text_thaw (GTK_TEXT (text)) ;
 
   gtk_widget_show_all (box);
   gdk_colormap_free_colors (cmap, &color, 1) ;

   return box ;
}
