/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: cpk_init.c
 * Author 	: Paolo Montegriffo
 * Created on	: May, 2002	
 * Language	: ANSI C
 * Description	: Initialize GTK interface for CataXcorr
 * Version      : 2.1.14 - May 05, 2003
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include <signal.h>
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif

#ifdef G_OS_WIN32
#include <process.h>		/* For _getpid() */
#endif

#include "common.h"

#include <gtk/gtk.h>

#include "catalib.h"
#include "colormaps.h"
#include "cpk_env.h"
#include "cpk_init.h"
#include "../pixmaps/on.xpm"

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#define LOGO_WIDTH_MIN  300
#define LOGO_HEIGHT_MIN 110
#define AUTHORS "Paolo Montegriffo"

#define SHOW_NEVER 0
#define SHOW_LATER 1
#define SHOW_NOW   2
#ifndef MAXPATHLEN
#define MAXPATHLEN 2048
#endif

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static void      make_initialization_status_window    (void);
static void      destroy_initialization_status_window (void);
static gboolean  cpk_logo_load                        (GtkWidget *window);
static gboolean  cpk_logo_load_size                   (GtkWidget *window);
static void      cpk_logo_draw                        (GtkWidget *widget);
static void      cpk_logo_text_draw                   (GtkWidget *widget);
static void      cpk_logo_expose                      (GtkWidget *widget);

static void      gimp_dialog_set_icon                 (GtkWindow *dialog);
static void      gimp_dialog_realize_callback         (GtkWidget *widget,
			                               gpointer   data) ;
static gint      cpk_let_close                        (gpointer data) ;

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern int        install_cmap ;
extern int        verbose ;
extern char      *module ;
extern boolean    no_interface ;
extern gint       wait_me ;

static gboolean   is_app_exit_finish_done = FALSE;

static GtkWidget *logo_area   = NULL;
static GdkPixmap *logo_pixmap = NULL;
static gchar     *logo = NULL;
static gint       logo_width        = 0;
static gint       logo_height       = 0;
static gint       logo_area_width   = 0;
static gint       logo_area_height  = 0;
static gint       show_logo         = SHOW_LATER;
static gint       max_label_length  = MAXPATHLEN;

static GtkWidget *win_initstatus = NULL ;
static GtkWidget *label1 = NULL ;
static GtkWidget *label2 = NULL ;
static GtkWidget *pbar   = NULL ;

gint       min_colors = 200 ;
gint       cycled_marching_ants = 0 ;

gboolean   we_are_exiting = FALSE ;
boolean    no_logo        = FALSE ;
boolean    no_logo_image  = FALSE ;

/*----------------------------------------------------------------------------
 *							Functions code
 *--------------------------------------------------------------------------*/


static gboolean
cpk_logo_load_size (GtkWidget *window)
{
   gchar buf[1024] ;
   FILE *fp ;

   CPKDEBUG5 ("cpk_logo_load_size", 4) ;
   if (logo_pixmap)
     return TRUE ;

   sprintf (buf, "%s" G_DIR_SEPARATOR_S "%s", cpk_data_directory (), logo) ;

   fp = fopen (buf, "rb") ;
   if (!fp)
     return FALSE ;

   fgets (buf, (int) sizeof (buf), fp) ;
   if (strcmp (buf, "P6\n") != 0) {
      fclose (fp) ;
      return FALSE ;
   }
   fgets (buf, (int) sizeof (buf), fp) ;
   fgets (buf, (int) sizeof (buf), fp) ;
   sscanf (buf, "%d %d", &logo_width, &logo_height) ;

   fclose (fp) ;

   return TRUE ;
}

static gboolean
cpk_logo_load (GtkWidget *window)
{
   GtkWidget *preview ;
   GdkGC     *gc ;
   gchar      buf[1024] ;
   guchar    *pixelrow ;
   FILE      *fp ;
   gint       count ;
   gint       i ;

   CPKDEBUG ("cpk_logo_load") ;
   if (logo_pixmap)
     return TRUE ;

   sprintf (buf, "%s" G_DIR_SEPARATOR_S "%s", cpk_data_directory (), logo) ;

   fp = fopen (buf, "rb") ;
   if (!fp)
     return FALSE ;

   fgets (buf, (int) sizeof (buf), fp) ;
   if (strcmp (buf, "P6\n") != 0) {
      fclose (fp) ;
      return FALSE ;
   }

   fgets (buf, (int) sizeof (buf), fp) ;
   fgets (buf, (int) sizeof (buf), fp) ;
   sscanf (buf, "%d %d", &logo_width, &logo_height) ;

   fgets (buf, (int) sizeof (buf), fp) ;
   if (strcmp (buf, "255\n") != 0) {
      fclose (fp) ;
      return FALSE ;
   }

   preview = gtk_preview_new (GTK_PREVIEW_COLOR) ;
   gtk_preview_size (GTK_PREVIEW (preview), logo_width, logo_height) ;
   pixelrow = g_new (guchar, (size_t)logo_width * 3) ;

   for (i = 0; i < logo_height; i++) {
      count = fread (pixelrow, sizeof (unsigned char), 
            (size_t)logo_width * 3, fp) ;
      if (count != (logo_width * 3)) {
         gtk_widget_destroy (preview) ;
	 g_free (pixelrow) ;
	 fclose (fp) ;
	 return FALSE ;
      }
      gtk_preview_draw_row (GTK_PREVIEW (preview), pixelrow, 0, i, logo_width) ;
   }

   gtk_widget_realize (window) ;
   logo_pixmap = gdk_pixmap_new (window->window, logo_width, logo_height,
				 gtk_preview_get_visual ()->depth) ;
   gc = gdk_gc_new (logo_pixmap) ;
   gtk_preview_put (GTK_PREVIEW (preview),
		    logo_pixmap, gc,
		    0, 0, 0, 0, logo_width, logo_height) ;
   gdk_gc_destroy (gc) ;

   gtk_widget_unref (preview) ;
   g_free (pixelrow) ;

   fclose (fp) ;

   return TRUE ;
}

static void
cpk_logo_text_draw (GtkWidget *widget)
{
   GdkFont *font ;
   CPKDEBUG ("cpk_logo_text_draw") ;

   /* this is a font, provide only one single font definition */
   font = gdk_font_load (("-*-helvetica-bold-r-normal--*-140-*-*-*-*-*-*")) ;
   if (!font) {
      GtkStyle *style = gtk_widget_get_style (widget) ;
      font = style->font ;
      gdk_font_ref (font) ;
   }

   gdk_draw_string (widget->window,
		    font,
		    widget->style->fg_gc[GTK_STATE_NORMAL],
		    ((logo_area_width - gdk_string_width (font,  (module))) / 2),
		    (gint)(0.25 * logo_area_height),
		    (module)) ;

   gdk_font_unref (font);

   /* this is a fontset, e.g. multiple comma-separated font definitions */
   font = gdk_fontset_load (("-*-helvetica-bold-r-normal--*-120-*-*-*-*-*-*,*")) ;
   if (!font) {
      GtkStyle *style = gtk_widget_get_style (widget) ;
      font = style->font ;
      gdk_font_ref (font) ;
   }

   gdk_draw_string (widget->window,
		    font,
		    widget->style->fg_gc[GTK_STATE_NORMAL],
		    ((logo_area_width - gdk_string_width (font, VERSION)) / 2),
		    (gint)(0.45 * logo_area_height),
		    VERSION) ;
   gdk_draw_string (widget->window,
		    font,
		    widget->style->fg_gc[GTK_STATE_NORMAL],
		    ((logo_area_width - gdk_string_width (font, ("brought to you by"))) / 2),
		    (gint)(0.65 * logo_area_height),
		    ("brought to you by")) ;
   gdk_draw_string (widget->window,
		    font,
		    widget->style->fg_gc[GTK_STATE_NORMAL],
		    ((logo_area_width - gdk_string_width (font, AUTHORS)) / 2),
		    (gint)(0.80 * logo_area_height),
		    AUTHORS) ;

   gdk_font_unref (font) ;
}

static void
cpk_logo_draw (GtkWidget *widget)
{
   gdk_draw_pixmap (widget->window,
		    widget->style->black_gc,
		    logo_pixmap,
		    0, 0,
		    ((logo_area_width - logo_width) / 2),
		    ((logo_area_height - logo_height) / 2),
		    logo_width, logo_height) ;
}

static void
cpk_logo_expose (GtkWidget *widget)
{
   switch (show_logo) {
      case SHOW_NEVER:
      case SHOW_LATER:
         cpk_logo_text_draw (widget);
         break;
      case SHOW_NOW:
         cpk_logo_draw (widget);
         break;
   }
}

static void
destroy_initialization_status_window (void)
{
   if (win_initstatus) {
      gtk_widget_destroy (win_initstatus) ;
      if (logo_pixmap != NULL)
         gdk_pixmap_unref (logo_pixmap) ;

      win_initstatus = label1 = label2 = pbar = logo_area = NULL ;
      logo_pixmap = NULL ;
   }
}

static void
make_initialization_status_window (void)
{
   GtkWidget *vbox ;
   GtkWidget *logo_hbox ;
   GtkStyle  *style ;
   gchar      buf[1024] ;

   if (no_interface || no_logo)
      return ;

   gtk_widget_push_visual (gdk_rgb_get_visual ()) ;
   gtk_widget_push_colormap (gdk_rgb_get_cmap ()) ;
   win_initstatus = gtk_window_new (GTK_WINDOW_DIALOG) ;
   gtk_widget_pop_visual () ;
   gtk_widget_pop_colormap () ;   

   sprintf (buf, "%s Startup", module) ;
   gtk_window_set_title (GTK_WINDOW (win_initstatus), buf) ;
   gtk_window_set_wmclass (GTK_WINDOW (win_initstatus), "cpk_startup", module) ;
   gtk_window_set_position (GTK_WINDOW (win_initstatus), GTK_WIN_POS_CENTER) ;
   gtk_window_set_policy (GTK_WINDOW (win_initstatus), FALSE, FALSE, FALSE) ;

   /************* FIXME !!!!!!!!!!! ************/
   gimp_dialog_set_icon (GTK_WINDOW (win_initstatus)) ;

   if (no_logo_image == FALSE &&
      cpk_logo_load_size (win_initstatus)) {
      show_logo = SHOW_LATER ;
   }

   vbox = gtk_vbox_new (FALSE, 4) ;
   gtk_container_add (GTK_CONTAINER (win_initstatus), vbox) ;

   logo_hbox = gtk_hbox_new (FALSE, 0) ;
   gtk_box_pack_start (GTK_BOX (vbox), logo_hbox, FALSE, TRUE, 0) ;

   logo_area = gtk_drawing_area_new () ;

   gtk_signal_connect (GTK_OBJECT (logo_area), "expose_event",
		       GTK_SIGNAL_FUNC (cpk_logo_expose),
		       NULL) ;

   logo_area_width  = MAX (logo_width, LOGO_WIDTH_MIN) ;
   logo_area_height = MAX (logo_height, LOGO_HEIGHT_MIN) ;

   gtk_drawing_area_size (GTK_DRAWING_AREA (logo_area),
			  logo_area_width, logo_area_height) ;
   gtk_box_pack_start (GTK_BOX(logo_hbox), logo_area, TRUE, FALSE, 0) ;

   label1 = gtk_label_new ("") ;
   gtk_box_pack_start_defaults (GTK_BOX (vbox), label1) ;
   label2 = gtk_label_new ("") ;
   gtk_box_pack_start_defaults (GTK_BOX (vbox), label2) ;

   pbar = gtk_progress_bar_new () ;
   gtk_box_pack_start_defaults (GTK_BOX (vbox), pbar) ;

   gtk_widget_show (vbox) ;
   gtk_widget_show (logo_hbox) ;
   gtk_widget_show (logo_area) ;
   gtk_widget_show (label1) ;
   gtk_widget_show (label2) ;
   gtk_widget_show (pbar) ;

   gtk_widget_show (win_initstatus) ;

   /*  This is a hack: we try to compute a good guess for the maximum 
    *  number of charcters that will fit into the splash-screen using 
    *  the default_font
    */
   style = gtk_widget_get_style (win_initstatus) ;
   max_label_length = (0.8 * (gfloat) strlen (AUTHORS) *
		       ((gfloat) logo_area_width /
		        (gfloat) gdk_string_width (style->font, AUTHORS))) ;
}

static void
gimp_dialog_set_icon (GtkWindow *dialog)
{
  g_return_if_fail (dialog);
  g_return_if_fail (GTK_IS_WINDOW (dialog));

  if (GTK_WIDGET_REALIZED (GTK_WIDGET (dialog)))
    gimp_dialog_realize_callback (GTK_WIDGET (dialog), NULL);
  else
    gtk_signal_connect (GTK_OBJECT (dialog), "realize",
			GTK_SIGNAL_FUNC (gimp_dialog_realize_callback),
			NULL);
}

static void
gimp_dialog_realize_callback (GtkWidget *widget,
			      gpointer   data) 
{
  static GdkPixmap *wilber_pixmap = NULL;
  static GdkBitmap *wilber_mask   = NULL;
  GtkStyle         *style;

  style = gtk_widget_get_style (widget);
  /* printf("gimp_dialog_realize_callback \n"); *//***FIXME***/
  return ;

  if (wilber_pixmap == NULL)
    wilber_pixmap =
      gdk_pixmap_create_from_xpm_d (widget->window,
				    &wilber_mask,
				    &style->bg[GTK_STATE_NORMAL],
				    on_xpm);

  gdk_window_set_icon (widget->window, NULL,
		       wilber_pixmap, wilber_mask);
}

void
app_init_update_status (gchar *label1val,
		        gchar *label2val,
		        gdouble  pct_progress)
{
  gchar *temp;

  if (!no_interface && !no_logo && win_initstatus)
    {
      if (label1val && strcmp (label1val, GTK_LABEL (label1)->label))
	gtk_label_set_text (GTK_LABEL (label1), label1val);

      if (label2val && strcmp (label2val, GTK_LABEL (label2)->label))
	{
	  while (strlen (label2val) > max_label_length)
	    {
	      temp = strchr (label2val, G_DIR_SEPARATOR);
	      if (temp == NULL)  /* for sanity */
		break;
	      temp++;
	      label2val = temp;
	    }

	  gtk_label_set_text (GTK_LABEL (label2), label2val);
	}

      if (pct_progress >= 0.0 && pct_progress <= 1.0 && 
	  gtk_progress_get_current_percentage (&(GTK_PROGRESS_BAR (pbar)->progress)) != pct_progress)
	/*
	 GTK_PROGRESS_BAR(pbar)->percentage != pct_progress)
	*/
	{
	  gtk_progress_bar_update (GTK_PROGRESS_BAR (pbar), pct_progress) ;
	}

      while (gtk_events_pending ())
	gtk_main_iteration ();

      /* We sync here to make sure things get drawn before continuing,
       * is the improved look worth the time? I'm not sure...
       */
      gdk_flush ();
    }
}

void
cpk_init (gchar *file_logo)
{
   gchar *filename ;
   gchar *path ;

   wait_me = 1 ;
   logo = file_logo ;
   /*  parse the systemwide gtkrc  */
   filename = cpk_gtkrc () ;
   if (verbose > 4)
      g_print (("parsing \"%s\"\n"), filename) ;

   gtk_rc_parse (filename) ;
   g_free (filename) ;

   if (!no_interface)
      get_standard_colormaps () ;

   make_initialization_status_window () ;

   if (!no_interface && !no_logo && win_initstatus)
      cpk_logo_text_draw (logo_area) ;

#ifdef DISPLAY_FILTERS
   printf("CAZCAZCAZCAZ\n");
   color_display_init () ;
#endif /* DISPLAY_FILTERS */

  /* Now we are ready to draw the splash-screen-image to the start-up window */
   if (no_interface == FALSE) {
      if (no_logo_image == FALSE && show_logo &&
         cpk_logo_load (win_initstatus)) {
         show_logo = SHOW_NOW;
         cpk_logo_draw (logo_area) ;
      }
   }

#ifdef cazcaz
  
   /*  Things to do only if there is an interface  */
   if (no_interface == FALSE)
    {
      session_restore () ;
    }
#endif


   gtk_timeout_add (1000, (GtkFunction) cpk_let_close, NULL) ;
}

static gint
cpk_let_close (gpointer data)
{
   if (wait_me)
      return TRUE ;
   destroy_initialization_status_window () ;
   return FALSE ;
}
