/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <gtk/gtk.h>

#include "cpk_gtk_util.h"
#include "cpk_plot.h"
#include "cpk_draw.h"

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#define PLOT_DELAY_1 (250)
#define PLOT_DELAY_2 (50)

#define CPK_PLOT_DRAWABLE(widget)    (CPK_DRAW (cpk_plot_drawable (widget)))

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

/* Local data */

static GtkBinClass  *parent_class = NULL;
/* static guint         plot_signals[PLOT_LAST_SIGNAL] = { 0 }; */


/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/
                                               
static void        cpk_plot_class_init       (CpkPlotClass      *klass) ;

static void        cpk_plot_init             (CpkPlot           *plot) ;
 
static gint        cpk_plot_expose           (GtkWidget         *widget,
		                              GdkEventExpose    *event) ;
static void        cpk_plot_paint            (GtkWidget         *widget,
	                                      GdkRectangle      *area) ;
static void        cpk_plot_paint_all        (CpkPlot           *plot) ;
static GtkWidget  *cpk_plot_drawable         (GtkWidget         *widget) ;
static gint        cpk_plot_timer1           (gpointer           data) ;
static gint        cpk_plot_timer2           (gpointer           data) ;

static void        cpk_plot_size_request     (GtkWidget         *widget,
		                              GtkRequisition    *requisition) ;
static void        cpk_plot_size_allocate    (GtkWidget         *widget,
	                                      GtkAllocation     *allocation) ;
static gint        cpk_plot_configure        (GtkWidget         *widget,
		                              GdkEventConfigure *event) ;
static gint        cpk_plot_motion_notify    (GtkWidget         *widget, 
                                              GdkEventMotion    *event) ;
static gint        cpk_plot_button_press     (GtkWidget         *widget, 
                                              GdkEventButton    *event) ;
static gint        cpk_plot_button_release   (GtkWidget         *widget, 
                                              GdkEventButton    *event) ;
static gint        cpk_plot_leave_notify     (GtkWidget         *widget, 
                                              GdkEventCrossing  *event) ;

static void        cpk_plot_pan_size_allocate(GtkWidget         *widget,
	                                      CpkPlot           *plot) ;
static void        cpk_plot_pan_paint        (GtkWidget         *widget,
	                                      GdkRectangle      *area,
                                              CpkPlot           *plot) ;
static void        cpk_plot_mgf_size_allocate(GtkWidget         *widget,
	                                      CpkPlot           *plot) ;
static void        cpk_plot_mgf_paint        (GtkWidget         *widget,
	                                      GdkRectangle      *area,
                                              CpkPlot           *plot) ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

GtkType
cpk_plot_get_type (void)
{
  static guint plot_type = 0;

  if (!plot_type)
    {
      static const GtkTypeInfo plot_info =
      {
	"CpkPlot",
	sizeof (CpkPlot),
	sizeof (CpkPlotClass),
	(GtkClassInitFunc) cpk_plot_class_init,
	(GtkObjectInitFunc) cpk_plot_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      plot_type = gtk_type_unique (gtk_bin_get_type (), &plot_info);
    }

  return plot_type;
}

static void
cpk_plot_class_init (CpkPlotClass *class)
{
   GtkObjectClass    *object_class ;
   GtkWidgetClass    *widget_class ;

   object_class = (GtkObjectClass*) class ;
   widget_class = (GtkWidgetClass*) class ;

   parent_class = gtk_type_class (gtk_bin_get_type ()) ;

   /* widget_class->draw = cpk_plot_draw ; */
   widget_class->expose_event = cpk_plot_expose ;
   widget_class->configure_event = cpk_plot_configure ;
   widget_class->size_request = cpk_plot_size_request;
   widget_class->size_allocate = cpk_plot_size_allocate;
   widget_class->button_press_event = cpk_plot_button_press ;
   widget_class->button_release_event = cpk_plot_button_release ;
   widget_class->enter_notify_event = NULL ;
   widget_class->leave_notify_event = cpk_plot_leave_notify ;
   widget_class->motion_notify_event = cpk_plot_motion_notify ;
}

static void
cpk_plot_init (CpkPlot *plot)
{
   gint i ;
   plot->label = NULL ;
   plot->pixmap = NULL ;
   plot->pixmap_pan = NULL ;
   plot->pixmap_mgf = NULL ;
   for (i = 0; i < 6; i++) plot->gc[i] = NULL ;

   plot->timer_1 = 0 ;
   plot->timer_2 = 0 ;

   CPK_PLOT_SET_FLAGS (plot, CPK_CONFIGURE 
         | CPK_PAINT
         | CPK_CENTER
         | CPK_CONFIGURE_PAN
         | CPK_CONFIGURE_MGF) ;
}

GtkWidget *
cpk_plot_new (GtkWidget *widget, const gchar *label)
{
   CpkPlot *plot ;
   
   plot = gtk_type_new (cpk_plot_get_type ());

   plot->parent = widget ;

   if (label) 
      plot->label = g_strdup (label) ;

   /* if (widget)
      gtk_container_add (GTK_CONTAINER (plot), widget) ; */

   return GTK_WIDGET (plot) ;
}

static gint
cpk_plot_expose (GtkWidget      *widget,
		 GdkEventExpose *event)
{
   GtkBin *bin ;
   GdkEventExpose child_event ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_PLOT (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
printf("** cpk_plot_expose **: %s\n", CPK_PLOT (widget)->label) ;

   if (GTK_WIDGET_DRAWABLE (CPK_PLOT (widget)->parent)) {
      bin = GTK_BIN (widget) ;

      cpk_plot_paint (widget, &event->area) ;

      child_event = *event ;
      if (bin->child &&
	  GTK_WIDGET_NO_WINDOW (bin->child) &&
	  gtk_widget_intersect (bin->child, &event->area, &child_event.area))
         gtk_widget_event (bin->child, (GdkEvent*) &child_event) ;
    }

   return FALSE ;
}

static void
cpk_plot_paint (GtkWidget    *widget,
	        GdkRectangle *area)
{
   CpkPlot   *plot ;
   GtkWidget *w ;

   printf("** cpk_plot_paint **: %s\n", CPK_PLOT (widget)->label) ;

   g_return_if_fail (widget != NULL) ;
   g_return_if_fail (CPK_IS_PLOT (widget)) ;
   g_return_if_fail (area != NULL) ;

   if (GTK_WIDGET_DRAWABLE (CPK_PLOT (widget)->parent)) {
      plot = CPK_PLOT (widget) ;
      w = plot->parent ;

      if (CPK_PLOT_CONFIGURE (plot)) {
   printf("configure...     ") ;

         /* Allocate pixmap */
         if (plot->pixmap) 
            gdk_pixmap_unref (plot->pixmap) ;

         plot->pixmap = gdk_pixmap_new (
               gtk_widget_get_parent_window (plot->parent),
               widget->allocation.width, widget->allocation.height, -1) ;
         CPK_PLOT_SET_FLAGS (plot, CPK_PAINT) ;
         CPK_PLOT_UNSET_FLAGS (plot, CPK_CONFIGURE) ;
      }
      if (CPK_PLOT_PAINT (plot)) {
   printf("paint...     ") ;
         /* Set pixmap backgound to white */
         gdk_draw_rectangle (plot->pixmap,
               w->style->white_gc, TRUE, 0, 0,
               widget->allocation.width, widget->allocation.height) ;

         /* Plot data */
         if (CPK_PLOT_DRAW (plot)) {
   printf("draw...     ") ;
            cpk_draw_paint (CPK_PLOT_DRAWABLE (widget), 
                            plot->pixmap, 
                            CPK_PLOT_MOVING (plot)) ;}
         CPK_PLOT_UNSET_FLAGS (plot, CPK_PAINT) ;
      }

   printf("draw_pixmap...     ") ;
      gdk_draw_pixmap (plot->parent->window,
                       w->style->fg_gc[GTK_WIDGET_STATE (widget)],
                       plot->pixmap,
                       area->x, area->y,
                       area->x, area->y,
                       area->width, area->height) ;

      if (CPK_PLOT_GRABBED (plot)) {
         GdkPoint points[10] ;
         cpk_poly_plus (points, widget->allocation.width / 2, 
               widget->allocation.height / 2, 0, 10) ;
         gdk_draw_polygon (plot->parent->window, 
                           w->style->mid_gc[GTK_WIDGET_STATE (widget)], 
                           FALSE, points, 9) ;
   printf("grabbed...     ") ;
      }

      if (CPK_PLOT_LABEL (plot)) {
         gint x, y ;
         x = (GTK_CONTAINER (plot)->border_width + 
               widget->style->klass->xthickness) ;
         y = (GTK_CONTAINER (plot)->border_width + 
               widget->style->font->ascent) ;
         gtk_paint_string (w->style, plot->parent->window, 
               GTK_WIDGET_STATE (widget), area, widget, "plot",
               x + 3, y + 3, plot->label) ;
      }

      if (CPK_PLOT_BORDER (plot)) {
         static GdkGC *border_gc = NULL ;
         if (!border_gc) {
            border_gc = gdk_gc_new (plot->parent->window) ;
            cpk_set_fg_color (plot->parent, border_gc, "blue") ;
            gdk_gc_set_line_attributes (border_gc, 3, GDK_LINE_SOLID, 
                  GDK_CAP_BUTT, GDK_JOIN_MITER) ;
         }   
         gdk_draw_rectangle (plot->parent->window, border_gc, FALSE, 0, 0, 
               widget->allocation.width-1, widget->allocation.height-1) ;
      }
   }
   printf("\n");
}

static GtkWidget *
cpk_plot_drawable (GtkWidget *widget)
{
   GtkBin *bin ;
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_PLOT (widget), FALSE) ;
   bin = GTK_BIN (widget) ;
   if (bin->child && GTK_WIDGET_VISIBLE (bin->child))
      return bin->child ;
   else
      return NULL ;
}   


static void
cpk_plot_size_request (GtkWidget      *widget,
		       GtkRequisition *requisition)
{
   CpkPlot *plot ;
   GtkBin *bin;

printf("** cpk_plot_size_request **\n") ;
   g_return_if_fail (widget != NULL);
   g_return_if_fail (CPK_IS_PLOT (widget));
   g_return_if_fail (requisition != NULL);

   plot = CPK_PLOT (widget);
   bin = GTK_BIN (widget);

   requisition->width = (GTK_CONTAINER (widget)->border_width +
                         GTK_WIDGET (widget)->style->klass->xthickness) * 2;

   requisition->height = (GTK_CONTAINER (widget)->border_width +
                          GTK_WIDGET (widget)->style->klass->ythickness) * 2;

   if (bin->child && GTK_WIDGET_VISIBLE (bin->child)) {
      GtkRequisition child_requisition;
      
      gtk_widget_size_request (bin->child, &child_requisition);

      requisition->width += child_requisition.width ;
      requisition->height += child_requisition.height ;
   }
}

static void
cpk_plot_size_allocate (GtkWidget     *widget,
	                GtkAllocation *allocation)
{
   CpkPlot *plot ;
   GtkBin *bin;
   GtkAllocation child_allocation;

printf("** cpk_plot_size_allocate **: %s\n", CPK_PLOT (widget)->label) ;
   g_return_if_fail (widget != NULL);
   g_return_if_fail (CPK_IS_PLOT (widget));
   g_return_if_fail (allocation != NULL);

   plot = CPK_PLOT (widget);
   bin = GTK_BIN (widget);

   if ((widget->allocation.width != allocation->width) ||
       (widget->allocation.height != allocation->height))
      CPK_PLOT_SET_FLAGS (plot, CPK_CONFIGURE) ;

   if (GTK_WIDGET_MAPPED (widget) &&
      ((widget->allocation.x != allocation->x) ||
       (widget->allocation.y != allocation->y) ||
       (widget->allocation.width != allocation->width) ||
       (widget->allocation.height != allocation->height)) &&
      (widget->allocation.width != 0) &&
      (widget->allocation.height != 0))
      gtk_widget_queue_clear (widget);

   widget->allocation = *allocation;

   if (bin->child && GTK_WIDGET_VISIBLE (bin->child)) {
      child_allocation.x = (GTK_CONTAINER (plot)->border_width +
			    GTK_WIDGET (plot)->style->klass->xthickness);
      child_allocation.width = MAX(1, (gint)allocation->width - child_allocation.x * 2);

      child_allocation.y = (GTK_CONTAINER (plot)->border_width +
			    GTK_WIDGET (plot)->style->klass->ythickness);
      child_allocation.height = MAX (1, ((gint)allocation->height - child_allocation.y -
					 (gint)GTK_CONTAINER (plot)->border_width -
					 (gint)GTK_WIDGET (plot)->style->klass->ythickness));

      child_allocation.x += allocation->x;
      child_allocation.y += allocation->y;

      gtk_widget_size_allocate (bin->child, &child_allocation);
   }

}

static gint
cpk_plot_configure (GtkWidget         *widget,
		    GdkEventConfigure *event)
{
   GtkAllocation allocation;
   
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_PLOT (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

printf("** cpk_plot_configure **: %s\n", CPK_PLOT (widget)->label) ;

   allocation.x = event->x ;
   allocation.y = event->y ;
   allocation.width = event->width ;
   allocation.height = event->height ;

   cpk_plot_size_allocate (widget, &allocation) ;

   return FALSE ;
}

static gint
cpk_plot_motion_notify (GtkWidget *widget, GdkEventMotion *event)
{
   CpkPlot *plot ;
   CpkDraw *draw = NULL ;
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_PLOT (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   plot = CPK_PLOT (widget) ;
   if (CPK_PLOT_DRAW (plot))
      draw = CPK_PLOT_DRAWABLE (widget) ;

   CPKDEBUG5 ("cpk_plot_motion_notify", 7) ;

   if (CPK_PLOT_NO_MOTION (plot))
      return FALSE ;

   /* Get cursor coordinates & state */
   if (event->is_hint) 
      gdk_window_get_pointer (event->window, &plot->crs_x, &plot->crs_y, &plot->state) ;
   else {
      plot->crs_x = event->x ;
      plot->crs_y = event->y ;
      plot->state = event->state ;
   }
   /* inserire check su reference per cxc_map */

   if (CPK_PLOT_DRAW (plot))
      cpk_draw_draw_to_world (draw, 
            (gdouble)plot->crs_x, (gdouble)plot->crs_y, 
            &draw->xcur, &draw->ycur) ;

   if (plot->crs_x < 0 || plot->crs_x > widget->allocation.width || 
       plot->crs_y < 0 || plot->crs_y > widget->allocation.height)
      return FALSE ;

   /* Deve notificare la posizione anche al magnifier e 
      se il bottone e' premuto anche al paner */

   if (plot->state & GDK_BUTTON1_MASK /* && map->pixmap != NULL */) {

      if (!CPK_PLOT_GRABBED (plot))
         return FALSE ;
   CPKDEBUG5 ("cpk_plot_motion_notify: button press", 5) ;

      if (CPK_PLOT_DRAW (plot)) {
         cpk_draw_geometry (draw, plot->parent, 
               (gdouble)plot->crs_x, (gdouble)plot->crs_y) ;

      }
         CPK_PLOT_SET_FLAGS (plot, CPK_PAINT | CPK_MOVING) ;
         if (plot->timer_1)
            gtk_timeout_remove (plot->timer_1) ;
         plot->timer_1 = gtk_timeout_add (PLOT_DELAY_1, cpk_plot_timer1, plot) ;
         cpk_plot_paint_all (plot) ;

      return FALSE ;
   }

   if (plot->timer_2)
      gtk_timeout_remove (plot->timer_2) ;
   plot->timer_2 = gtk_timeout_add (PLOT_DELAY_2, cpk_plot_timer2, plot) ;
   return FALSE ;
}

static gint
cpk_plot_button_press (GtkWidget *widget, GdkEventButton *event)
{
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_PLOT (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
printf("Button press **\n");
   CPK_PLOT_SET_FLAGS (CPK_PLOT (widget), CPK_GRABBED) ;
   if (CPK_PLOT_DRAW (CPK_PLOT (widget))) 
      cpk_draw_button_press (CPK_PLOT_DRAWABLE (widget), 
            CPK_PLOT (widget)->parent, 
            (gdouble)event->x, (gdouble)event->y, event->state) ;

   return FALSE ;
}

static gint
cpk_plot_button_release (GtkWidget *widget, GdkEventButton *event)
{
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_PLOT (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

printf("Button release **\n");
   CPK_PLOT_UNSET_FLAGS (CPK_PLOT (widget), CPK_GRABBED) ;
   if (CPK_PLOT (widget)->timer_1)
      cpk_plot_timer1 (widget) ;
   else
      cpk_plot_paint_all (CPK_PLOT (widget)) ;

   if (CPK_PLOT_DRAW (CPK_PLOT (widget)))
      cpk_draw_button_release (CPK_PLOT_DRAWABLE (widget), 
            CPK_PLOT (widget)->parent, (gdouble)event->x, (gdouble)event->y) ;

   return FALSE ;
}

static gint
cpk_plot_leave_notify (GtkWidget *widget, GdkEventCrossing *event)
{
   CpkPlot *plot ;
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_PLOT (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   plot = CPK_PLOT (widget) ;
   CPK_PLOT_UNSET_FLAGS (plot, CPK_GRABBED) ;
   cpk_plot_timer1 (plot) ;

   if (plot->timer_2)
      gtk_timeout_remove (plot->timer_2) ;
   plot->timer_2 = 0 ;
   return FALSE ;
}

static gint
cpk_plot_timer1 (gpointer data)
{
   CpkPlot *plot = CPK_PLOT (data) ;
   CPKDEBUG5 ("cpk_plot_timer1", 4) ;
printf ("%s\n", plot->label);
   GDK_THREADS_ENTER ();
   if (plot->timer_1) {
      plot->timer_1 = 0 ;
      CPK_PLOT_SET_FLAGS (plot, CPK_SCALE | CPK_PAINT) ;
      CPK_PLOT_UNSET_FLAGS (plot, CPK_MOVING) ;
      cpk_plot_paint_all (plot) ;
   }
   GDK_THREADS_LEAVE ();

   return FALSE ;
}

static gint
cpk_plot_timer2 (gpointer data)
{
   CpkPlot *plot = CPK_PLOT (data) ;
   CPKDEBUG5 ("cpk_plot_timer2", 4) ;

   GDK_THREADS_ENTER ();
   if (plot->timer_2) {
      plot->timer_2 = 0 ;

      if (CPK_PLOT_DRAW (plot) && 
          GTK_WIDGET_MAPPED (plot->parent) && 
          GTK_WIDGET_VISIBLE (plot->parent->parent)) {
         gint id ;
         CpkDraw *draw = CPK_PLOT_DRAWABLE (GTK_WIDGET (plot)) ;
         id = cpk_draw_find_nearest (draw, draw->xcur, draw->ycur) ;
         cpk_draw_draw_mark (draw, plot->parent, id, TRUE) ;
      }
   }
   GDK_THREADS_LEAVE ();

   return FALSE ;
}

static void
cpk_plot_paint_all (CpkPlot *plot)
{
   GdkRectangle area ;
   area.x = plot->parent->allocation.x ;
   area.y = plot->parent->allocation.y ;
   area.width = plot->parent->allocation.width ;
   area.height = plot->parent->allocation.height ;
   printf("%d %d %d %d\n",area.x, area.y, area.width, area.height);
   cpk_plot_paint (GTK_WIDGET (plot), &area) ;
}

/*----------------------------------------------------------------------------
 *							Pan functions
 *--------------------------------------------------------------------------*/

gint
cpk_plot_pan_configure (GtkWidget         *widget,
		        GdkEventConfigure *event, 
                        CpkPlot           *plot)
{
   g_return_val_if_fail (widget != NULL, FALSE) ;
  /*  g_return_val_if_fail (GTK_WIDGET_DRAWABLE (widget), FALSE) ; */
   g_return_val_if_fail (event != NULL, FALSE) ;

printf("** cpk_plot_pan_configure **: %s\n", plot->label) ;

   if (CPK_PLOT_CONFIGURE_PAN (plot))
      cpk_plot_pan_size_allocate (widget, plot) ; 
   

   return FALSE ;
}

static void
cpk_plot_pan_size_allocate (GtkWidget     *widget,
	                    CpkPlot       *plot)
{
   if (plot->pixmap_pan) 
         gdk_pixmap_unref (plot->pixmap_pan) ;
   plot->pixmap_pan = gdk_pixmap_new (widget->window,
            widget->allocation.width, widget->allocation.height, -1) ;
   CPK_PLOT_UNSET_FLAGS (plot, CPK_CONFIGURE_PAN) ;
   CPK_PLOT_SET_FLAGS (plot, CPK_PAINT_PAN) ;
}


gint
cpk_plot_pan_expose    (GtkWidget         *widget,
		        GdkEventExpose    *event, 
                        CpkPlot           *plot)
{
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
printf("** cpk_plot_pan_expose **: %s\n", plot->label) ;

   if (CPK_PLOT_CONFIGURE_PAN (plot)) 
      cpk_plot_pan_size_allocate (widget, plot) ; 

   cpk_plot_pan_paint (widget, &event->area, plot) ;

   gdk_draw_pixmap (widget->window, 
         widget->style->fg_gc[GTK_WIDGET_STATE (widget)], plot->pixmap_pan,
         event->area.x, event->area.y, event->area.x, event->area.y,
         event->area.width, event->area.height) ;

   /* Plot frame border */
   if (CPK_PLOT_DRAW (plot))
      cpk_draw_pan_border (CPK_PLOT_DRAWABLE (GTK_WIDGET (plot)), widget) ;

   return FALSE ;
}


static void
cpk_plot_pan_paint (GtkWidget    *widget,
	            GdkRectangle *area,
                    CpkPlot      *plot)
{
   g_return_if_fail (widget != NULL) ;
   g_return_if_fail (CPK_IS_PLOT (plot)) ;
   g_return_if_fail (area != NULL) ;

   if (GTK_WIDGET_DRAWABLE (widget)) {

      if (CPK_PLOT_PAINT_PAN (plot)) {

         /* Set pixmap backgound to white */
         gdk_draw_rectangle (plot->pixmap_pan,
               widget->style->white_gc, TRUE, 0, 0,
               widget->allocation.width, widget->allocation.height) ;

         /* Plot data */
         if (CPK_PLOT_DRAW (plot)) 
            /* cpk_draw_pan (CPK_PLOT_DRAWABLE (GTK_WIDGET (plot))) */ ;

         if (CPK_PLOT_LABEL (plot)) {
            gint x, y ;
            x = (GTK_CONTAINER (plot)->border_width + 
                  widget->style->klass->xthickness) ;
            y = (GTK_CONTAINER (plot)->border_width + 
                  widget->style->font->ascent) ;
            gdk_draw_string (plot->pixmap_pan, widget->style->font, 
                  widget->style->black_gc,x + 3, y + 3, plot->label) ;
         }

         CPK_PLOT_UNSET_FLAGS (plot, CPK_PAINT_PAN) ;
      }


   }
}

/*----------------------------------------------------------------------------
 *							Magnifier functions
 *--------------------------------------------------------------------------*/

gint
cpk_plot_mgf_configure (GtkWidget         *widget,
		        GdkEventConfigure *event, 
                        CpkPlot           *plot)
{
   g_return_val_if_fail (widget != NULL, FALSE) ;
  /*  g_return_val_if_fail (GTK_WIDGET_DRAWABLE (widget), FALSE) ; */
   g_return_val_if_fail (event != NULL, FALSE) ;

printf("** cpk_plot_mgf_configure **: %s\n", plot->label) ;

   if (CPK_PLOT_CONFIGURE_MGF (plot))
      cpk_plot_mgf_size_allocate (widget, plot) ; 

   return FALSE ;
}

static void
cpk_plot_mgf_size_allocate (GtkWidget     *widget,
	                    CpkPlot       *plot)
{
   if (plot->pixmap_mgf) 
         gdk_pixmap_unref (plot->pixmap_mgf) ;
   plot->pixmap_mgf = gdk_pixmap_new (widget->window,
            widget->allocation.width, widget->allocation.height, -1) ;
   CPK_PLOT_UNSET_FLAGS (plot, CPK_CONFIGURE_MGF) ;
   CPK_PLOT_SET_FLAGS (plot, CPK_PAINT_MGF) ;
}


gint
cpk_plot_mgf_expose    (GtkWidget         *widget,
		        GdkEventExpose    *event, 
                        CpkPlot           *plot)
{
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
printf("** cpk_plot_mgf_expose **: %s\n", plot->label) ;

   if (CPK_PLOT_CONFIGURE_MGF (plot)) 
      cpk_plot_mgf_size_allocate (widget, plot) ; 

   cpk_plot_mgf_paint (widget, &event->area, plot) ; 

   gdk_draw_pixmap (widget->window, 
         widget->style->fg_gc[GTK_WIDGET_STATE (widget)], plot->pixmap_mgf,
         event->area.x, event->area.y, event->area.x, event->area.y,
         event->area.width, event->area.height) ;
  return FALSE ;
}


static void
cpk_plot_mgf_paint (GtkWidget    *widget,
	            GdkRectangle *area,
                    CpkPlot      *plot)
{
   static GdkPoint *points = NULL ;
   g_return_if_fail (widget != NULL) ;
   g_return_if_fail (CPK_IS_PLOT (plot)) ;
   g_return_if_fail (area != NULL) ;


   if (GTK_WIDGET_DRAWABLE (widget)) {

      if (CPK_PLOT_PAINT_MGF (plot)) {

         /* Set pixmap backgound to white */
         gdk_draw_rectangle (plot->pixmap_mgf,
               widget->style->white_gc, TRUE, 0, 0,
               widget->allocation.width, widget->allocation.height) ;

         /* Plot center cross */
         if (!points) {
            points = (GdkPoint *) malloc ((size_t)10 * sizeof (GdkPoint)) ;
            cpk_poly_plus (points, widget->allocation.width / 2, 
                     widget->allocation.height / 2, 0, 10) ;
         }
         gdk_draw_polygon (plot->pixmap_mgf, 
                           widget->style->mid_gc[GTK_WIDGET_STATE (widget)], 
                           FALSE, points, 9) ;

         /* Plot data */
         if (CPK_PLOT_DRAW (plot)) {
            /* cpk_draw_mgf () */;
         }

         /* Solo per debug */
         if (CPK_PLOT_LABEL (plot)) {
            gint x, y ;
            x = (GTK_CONTAINER (plot)->border_width + 
                  widget->style->klass->xthickness) ;
            y = (GTK_CONTAINER (plot)->border_width + 
                  widget->style->font->ascent) ;
            gdk_draw_string (plot->pixmap_mgf, widget->style->font, 
                  widget->style->black_gc,x + 3, y + 3, plot->label) ;
         }

         CPK_PLOT_UNSET_FLAGS (plot, CPK_PAINT_MGF) ;
      }

   }
}
