/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <math.h>

#include <gtk/gtk.h>

#include "wcs.h"
#include "platemodel.h"
#include "catalib.h"
#include "cxc_data.h"
#include "cxc_frames.h"
#include "cpk_gtk_util.h"
#include "cxc_gtk_main.h"
#include "cxc_match_set.h"
#include "cxc_model_set.h"

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern boolean   suspend ; /***FIXME: serve?*/
extern gdouble   tunits ;
guint  *mord[4] = { &m_type[POLY].maxord, 
                    &m_type[POLYRAD].maxord, 
                    &m_type[SAO].maxord, NULL} ;

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static void       get_selection   ( GtkWidget *item, Smodel *ModelSet) ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

static void
get_page (GtkNotebook *notebook, GtkNotebookPage *page, gint page_num, 
      gpointer data)
{
   Smodel *ModelSet = data ;
   ModelSet->page = page_num ;
}

static void
get_matchs_val (Smatchs *MatchsSet, CxcData *data)
{
   gboolean spheric = data->f_set->info[0]->spheric ;
   gdouble  tol ;
   gdouble  dev = 0.0 ;

   dev = MAX (data->pmodel->dxdev, data->pmodel->dydev) ;
   tol = MIN (6.0, data->gparam->sr_2+1) * dev ;
   tol = MAX (data->gparam->min_tol2, tol) ;
   cpk_entry_with_units (&MatchsSet->uval[0], tol, spheric) ;
   /* cpk_entry_with_units (&MatchsSet->uval[1], data->gparam->m2_rad, spheric) ;*/
   /*
    * modified 22/07/13
    * original:
    * cpk_entry_with_units (&MatchsSet->uval[1], data->list2->param->m2_rad, spheric) ;
    * problem: the mix_radius m1_rad must be taken from the default value of reference frame
    * and not that of the rotating list
    */
   cpk_entry_with_units (&MatchsSet->uval[1], data->list1->param->m2_rad, spheric) ;

   if (data->gparam->t2_mag < 500.0) {
      sprintf (MatchsSet->value, "%g", data->gparam->t2_mag) ;
      MatchsSet->active = TRUE ;
   }
   else {
      data->gparam->t2_mag = 0.0 ;
      sprintf (MatchsSet->value, "*") ;
      MatchsSet->active = FALSE ;
   }
   MatchsSet->tol_mag = data->gparam->t2_mag ;

   MatchsSet->values[0] =  data->gparam->sr_2 ;
   MatchsSet->values[1] =  data->gparam->nsr_2 ;
}

void
get_model_val (Smodel *ModelSet, CxcData *data)
{
   switch (data->gparam->fitmodel) {
      case POLY :
         ModelSet->m = 0 ;
         break ;
      case POLYRAD :
         ModelSet->m = 1 ;
         break ;
      case SAO :
         ModelSet->m = 2 ;
         break ;
      case NONE :
         ModelSet->m = 3 ;
         break ;
      default :
         ModelSet->m = 0 ;
   }
   ModelSet->degree = (ModelSet->m < 3) ? data->gparam->max_ord : 0 ;

   if (bin_get_count (data->stack->nf) + bin_get_count (data->stack->nf2) > 1) {
      int nf = 0 ;
      while (!bin_check_n (data->stack->nf, nf) && 
             !bin_check_n (data->stack->nf2, nf-32))
         nf++ ;
      printf("HEIAHEIAALLALA nf %d\n",nf);
      memcpy (data->pmodel->crval, data->model[nf]->crval, 2*sizeof(double)) ;
   }
   sprintf (ModelSet->value[0], "%.10f", data->pmodel->crval[0]) ;
   sprintf (ModelSet->value[1], "%.10f", data->pmodel->crval[1]) ;
   sprintf (ModelSet->value[2], "%.10f", data->pmodel->crpix[0]) ;
   sprintf (ModelSet->value[3], "%.10f", data->pmodel->crpix[1]) ;
}

static int
set_matchs_val (Smatchs *MatchsSet, CxcData *data)
{
   gdouble  tol_mag ;

   if (MatchsSet->active) {
      tol_mag = atof (MatchsSet->value) ;
      if (tol_mag <= 0.0) {
         (void) cpk_warning (MatchsSet->window, 
               "Set a valid 'Tol Mag' or switch off the 'Active' button!") ;
         return 1 ;
      }
   }
   else
      tol_mag = 500 ;

   data->gparam->t2_mag = tol_mag ;
   data->gparam->min_tol2 = MatchsSet->uval[0].value ;
   data->gparam->m2_rad = MatchsSet->uval[1].value ; /* obsoleto */
   data->list2->param->m2_rad = MatchsSet->uval[1].value ;
   data->gparam->sr_2 = gtk_spin_button_get_value_as_float (
         GTK_SPIN_BUTTON (MatchsSet->spin[0])) ;
   data->gparam->nsr_2 = gtk_spin_button_get_value_as_int (
         GTK_SPIN_BUTTON (MatchsSet->spin[1])) ;

   return 0 ;
}

void
set_model_val (Smodel *ModelSet, CxcData *data)
{
   gdouble new_val ;
   switch (ModelSet->m) {
      case 0 :
         data->gparam->fitmodel = POLY ;
         break ;
      case 1 :
         data->gparam->fitmodel = POLYRAD ;
         break ;
      case 2 :
         data->gparam->fitmodel = SAO ;
         break ;
      case 3 :
         data->gparam->fitmodel = NONE ;
         break ;
      default :
         data->gparam->fitmodel = POLY ;
   }
   data->gparam->max_ord =  gtk_spin_button_get_value_as_int (
         GTK_SPIN_BUTTON (ModelSet->spin)) ;
   if (ModelSet->page == 0) {
      data->gparam->recenter = FALSE ;
      new_val = atof (gtk_entry_get_text (GTK_ENTRY (ModelSet->entry[0]))) ;
      if (ABS (data->pmodel->crval[0] - new_val) > 1e-10)
         data->pmodel->crval[0] = new_val ;
      new_val = atof (gtk_entry_get_text (GTK_ENTRY (ModelSet->entry[1]))) ;
      if (ABS (data->pmodel->crval[1] - new_val) > 1e-10)
         data->pmodel->crval[1] = new_val ;
   }
   else {
      data->gparam->recenter = TRUE ;
      data->pmodel->crpix[0] = atof (gtk_entry_get_text (
            GTK_ENTRY (ModelSet->entry[2]))) ;
      data->pmodel->crpix[1] = atof (gtk_entry_get_text (
            GTK_ENTRY (ModelSet->entry[3]))) ;
   }
}

static void
get_selection (GtkWidget *item, Smodel *ModelSet)
{
   gint i ;
   for (i = 0; i < 4; i++)
      if (item == ModelSet->item[i]) {
         ModelSet->m = i ;
         GTK_ADJUSTMENT (ModelSet->adj)->upper = (i < 3) ? *mord[i] : 0 ;
         gtk_spin_button_update (GTK_SPIN_BUTTON (ModelSet->spin)) ;
      }
}

GtkWidget *
model_settings (Smodel *ModelSet, cxc_display *display)
{
   GtkWidget   *box, *vbox, *hbox, *bbox, *frame, *label, *spin ;
   GtkWidget   *entry, *option, *menu, *notebook, *event_box ;
   GtkTooltips *ttip ;
   guint        vmax ;

   box = gtk_vbox_new (FALSE, 0) ;

   /* Model Settings */

   ttip = gtk_tooltips_new () ;
   set_font ("-*-times-medium-r-normal--18-140-*-*-*-*-*-*") ;
   frame = gtk_frame_new ("Plate Model settings") ;
   gtk_container_set_border_width (GTK_CONTAINER (frame), 8) ;
   gtk_container_add (GTK_CONTAINER (box), frame) ;
   set_font ("-*-times-medium-r-normal--17-140-*-*-*-*-*-*") ;

   vbox = gtk_vbox_new (FALSE, 0) ;
   gtk_container_set_border_width (GTK_CONTAINER (vbox), 5) ;
   gtk_container_add (GTK_CONTAINER (frame), vbox) ;

   hbox = gtk_hbox_new (FALSE, 0) ;
   gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0) ;

   bbox = gtk_hbox_new (FALSE, 2) ;
   gtk_box_pack_start (GTK_BOX (hbox), bbox, FALSE, FALSE, 3) ;

   label = gtk_label_new ("Function:") ;
   gtk_box_pack_start (GTK_BOX (bbox), label, FALSE, FALSE, 0) ;
   option = gtk_option_menu_new () ;
   menu = gtk_menu_new () ;
   ModelSet->item[0] = make_menu_item ("POLY", 
         GTK_SIGNAL_FUNC (get_selection), ModelSet) ;
   gtk_menu_append (GTK_MENU (menu), ModelSet->item[0]) ;
   ModelSet->item[1] = make_menu_item ("POLYRAD", 
         GTK_SIGNAL_FUNC (get_selection), ModelSet) ;
   gtk_menu_append (GTK_MENU (menu), ModelSet->item[1]) ;
   ModelSet->item[2] = make_menu_item ("SAO", 
         GTK_SIGNAL_FUNC (get_selection), ModelSet) ;
   gtk_menu_append (GTK_MENU (menu), ModelSet->item[2]) ;
   ModelSet->item[3] = make_menu_item ("NONE", 
         GTK_SIGNAL_FUNC (get_selection), ModelSet) ;
   gtk_menu_append (GTK_MENU (menu), ModelSet->item[3]) ;
   gtk_menu_set_active (GTK_MENU (menu), ModelSet->m) ;
   gtk_option_menu_set_menu (GTK_OPTION_MENU (option), menu) ;
   gtk_tooltips_set_tip (GTK_TOOLTIPS (ttip), option, 
         "Model function for plate solution", NULL) ;
   gtk_container_add (GTK_CONTAINER (bbox), option) ;

   bbox = gtk_hbox_new (FALSE, 2) ;
   gtk_box_pack_start (GTK_BOX (hbox), bbox, FALSE, FALSE, 3) ;

   label = gtk_label_new ("Degree:") ;
   gtk_box_pack_start (GTK_BOX (bbox), label, TRUE, TRUE, 0) ;
   vmax = (ModelSet->m < 3) ? *mord[ModelSet->m] : 0 ;
   ModelSet->adj = GTK_OBJECT (
         gtk_adjustment_new (ModelSet->degree, 0, vmax, 1, 1, 1)) ;
   ModelSet->spin = spin = gtk_spin_button_new (
         GTK_ADJUSTMENT (ModelSet->adj), 1, 0) ; 
   gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (spin), TRUE) ;
   gtk_tooltips_set_tip (GTK_TOOLTIPS (ttip), spin, 
         "Model polynomial degree", NULL) ;
   gtk_box_pack_start (GTK_BOX (bbox), spin, FALSE, FALSE, 0) ;


   event_box = gtk_event_box_new () ;
   gtk_container_set_border_width (GTK_CONTAINER (event_box), 5);
   gtk_box_pack_start (GTK_BOX (vbox), event_box, FALSE, FALSE, 0) ;

   bbox = gtk_vbox_new (FALSE, 5) ;
   gtk_container_add (GTK_CONTAINER (event_box), bbox) ;
 
   hbox = gtk_hbox_new (FALSE, 0) ;
   gtk_box_pack_start (GTK_BOX (bbox), hbox, FALSE, FALSE, 0) ;
   label = gtk_label_new ("WCS tangent point:") ;
   gtk_label_set_justify (GTK_LABEL (label),  GTK_JUSTIFY_LEFT) ; /*FIXME ?? */
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0) ;

   ModelSet->notebook = notebook = gtk_notebook_new ();
   gtk_tooltips_set_tip (GTK_TOOLTIPS (ttip), event_box, 
         "Set the origin for World Coordinate System", NULL) ;
   gtk_box_pack_start (GTK_BOX (bbox), notebook, TRUE, TRUE, 0) ;

   frame = gtk_frame_new (NULL) ;
   gtk_container_set_border_width (GTK_CONTAINER (frame), 5);
   label = gtk_label_new ("CRVAL");
   gtk_notebook_append_page (GTK_NOTEBOOK (notebook), frame, label);
   vbox = gtk_vbox_new (FALSE, 3) ;
   gtk_container_add (GTK_CONTAINER (frame), vbox) ;

   bbox = gtk_hbox_new (FALSE, 0) ;
   gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 3) ;
   label = gtk_label_new ("CRVAL1:") ;
   gtk_box_pack_start (GTK_BOX (bbox), label, FALSE, FALSE, 3) ;
   ModelSet->entry[0] = entry = gtk_entry_new () ;
   gtk_entry_set_max_length (GTK_ENTRY (entry), 254) ;
   gtk_entry_set_text (GTK_ENTRY (entry), ModelSet->value[0]) ;
   gtk_signal_connect (GTK_OBJECT (entry), "activate",
         GTK_SIGNAL_FUNC (entry_changed), ModelSet->value[0]) ;
   gtk_signal_connect (GTK_OBJECT (entry), "focus-out-event",
         GTK_SIGNAL_FUNC (entry_focus_out), ModelSet->value[0]) ;
   gtk_widget_set_usize (entry, 120, 0) ;
   gtk_box_pack_end (GTK_BOX (bbox), entry, FALSE, FALSE, 3) ;
   bbox = gtk_hbox_new (FALSE, 0) ;
   gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 3) ;
   label = gtk_label_new ("CRVAL2:") ;
   gtk_box_pack_start (GTK_BOX (bbox), label, FALSE, FALSE, 3) ;
   ModelSet->entry[1] = entry = gtk_entry_new () ;
   gtk_entry_set_max_length (GTK_ENTRY (entry), 254) ;
   gtk_entry_set_text (GTK_ENTRY (entry), ModelSet->value[1]) ;
   gtk_signal_connect (GTK_OBJECT (entry), "activate",
         GTK_SIGNAL_FUNC (entry_changed), ModelSet->value[1]) ;
   gtk_signal_connect (GTK_OBJECT (entry), "focus-out-event",
         GTK_SIGNAL_FUNC (entry_focus_out), ModelSet->value[1]) ;
   gtk_widget_set_usize (entry, 120, 0) ;
   gtk_box_pack_end (GTK_BOX (bbox), entry, FALSE, FALSE, 3) ;

   frame = gtk_frame_new (NULL) ;
   gtk_container_set_border_width (GTK_CONTAINER (frame), 5);
   label = gtk_label_new ("CRPIX");
   gtk_notebook_append_page (GTK_NOTEBOOK (notebook), frame, label);
   vbox = gtk_vbox_new (FALSE, 3) ;
   gtk_container_add (GTK_CONTAINER (frame), vbox) ;
   bbox = gtk_hbox_new (FALSE, 0) ;
   gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 3) ;
   label = gtk_label_new ("CRPIX1:") ;
   gtk_box_pack_start (GTK_BOX (bbox), label, FALSE, FALSE, 3) ;
   ModelSet->entry[2] = entry = gtk_entry_new () ;
   gtk_entry_set_max_length (GTK_ENTRY (entry), 254) ;
   gtk_entry_set_text (GTK_ENTRY (entry), ModelSet->value[2]) ;
   gtk_signal_connect (GTK_OBJECT (entry), "activate",
         GTK_SIGNAL_FUNC (entry_changed), ModelSet->value[2]) ;
   gtk_signal_connect (GTK_OBJECT (entry), "focus-out-event",
         GTK_SIGNAL_FUNC (entry_focus_out), ModelSet->value[2]) ;
   gtk_widget_set_usize (entry, 120, 0) ;
   gtk_box_pack_end (GTK_BOX (bbox), entry, FALSE, FALSE, 3) ;
   bbox = gtk_hbox_new (FALSE, 0) ;
   gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 3) ;
   label = gtk_label_new ("CRPIX2:") ;
   gtk_box_pack_start (GTK_BOX (bbox), label, FALSE, FALSE, 3) ;
   ModelSet->entry[3] = entry = gtk_entry_new () ;
   gtk_entry_set_max_length (GTK_ENTRY (entry), 254) ;
   gtk_entry_set_text (GTK_ENTRY (entry), ModelSet->value[3]) ;
   gtk_signal_connect (GTK_OBJECT (entry), "activate",
         GTK_SIGNAL_FUNC (entry_changed), ModelSet->value[3]) ;
   gtk_signal_connect (GTK_OBJECT (entry), "focus-out-event",
         GTK_SIGNAL_FUNC (entry_focus_out), ModelSet->value[3]) ;
   gtk_widget_set_usize (entry, 120, 0) ;
   gtk_box_pack_end (GTK_BOX (bbox), entry, FALSE, FALSE, 3) ;

   gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), GTK_POS_LEFT) ;
   gtk_signal_connect (GTK_OBJECT (notebook), "switch-page",
                       GTK_SIGNAL_FUNC (get_page), (gpointer) ModelSet) ;

   gtk_widget_show_all (box) ;

   if (bin_get_count (CXC_DATA (display->data)->stack->nf) > 1) {
      CXC_DATA (display->data)->gparam->recenter = FALSE ;
      gtk_widget_set_sensitive (notebook, FALSE) ;
   }
      
   gtk_notebook_set_page (GTK_NOTEBOOK (notebook), 
         (gint) CXC_DATA (display->data)->gparam->recenter) ;

   return box ;
}

/*----------------------------------------------------------------------------
 * 							Private Types
 *--------------------------------------------------------------------------*/

typedef struct tmp_tag {
   GtkWidget   *main_window ;
   GtkWidget   *dialog ;
   Smodel      *ModelSet ;
   cxc_display *display ;
} Mparam ;

static void
model_accept (GtkWidget *widget, Mparam *ModelParam)
{
   CxcData   *data = CXC_DATA (ModelParam->display->data) ;

   if (set_matchs_val (&ModelParam->ModelSet->MatchsSet, data))
      return ;
   set_model_val (ModelParam->ModelSet, data) ;
   data->gparam->min_tol = data->gparam->min_tol2 ;

   data->pmodel->status = BUSY ;
   cxc_buttons_wcs_status (ModelParam->display) ;
   gtk_widget_destroy (ModelParam->dialog) ;

   /* Remember user setted reference */
   if (data->list1 == data->stack->ref)
      ModelParam->display->user_reference = ModelParam->display->nframes ;
   else
      ModelParam->display->user_reference = data->frm1 ;

   /* update the GUI */
   while (gtk_events_pending ())
      gtk_main_iteration () ;

   cxc_data_model_fit (widget, data) ;
   if (ModelParam->display->we_are_exiting)
      quit_program (NULL, ModelParam->display) ;
}

static void
close_window (GtkWidget *window, gpointer data)
{
   cxc_display *display = data ;
   display->model_frames_dialog = NULL ;
}

static void
get_buttons (GtkWidget *dialog, Mparam *ModelParam)
{
   GtkWidget *button ;
   button = gtk_button_new_with_label ("OK") ;
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (model_accept), ModelParam) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_show (button) ;

   button = gtk_button_new_with_label ("Cancel") ;
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) dialog) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_grab_default (button) ;
   gtk_widget_show (button) ;

   button = gtk_button_new_with_label ("Help") ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_show (button) ;
}

static void
get_buttons0 (GtkWidget *dialog, Mparam *ModelParam)
{
   GtkWidget *button ;
   button = gtk_button_new_with_label ("< Back") ;
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (model_accept), ModelParam) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_show (button) ;

   button = gtk_button_new_with_label ("Forward >") ;
   gtk_signal_connect (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (model_accept), ModelParam) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_grab_default (button) ;
   gtk_widget_show (button) ;

   button = gtk_button_new_with_label ("Cancel") ;
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) dialog) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_show (button) ;

   button = gtk_button_new_with_label ("Help") ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_show (button) ;
}

void
plate_model (GtkWidget *widget, cxc_display *display)
{
   GtkWidget *dialog, *box, *vbox ;
   static Smodel  ModelSet ;
   static Mparam  ModelParam ;

   if (display->model_frames_dialog) {
      /* Window is already opened */
      gtk_window_set_transient_for (GTK_WINDOW (display->model_frames_dialog), 
            GTK_WINDOW (display->window)) ;
      return ;
   }
   if (cxc_frame_select (display)) {
      cpk_warning (display->window, 
            "Please select a frame other than reference one.") ;
      return ;
   }

   /* Create a new window */
   display->model_frames_dialog = dialog = gtk_dialog_new () ;

   /* Initialize Salign struct */
   get_matchs_val (&ModelSet.MatchsSet, CXC_DATA (display->data)) ;
   get_model_val (&ModelSet, CXC_DATA (display->data)) ;

   ModelParam.main_window = display->window ;
   ModelParam.dialog = dialog ;
   ModelParam.ModelSet = &ModelSet ;
   ModelParam.display = display ;
   ModelSet.MatchsSet.window = dialog ;

   gtk_window_set_title (GTK_WINDOW (dialog), "CataXcorr: Fit Plate Model") ;
   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE) ;
   gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
         GTK_SIGNAL_FUNC (close_window), (gpointer)display) ;

   /* Main container */
   box = gtk_vbox_new (FALSE, 0) ;
   gtk_container_set_border_width (GTK_CONTAINER (box), 8) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), box, TRUE, TRUE, 0) ;
   cxc_set_reference (box, TRUE, 8, display->user_reference, display) ;

   /* Match Settings */
   vbox = matchs_settings (&ModelSet.MatchsSet, display) ;
   gtk_container_add (GTK_CONTAINER (box), vbox) ;

   /* Plate Model Settings */
   vbox = model_settings (&ModelSet, display) ;
   gtk_container_add (GTK_CONTAINER (box), vbox) ;

   gtk_widget_show_all (box) ;

   /* buttons */
   get_buttons (dialog, &ModelParam) ;

   gtk_widget_show_all (dialog) ;
}
