/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: CataXcorr.c
 * Author 	: Paolo Montegriffo
 * Created on	: October, 2000	
 * Language	: ANSI C
 * Description	: 
 * Version      : 2.0.0 - November 21, 2000
 * Version      : 2.1.0 - July 04, 2001
 * Version      : 2.1.1 - July 11, 2001
 * Version      : 2.1.3 - October 22, 2001
 * Version      :         November 13, 2000 - last revision
 * Version      : 2.1.6 - May 16, 2002 - highly improved version
 * Notes        : 
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <ctype.h>
#include <math.h>
#include <libgen.h>
#include <regex.h>

#if HAVE_GTK
#include <gtk/gtk.h>
#include "cxc_init.h"
#endif

#include "catalib.h"
#include "cataxcorr.h"
#include "cxc_decode.h"
#include "cp_messages.h"
#include "platemodel.h"
#include "starlist.h"
#include "coordinates.h"
#include "model.h"

#include "cataxcorr_p.h"

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

extern StarMatch   *FindShift            (StarList    *sl1,
                                          StarList    *sl2,
                                          int         *nmatch) ;
extern StarMatch   *FindMatch            (StarList    *sl1,
                                          StarList    *sl2,
                                          int         *nmatch) ;
extern StarMatch   *SearchPairs          (StarList    *sl1,
                                          StarList    *sl2,
                                          double       tol_mag,
                                          double       n_sr,
                                          int          n_iter,
                                          double       tol_pos,
                                          double       mix_rad,
                                          double      *pdx,
                                          double      *pdy,
                                          int         *pnmatch) ;
static int          star_list_flag       (StarList    *sl1,
                                          StarList    *sl2) ;
static double       star_match_get_dmag  (StarMatch   *match,
                                          int          nmatch) ;
static double       star_list_calib      (StarMatch   *match,
                                          int          nmatch,
                                          StarList    *sl2) ;
void                model_set_lin_key    (PlateMod    *pmod) ;

static int          model_fit_once       (StarList    *sl1,
                                          StarList    *sl2,
                                          PlateMod    *model,
                                          double      *chisq) ;
static int          model_recenter       (StarList    *sl1,
                                          StarList    *sl2,
                                          PlateMod    *model,
                                          StarMatch   *match,
                                          int          nmatch) ;
static int          model_get_max_order  (PlateMod    *model, 
                                          int          nmatch) ;
static void         model_set_linear     (PlateMod    *model) ;
static void         model_fits_keywords  (PlateMod    *model) ;

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#ifndef MIN
#define MIN(a,b) (((a) <= (b)) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b) (((a) >= (b)) ? (a) : (b))
#endif
#ifndef ABS
#define ABS(a) ((a) < 0 ? (-(a)) : (a))
#endif
#ifndef SQR
#define SQR(a) ((a)*(a))
#endif
#ifndef M_PI
#define M_PI 3.141592653589793238
#endif

#define HYPOT(a,b) (sqrt ((a)*(a)+(b)*(b)))

#define MAXLENINPUT 2048
#define CHI2_TOL 0.001              /* Chi-2 convergence tolerance */

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

#if HAVE_GTK
boolean   no_interface = FALSE ;  /* Activate display for interactive matching */
#else
boolean   no_interface = TRUE ;
#endif
boolean   suspend      = FALSE ;
boolean   install_cmap = FALSE ;

double    IntNull   = 99.999 ; /* Internal Magnitude Null value */

char      command[MAXLENINPUT] = "\0" ;
char     *module        = "CataXcorr" ;
char     *version       = VERSION " - May, 2003" ;

extern    boolean    mag_flag ;
extern    boolean    spheric ;
extern    int        nrif ;
extern    int        verbose ;
extern    FileSet   *f_set ;
extern    def_pars  *param ;
extern    g_pars     gparam ;

double    xx0, xx1, yy0, yy1 ;    /* FIXME: only for Legendre polynomials */

/*----------------------------------------------------------------------------
 * 							Private functions
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 *							Functions code
 *--------------------------------------------------------------------------*/

int 
main (int ac, char **av)
{
   int        i, status = 1 ;

   strcpy (command, *av) ;
   for (i = 1; i < ac; i++) 
      sprintf (command, "%s %s", command, *(av+i)) ;

   cxc_arg_check_help (ac, av) ;

   if (cxc_arg_check_display (&ac, &av))
      no_interface = TRUE ;

   if (no_interface == FALSE) {
#if HAVE_GTK
      /* Initialize Gtk application */
      cxc_gtk_init (ac, av) ;
      status = 0 ;
#else
      cp_error ("Cannot open display: you must first install Gtk library") ;
#endif
   }
   else {
      char       line[MAX_RECORD_LENGTH] = "\n" ;
      char       parfile[1024] = "\0" ;
      int        j ;
      StarMatch *match = (StarMatch*)NULL ;
      int        nmatch = 0 ;
      StarList  *list1 = NullStarList ;
      StarList  *list2 = NullStarList ;
      StarList **list = NULL ;
      StarList  *out ;
      StackTab  *stack = NULL ;
      PlateMod  *pmodel = (PlateMod*)NULL ;

      /* Decode list arguments */
      cxc_arg_list_decode (ac, av) ;

#ifdef debug
   for (i=0;i<f_set->n;i++)
      printf("%g %g %g %g %g %g %2d %g %g\n",
            param[i].eps,param[i].crow_rad,param[i].min_side,param[i].max_side,param[i].scale,param[i].max_shift,
            param[i].mask_shape,param[i].mask_p[0],param[i].mask_p[1]);
#endif
      /* Initialize StackTab structure with reference star list */
      list = (StarList**) malloc ((size_t)f_set->n*sizeof(StarList*)) ;
      list[0] = list1 = star_list_read (f_set->info[nrif], &param[nrif]) ;
      if (list1 == NullStarList) goto exit ;
      /* list1->param = param[0] ; */

      cp_message (1,2,"Initialize StackTab structure\n");
      stack = stack_init (list1, 0, f_set->n) ;
      if (stack == NULL) goto exit ;

      cp_message (0,5,"index    id    mag       xpix      ypix\n") ;
      for (i = 0; i < MIN (6,(int)stack->ref->nstar); i++)
         cp_message (0,4,"%5d %5d %5s %f %f %f\n",i,stack->ref->data[i]->index,
               stack->ref->data[i]->id, stack->ref->data[i]->mag,
               stack->ref->data[i]->xpix, stack->ref->data[i]->ypix);

      /* Loop over files */
      for (j = 1; j < f_set->n; j++){

         if (gparam.mosaic) 
            list1 = list[j-1] ;
         else 
            list1 = list[0] ;

         /* Read star list */
         list[j] = list2 = star_list_read (f_set->info[j], &param[j]) ;
         if (list2 == NullStarList) 
            goto exit ;
         /* list2->param = param[j] ; */

         /* Initialize Plate Model */
         pmodel = model_init (list1, list2, NULL) ;

         /* Project frames on tangent plane */
         star_list_pix2prj (list1, pmodel->crval) ;
         star_list_pix2prj (list2, pmodel->crpix) ;

         /* Flag crowded and masked stars */
         star_list_flag (list1, list2) ;

         /* Align star lists */
         gparam.min_tol = MAX (0.5*list1->param->crow_rad, 5.0*list1->param->eps) ;
         if (star_list_align (list1, list2, pmodel))
            goto exit ;

         /* Set list1 to coinc's cumulative star list */
         list1 = star_list_get_coinc (stack) ;

         /* Must we find and refine plate model transformation ? */
         gparam.min_tol = 10.0 * MAX (pmodel->dxdev, pmodel->dydev) ;
         star_list_register (list1, list2, pmodel) ;

         /* Fit plate model transformation */
         gparam.min_tol = 6.0 * MAX (pmodel->dxdev, pmodel->dydev) ;
         if (gparam.fitmodel != LIN)
            model_fit (list1, list2, pmodel) ;
         if (pmodel->match) 
            free (pmodel->match) ;
         pmodel->match = NULL ;

         /* Search for matching stars */
         cp_message (1,2,"Searching for matching stars\n");
         match = star_list_match (list1, list2, pmodel, &nmatch) ;

         /* Get xyrms ***FIXME****/
         model_get_rms (match, nmatch, pmodel) ;

         /* Resume info */
         if (spheric) strcpy (line, "arcsec\0") ;
         else strcpy (line, "pixel\0") ;
         i = (spheric) ? 3600 : 1 ;
         cp_message (1,1,"Number of matching pairs is \33[1m%d\33[0m\n",nmatch);
         cp_message (0,1,"Computed residuals from model fit in X and Y:\n") ;
         cp_message (0,1,"   X_RMS error = %#.6g [%s]\n",
                i * ABS(pmodel->cdelt[0])*pmodel->dxdev, line) ;
         cp_message (0,1,"   Y_RMS error = %#.6g [%s]\n", 
                i * ABS(pmodel->cdelt[1])*pmodel->dydev, line) ;
         cp_message (0,1,"Computed PIXSIZE = %#.4g [%s]\n", i * 
                    (ABS(pmodel->cdelt[0])+ABS(pmodel->cdelt[1]))/2.0, line) ;

         /* Put the current star list into the coinc structure */
         cp_message (1,2,"Adding  star list into coinc struct\n") ;
         if (stack_add (&stack, list2, j, match, nmatch)) goto exit ;

         STARCHECK (stack->ref->data,stack->ref->nstar) ;

         /* Save Plate Models */
         if (gparam.savetrans == TRUE) {
            sprintf (parfile, "%s.plm", f_set->info[j]->filename) ;
	    model_save (parfile, f_set->info[nrif]->filename, pmodel, f_set->info[j]) ;
         }

         /* Save rotated star list */
         gparam.append = (list1->spheric && !list2->spheric) ;
         if (gparam.fitmodel != NONE) {
            sprintf (parfile, "%s.rot", f_set->info[j]->filename) ;
            out = star_list_transform (list2, pmodel) ;
	    star_list_save (parfile, f_set->info[j], out) ;
            star_list_destroy (out) ;
         }

         /* free memory */
         cp_message (0, 5, "Deallocating correlated index tables\n") ;
         free (match) ;
         match = (StarMatch*)NULL ;
         free (pmodel) ;
         pmodel = (PlateMod*)NULL ;
      }

      /* Save Coincidence Table */
      cp_message (1, 2, "Saving Coincidence Table\n") ;
      coinc_tab_save (stack, f_set->info, gparam.outfile) ;

      /* If we reach this point, then everything was ok */
      status = 0 ;

      exit:

      /* Free memory */
      cp_message (1,4,"Deallocating memory\n") ;
      if (match != (StarMatch*)NULL) free (match) ;
      if (pmodel != (PlateMod*)NULL) free (pmodel) ;
      cp_message (1,4,"Deallocating list\n") ;
      if (list) {
         for (i = 0; i < f_set->n; i++)
            if (list[i]) star_list_destroy (list[i]) ;
         free (list) ;
      }
      cp_message (0, 5, "Deallocating coinc\n") ;
      stack_destroy (stack) ;
      cp_message (0, 5, "Deallocating f_info\n") ;
      if (f_set->info) {
         for (i = 0; i < f_set->n; i++)
            (void) free_carray (f_set->info[i]->labels, f_set->info[i]->n_cols) ;
         free (f_set->info) ;
      }

   } /* End no_interface == FALSE test */

   if (status == 0) 
      cp_message (1, 1, "Done...\n") ;
   else 
      cp_message (1, 1, "Execution failed...\n") ;

   exit (status) ;
}

void
gui_update (void)
{
#if HAVE_GTK
      /* update the GUI */
   if (no_interface == FALSE)
      while (gtk_events_pending ())
         gtk_main_iteration ();
#endif
   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : model_init
 * In        : 
 * Out       : 
 * Purpouse  : 
 * Notes     : If necessary allocate memory for returning model
 *             
 *---------------------------------------------------------------------------*/

PlateMod *
model_init (StarList *sl1, StarList *sl2, PlateMod *model)
{
   int       i ;
   double    tol_pos = 0.99 * MAX (param[nrif].crow_rad, 5.0*param[nrif].eps) ;
   gboolean  recenter = gparam.recenter ;
   CPKDEBUG("model_init") ;

   /* Initialize Plate Model to Null transformation */
   if (model == (PlateMod*)NULL) 
      model = (PlateMod*) malloc (sizeof(PlateMod)) ;
   else {
      if (model->match) 
         free (model->match) ;
   }
   model->match = NULL ;
   model->nmatch = 0 ;

   model->fitmodel = LIN ;
   model->set = m_type[LIN] ;
   model_set_linear (model) ;
   for (i = 0; i < FN_MAXPAR; i++) 
      model->xfit_coeff[i] = model->yfit_coeff[i] = 0.0 ;
   model->xfit_coeff[1] = model->yfit_coeff[2] = 1.0 ;
   model->dxdev = model->dydev = tol_pos/5.0 ;
   model->xchisq = model->ychisq = model->dmag = 0.0 ;

   model->spheric = sl1->spheric ;

   if (sl1 == sl2) 
      recenter = FALSE ;

   if (recenter == FALSE) {

      if (gparam.CRVAL[0] != DBL_MAX && gparam.CRVAL[1] != DBL_MAX) 
         memcpy (model->crval, gparam.CRVAL, (size_t)2*sizeof(double)) ;
      else
         memcpy (model->crval, sl1->origin, (size_t)2*sizeof(double)) ;

      if (sl2->spheric == sl1->spheric)
         memcpy (model->crpix, model->crval, (size_t)2*sizeof(double)) ;
      else
         memcpy (model->crpix, sl2->origin, (size_t)2*sizeof(double)) ;
   }
   else {

      if (gparam.CRPIX[0] != DBL_MAX && gparam.CRPIX[1] != DBL_MAX) 
         memcpy (model->crpix, gparam.CRPIX, (size_t)2*sizeof(double)) ;
      else
         memcpy (model->crpix, sl2->origin, (size_t)2*sizeof(double)) ;

      if (sl2->spheric == sl1->spheric)
         memcpy (model->crval, model->crpix, (size_t)2*sizeof(double)) ;
      else
         memcpy (model->crval, sl1->origin, (size_t)2*sizeof(double)) ;
   }

   memcpy (sl1->origin, model->crval, (size_t)2*sizeof(double)) ;
   if (sl1 != sl2) 
      memcpy (sl2->origin, model->crpix, (size_t)2*sizeof(double)) ;

   model->status = INIT ;

   return model ;
}

static int
star_list_flag (StarList *sl1, StarList *sl2)
{
   /* Flag crowded stars in lists */
   cp_message (1,2,"Flag crowded stars in lists\n");
   star_flag_crowded (sl1, /* 2.0 *  */sl1->param->crow_rad) ;
   IF_ABORT_RETURN (1) ;
   star_flag_crowded (sl2, /* 2.0 *  */sl2->param->crow_rad) ;
   IF_ABORT_RETURN (1) ;

   /* Apply file masks */
   cp_message (1,2,"Apply file masks\n") ;
   star_flag_mask (sl1, sl1->param->mask_shape, sl1->param->mask_p) ;
   IF_ABORT_RETURN (1) ;
   star_flag_mask (sl2, sl2->param->mask_shape, sl2->param->mask_p) ;
   return 0 ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_list_align
 * In        : 
 * Out       : 
 * Purpouse  : 
 *             
 *---------------------------------------------------------------------------*/

int
star_list_align (StarList *sl1, StarList *sl2, PlateMod *model)
{
   StarList  *ss1, *ss2 ;
   StarMatch *match = NULL ;
   int        nmatch = 0 ;
   int        status = 1 ;
   double     chisq, init_tol = 0.0 ;

   if (model->match) free (match) ;
   model->match = NULL ;
   model->nmatch = 0 ;

   /* Get working subsample from input star lists */
   ss1 = star_list_sample (sl1, TRUE) ;
   ss2 = star_list_sample (sl2, TRUE) ;
   IF_ABORT_EXIT ;

   /* Find 1st order transformation */
   if (gparam.find == SHIFT) {
      /* Cross-correlate the two set of positions to find best match */
      if (gparam.mosaic)
         CXC_MESSAGE (0, 2, "Mode: MOSAIC\n") ;
      else
         CXC_MESSAGE (0, 2, "Mode: SHIFT\n") ;
      cp_message (1,2,"Cross-correlating the two set of positions to find best match\n");
      match = FindShift (ss1, ss2, &nmatch) ;
      IF_ABORT_EXIT ;
      if (match == (StarMatch*)NULL) {
         cp_error ("Restart modifying 'max_shift' parameter") ;
         goto exit ;
      }
      cp_message(0,2,"Matched offset between tables with %d bins;\n", nmatch);
   }
   if (gparam.find == MATCH) {
      /* Use triangulations to find best match */
      CXC_MESSAGE (0, 2, "Mode: MATCH\n") ;
      cp_message (1,2,"Looking for matching triangles to guess 1st order transformation\n");
      match = FindMatch (ss1, ss2, &nmatch) ;
      IF_ABORT_EXIT ;
      if (match == (StarMatch*)NULL) {
         cp_error ("Cannot find any reliable transformation: aborting") ;
         cp_error ("Restart modifying 'eps','tune' or 'max_side' parameters") ;
         goto exit ;
      }
      cp_message (0,2,"Matched tables with %d star pairs\n",nmatch);
   }
   if (gparam.find == OTHER) {
      CXC_MESSAGE (0, 2, "Mode: MANUAL\n") ;
      init_tol = MAX (param[nrif].crow_rad, 5.0*param[nrif].eps) ;
      /* Simple search to find matching pairs */
      init_tol = gparam.min_tol ;
      printf("**************init_tol %g\n",init_tol);
      match = SearchPairs (sl1, sl2, 500.0, 2.5, 10, init_tol, 0.0, 
               &model->dxdev, &model->dydev, &nmatch) ;
      if (match == (StarMatch*)NULL) {
         cp_error ("Cannot find any matching star: aborting") ;
         goto exit ;
      }
      IF_ABORT_EXIT ;
   }

   if (gparam.find == SHIFT || gparam.find == MATCH || gparam.find == OTHER) {

      /* Find plate solutions */
      model_set_order (model, 1) ;
      cp_message (1,2,"Fitting %d matching stars with a 1-st order polynomial\n",nmatch);
      if (model_fit_distortion (match, nmatch, model, &chisq))
         goto exit ;
      cp_message (0,2,"Applying current Plate Model to data set\n") ;
      coord_prj2rot (sl2->data, (int)sl2->nstar, model) ;
      if (match) free (match) ;
      match = NULL ;
      nmatch = 0 ;
      init_tol = 0.0 ;
   }

   if (gparam.find == NONE) {
      CXC_MESSAGE (0, 2, "Mode: NONE\n") ;
      /*** FIXME: uguaglia crpix/crval ***/
      if (model->crval[0] != model->crpix[0] || 
          model->crval[1] != model->crpix[1]) {
         model->crpix[0] = model->crval[0] ;
         model->crpix[1] = model->crval[1] ;
         sl2->proj = FALSE ;
         star_list_pix2prj (sl2, model->crpix) ;
      }
      init_tol = MAX (param[nrif].crow_rad*0.5, 5.0*param[nrif].eps) ;
   }
   
   /* Simple search to find matching pairs */
   match = SearchPairs (sl1, sl2, 500.0, 2.5, 10, init_tol, 0.0, 
            &model->dxdev, &model->dydev, &nmatch) ;

   cp_message (0,4,"dxdev: %g; dydev %g\n", model->dxdev, model->dydev) ;

   IF_ABORT_EXIT ;

   /* Calibrate 2nd star list */
   if (match)
      model->dmag = star_list_calib (match, nmatch, sl2) ;
   IF_ABORT_EXIT ;

   model->status = LOW_ALIGN ;

   /* Flag star lists for display mode only */
   if (no_interface == FALSE) {
      int i ;
      for (i = 0; i < (int)sl1->nstar; i++) sl1->data[i]->flag = 1 ;
      for (i = 0; i < (int)sl2->nstar; i++) sl2->data[i]->flag = 1 ;
      for (i = 0; i < nmatch; i++) {
         match[i].s1->flag = 0 ;
         match[i].s2->flag = 0 ;
      }
   }

   status = 0 ;

   exit :

   if (status) {
      if (match != (StarMatch*)NULL) free (match) ;
   }
   else {
      model->match = match ;
      model->nmatch = nmatch ;
   }

   star_list_free (ss1) ;
   star_list_free (ss2) ;

   return status ;
}

static double 
star_match_get_dmag (StarMatch *match, int nmatch)
{
   int    i, n ;
   double dmave = 0.0 ;
   if (nmatch == 0) return dmave ;
   n = 0 ;
   for (i = 0; i < nmatch; i++) {
      if (match[i].s1->mag != IntNull && match[i].s2->mag != IntNull) {
         dmave += match[i].s1->mcal - match[i].s2->mag ;
         n++ ;
      }
   }
   dmave /= (double) MAX (n, 1) ;
   return dmave ;
}

static double 
star_list_calib (StarMatch *match, int nmatch, StarList *sl2)
{
   int    i ;
   double dmave =  star_match_get_dmag (match, nmatch) ;

   cp_message (0,3,"Calibrating 2nd star list: adding %g\n", dmave) ;
   for (i = 0; i < (int)sl2->nstar; i++) 
      sl2->data[i]->mcal = (sl2->data[i]->mag != IntNull) ? 
            sl2->data[i]->mag+dmave : IntNull ;
   return dmave ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_list_register
 * In        : 
 * Out       : 
 * Purpouse  : 
 * Note      : Sigma-rejection factor is fixed to 4.0
 *             
 *---------------------------------------------------------------------------*/

int 
star_list_register (StarList *sl1, StarList *sl2, PlateMod *model)
{         
   int       k, status = 1 ;
   int       nstop, iterate ;
   PlateMod  Model, Model0 ;
   double    chisq = 0.0, chi0 = 0.0, dchi = 1.0, flag_rad ;
   int       tran_order, max_order, ncoeff = 0 ;
   StarList *ss1, *ss2 ;
   int       delete ;

   if (gparam.fitmodel == NONE)
      return 0 ;

   /* Flag crowded stars in lists */
   if (model->status == LOW_ALIGN) {
      cp_message (1,2,"Flag crowded stars in lists\n");
      star_flag_crowded (sl1, sl1->param->crow_rad) ;
      IF_ABORT_RETURN (status) ;

      star_flag_crowded (sl2, sl2->param->crow_rad) ;
      IF_ABORT_RETURN (status) ;

      if (model_fit_once (sl1, sl2, model, &chisq))
         return status ;
   }

   /* Flag crowded stars in lists with unique crowding radius */
   cp_message (1,2,"Flag crowded stars in lists with unique crowding radius\n");
   flag_rad = 5.0 * MAX (model->dxdev, model->dydev) ;
   flag_rad = MAX (flag_rad, gparam.min_tol) ;
   star_flag_crow_r (sl1, flag_rad) ;
   IF_ABORT_RETURN (status) ;

   star_flag_crow_r (sl2, flag_rad) ;
   IF_ABORT_RETURN (status) ;

   /* Get working subsample from input star lists */
   cp_message (1,3,"Get working subsample from input star lists\n") ;
   ss1 = star_list_sample (sl1, FALSE) ;
   ss2 = star_list_sample (sl2, FALSE) ;

   /* Set model */
   Model = *model ;
   Model.fitmodel = POLY ;
   Model.set = m_type[POLY] ;
   max_order = MIN (Model.set.maxord, gparam.Imax_ord) ;
   /* tran_order = 1 ; */
   tran_order = MIN (model->order, max_order) ;
   model_set_order (&Model, tran_order) ;
   CXC_MESSAGE (0, 2, "Resetting model to %d fitting coeff...", 
      Model.set.fncoef[tran_order]) ;
   if (model_fit_once (ss1, ss2, &Model, &chisq))
      goto exit ;
   CXC_MESSAGE (1, 0, "\n") ;

   Model0 = Model ;

   /*
    * Begin the iteration to find matching pairs and transformation coeff.
    */ 
   cp_message (1,2,"Begin iteration to fit star list on reference frame\n") ;
   k = 0 ;
   nstop = 0 ;
   iterate = TRUE ;
   do
   {
      /* Loop at fixed number of fitting coefficents */
      delete = 0 ;
      do
      {
         /* Delete and move up cursor for 11 lines */
         if (delete && verbose == 2) cp_message(0,2,"\33[12M\33[12A\33[0G") ;
         delete = 1 ;
         /* CXC_MESSAGE (0,2,"\33[1m---> Step # %d\33[0m;\n",++k) ; */
         CXC_MESSAGE (0, 2, "---> Step # %d... ", ++k) ;

         switch (model_fit_once (ss1, ss2, &Model, &chisq)) {
            case 1 : case 2 :
               goto exit ;
               break ;
            case 3 :
               iterate = FALSE ;
               break ;
            default :
               ;
         }
         IF_ABORT_EXIT ;

         /* Chi2 convergence test: must be achieved for 2 consecutive times */
         /* dchi = 2.0 * fabs(chisq-chi0) / (chisq+chi0) ; */
         dchi = fabs(chisq-chi0) ;
         chi0 = chisq ;
         CXC_MESSAGE (0,4,"Chi2 variation: %15g\n", dchi) ;
         if (dchi < CHI2_TOL) nstop++ ;
         else nstop = 0 ;
      } while ((dchi > CHI2_TOL || nstop < 2) && k < gparam.max_iter) ;

#ifdef debug
      printf("xfit_coeff: %f %f %f\nyfit_coeff: %f %f %f\ndmag: %f\n",
      Model.xfit_coeff[0],Model.xfit_coeff[1],Model.xfit_coeff[2],
      Model.yfit_coeff[0],Model.yfit_coeff[1],Model.yfit_coeff[2],
      Model.dmag);
#endif

      if (iterate == FALSE) break;

      if (dchi <= CHI2_TOL)
         CXC_MESSAGE (1,2,"Reduced Chi-2 unchanged\n");
      else
         CXC_MESSAGE (1,2,"Done maximum number of iterations\n");

      /* Save current Plate Model and index table */
      Model0 = Model ;
      ncoeff = Model.set.fncoef[tran_order] ;

      /* Can we increase fit polynomial order ? */
      if (tran_order < max_order) {
         tran_order++ ;
         model_set_order (&Model, tran_order) ;
         Model.dxdev *= 1.5 ;
         Model.dydev *= 1.5 ;
         if (k < gparam.max_iter) 
            CXC_MESSAGE (1, 2, "Increasing no. of fitting polynomial coeffs to %d\n", 
                    Model.set.fncoef[tran_order]) ;
      }
      else {
	 iterate = FALSE ;
         CXC_MESSAGE (1,2,"Reached maximum allowed transformation order [%d]: "
               "stop iterating\n", tran_order ) ;
      }
   } while (iterate == TRUE && k < gparam.max_iter) ;

   if (k == gparam.max_iter) 
      cp_message (1,2,"Done maximum number of iterations [%d]: stop iterating\n", gparam.max_iter) ;

   if (model->dxdev < Model0.dxdev || model->dydev < Model0.dydev)
      cp_warning ("Tolerance radius didn't converged: be carefull") ;

#ifdef debug
   printf("Starting  dx %#g dy %#g\n", model->dxdev, model->dydev) ;
   printf("Outcoming dx %#g dy %#g\n", Model0.dxdev, Model0.dydev) ;
#endif

   /* Recover last saved Plate Model and apply to 2nd star list */
   if (ncoeff == 0) {
      cp_warning("Could not refine plate model") ;
   }
   else {
      cp_message (1,2,"Assumed plate model with %d fitted (over %d) "
            "polynomial coeffs\n", ncoeff, Model.set.fncoef[gparam.max_ord]) ;
      *model = Model0 ;
   }
   coord_prj2rot (sl2->data, (int)sl2->nstar, model) ;

   status = 0 ;

   STARCHECK (sl1->data, sl1->nstar) ;
   STARCHECK (sl2->data, sl2->nstar) ;

   exit :

   /* Free memory */
   star_list_free (ss1) ;
   star_list_free (ss2) ;

   return status ;
}

static int
model_fit_once (StarList *sl1, StarList *sl2, PlateMod *model, double *chisq)
{
   StarMatch *match = NULL ;
   int        nmatch = 0, status = 1 ;

   CPKDEBUG("model_fit_once") ;

   if (model->match) free (match) ;
   model->match = NULL ;
   model->nmatch = 0 ;

   /* Search for matching stars */
   cp_message(0,2,"Searching for strictly matching stars\n") ;
   match = SearchPairs (sl1, sl2, gparam.t1_mag, gparam.sr_1, gparam.nsr_1, 
         gparam.min_tol, sl2->param->m1_rad, &model->dxdev, &model->dydev, &nmatch) ;
   if (nmatch == 0) {
      cp_warning("Found 0 matching stars: breaking sequence") ;
      status = 2 ;
      goto exit ;
   }
   if (nmatch < minstar[model->order]) {
      cp_warning("Insufficent number of matching stars [%d found]: breaking loop",nmatch) ;
      status = 3 ;
      goto exit ;
   }
   CXC_MESSAGE (0, 2, "Matched %d stars... ", nmatch) ;
   IF_ABORT_EXIT ;

   /* Find plate solutions */
   CXC_MESSAGE (0, 3, "Fitting... ",
              nmatch, model->set.fncoef[model->order]) ;
   /* CXC_MESSAGE (0, 3, "Fitting %d matches with a %d-coef polynomial... ",
              nmatch, model->set.fncoef[model->order]) ; */
   if (model_fit_distortion (match, nmatch, model, chisq)) {
      status = 3 ;
      goto exit ;
   }
   CXC_MESSAGE (0,2,"Reduced Chi2 = %12g... ", *chisq) ;
   IF_ABORT_EXIT ;

   /* Apply current Plate Model */
   cp_message (0,2,"Applying current Plate Model to data set\n") ;
   coord_prj2rot (sl2->data, (int)sl2->nstar, model) ;
   IF_ABORT_EXIT ;
   status = 0 ;

   exit :

   if (status) {
      if (match) free (match) ;
   }
   else {
      model->match = match ;
      model->nmatch = nmatch ;
   }
      
   return status ;
}

/*-----------------------------------------------------------------------------
 * Function  : model_fit
 * In        : 
 * Out       : 
 * Purpouse  : Fit linear terms of transformation and general plate model
 * Note      : 
 *             
 *---------------------------------------------------------------------------*/

int 
model_fit (StarList *sl1, StarList *sl2, PlateMod *model)
{
   int        i ;
   double     chisq = 0.0 ;
   StarList  *ss1, *ss2 ;
   double     flag_rad ;

   StarMatch *match = NULL ;
   int        nmatch ;

   CPKDEBUG("model_fit") ;

   /* Flag crowded stars in lists */
   cp_message (1,2,"Flag crowded stars in lists\n");
   flag_rad = 2.0 * MAX (model->dxdev, model->dydev) ;
   flag_rad = MAX (flag_rad, gparam.min_tol) ;

   star_flag_crow_r (sl1, flag_rad) ;
   IF_ABORT_RETURN (1) ;

   star_flag_crow_r (sl2, flag_rad) ;
   IF_ABORT_RETURN (1) ;

   /* Get working subsample from input star lists */
   ss1 = star_list_sample (sl1, FALSE) ;
   ss2 = star_list_sample (sl2, FALSE) ;

   /* Search for matching stars */
   cp_message (1,2,"Searching for matching stars\n");
   match = SearchPairs (ss1, ss2, gparam.t2_mag, gparam.sr_2, gparam.nsr_2, 
         gparam.min_tol, sl2->param->m2_rad, &model->dxdev, &model->dydev, &nmatch) ;
   star_list_free (ss1) ;
   star_list_free (ss2) ;
   IF_ABORT_RETURN(1) ;

   if (nmatch == 0) {
      cp_warning ("Found 0 matching stars") ;
      return 1 ;
   }

   if (model->match) 
      free (model->match) ;
   model->match = match ;
   model->nmatch = nmatch ;

   /* Compute projected coordinates ; ***FIXME: CHECKME *** */
   if (gparam.recenter) 
      star_list_pix2prj (sl2, model->crpix) ;
   else
      star_list_pix2prj (sl1, model->crval) ;
   IF_ABORT_RETURN (1) ;

   /* First we fit a linear transformation */
   model_set_linear (model) ;
   model_fit_distortion (match, nmatch, model, &chisq) ;
   IF_ABORT_RETURN (1) ;

   /* Set the requested WCS tan point */
   CXC_MESSAGE (0, 2, "Setting requested WCS tangent point...   ") ;
   if (model_recenter (sl1, sl2, model, match, nmatch))
      return 1 ;
   IF_ABORT_RETURN (1) ;
   MODEL_CHECK (model) ;

   /* Computes FITS keyws ( PCi_j, CDELTi ) in model struct */
   model_fits_keywords (model) ;

   if (gparam.fitmodel == NONE) 
      return 0 ;

   /* Set the plate model function */
   model->fitmodel = gparam.fitmodel ;
   model->set = m_type[gparam.fitmodel] ;
   model_set_order (model, model_get_max_order (model, nmatch)) ;

   /* Fit the full plate model distortion function */
   model_fit_distortion (match, nmatch, model, &chisq) ;
   CXC_MESSAGE (0, 2, "Reduced Chi-2 = %12g;\n", chisq) ;
  
   /* Apply current model to 2nd star list */
   coord_prj2rot (sl2->data, (int)sl2->nstar, model) ;

   /* Flag star lists for display mode only */
   if (no_interface == FALSE) {
      for (i = 0; i < (int)sl1->nstar; i++) sl1->data[i]->flag = 1 ;
      for (i = 0; i < (int)sl2->nstar; i++) sl2->data[i]->flag = 1 ;
      for (i = 0; i < nmatch; i++) {
         match[i].s1->flag = 0 ;
         match[i].s2->flag = 0 ;
      }
   }

   /* model_stats(sl1, sl2, match, nmatch, model, &dxdev, &dydev) ; */
   model_get_rms (match, nmatch, model) ;
   CXC_MESSAGE (2, 2, "Max displacement: dx: %g, dy: %g\n", model->dx, model->dy) ;
   MODEL_CHECK (model) ;

   return 0 ;
}

/*-----------------------------------------------------------------------------
 * Function  : model_recenter
 * In        : 
 *             
 * Out       : 
 * Purpouse  : 
 *---------------------------------------------------------------------------*/

static int
model_recenter (StarList *sl1, StarList *sl2, PlateMod *model, 
                StarMatch *match, int nmatch)
{
   int     i ;
   double  chisq = 0.0 ;
   Star    origin = {NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL} ;
   Star   *Origin = &origin ;

   CPKDEBUG ("model_recenter") ;

#ifdef debug
   printf("old crval %g %g\n", sl1->origin[0], sl1->origin[1]) ;
   printf("old crpix %g %g\n", sl2->origin[0], sl2->origin[1]) ;
#endif
   if (gparam.recenter==FALSE)
      CXC_MESSAGE (0, 2, "Computing new CRPIXi...\n") ;
   else
      CXC_MESSAGE (0, 2, "Computing new CRVALi...\n") ;
   i = 0 ;
   while ((ABS (model->xfit_coeff[0]) > ABS (model->cdelt[0] * 
                model->xfit_coeff[1] + model->cdelt[1] * model->xfit_coeff[2]) /
                1e6 ||
           ABS (model->yfit_coeff[0]) > ABS (model->cdelt[1] * 
                model->yfit_coeff[1] + model->cdelt[1] * model->yfit_coeff[2]) /
                1e6) && i++ < 5) {
      if (gparam.recenter == FALSE) {
         /* Recenter: Get CRPIXs and fit again */
         coord_rot2prj (0.0, 0.0, model, &origin.xprj, &origin.yprj) ;
         coord_prj2pix (&Origin, 1, sl2->spheric, model->crpix) ;
         model->crpix[0] = origin.xpix ;
         model->crpix[1] = origin.ypix ;
         /* Project 2nd starlist again around new origin */
         star_list_pix2prj (sl2, model->crpix) ;
         CXC_MESSAGE (4, 3, "New CRPIXi: %.10f, %.10f\n", 
               model->crpix[0], model->crpix[1]) ;
      }
      else {
         /* Recenter: Get CRVALs and fit again */
         cp_message (0,3,"current crpix %.10f %.10f\n", 
               model->crpix[0], model->crpix[1]) ;
         cp_message (0,3,"current crval %.10f %.10f\n", 
               model->crval[0], model->crval[1]) ;
         cp_message (0,3,"current coeff0 %.10f %.10f\n", 
               model->xfit_coeff[0], model->yfit_coeff[0]) ;
         coord_prj2rot (&Origin, 1, model) ; 
         coord_rot2pix (&Origin, 1, sl1->spheric, model->crval) ;
         
         model->crval[0] = origin.xpix ;
         model->crval[1] = origin.ypix ;
         /* Project 1st starlist again around new origin */
         star_list_pix2prj (sl1, model->crval) ;
         CXC_MESSAGE (4, 3, "New CRVALi %.10f, %.10f\n", 
               model->crval[0], model->crval[1]) ;
      }
      model_fit_distortion (match, nmatch, model, &chisq) ;
      IF_ABORT_RETURN (1) ;
   }
   return 0 ;
}

int 
model_recenter_crpix (StarList *sl2, PlateMod *model) 
{
   double     chisq = 0.0 ;
   int        i, order ;

   StarMatch *match = model->match ;
   int        nmatch = model->nmatch ;

   star_list_pix2prj (sl2, model->crpix) ;
   IF_ABORT_RETURN (1) ;

   /* First we fit a linear transformation */
   order = model->order ;
   model_set_linear (model) ;
   model_fit_distortion (match, nmatch, model, &chisq) ;
   IF_ABORT_RETURN (1) ;

   /* Set the requested WCS tan point */
   CXC_MESSAGE (0, 2, "Setting requested WCS tangent point...   ") ;

   i = 0 ;
   while ((ABS (model->xfit_coeff[0]) > ABS (model->cdelt[0] * 
                model->xfit_coeff[1] + model->cdelt[1] * model->xfit_coeff[2]) /
                1e6 ||
           ABS (model->yfit_coeff[0]) > ABS (model->cdelt[1] * 
                model->yfit_coeff[1] + model->cdelt[1] * model->yfit_coeff[2]) /
                1e6) && i++ < 5) {
      Star  origin = {NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL} ;
      Star *Origin = &origin ;
      coord_rot2prj (0.0, 0.0, model, &origin.xprj, &origin.yprj) ;
      coord_prj2pix (&Origin, 1, sl2->spheric, model->crpix) ;
      
      model->crpix[0] = origin.xpix ;
      model->crpix[1] = origin.ypix ;
      /* Project 2nd starlist again around new origin */
      star_list_pix2prj (sl2, model->crpix) ;
      CXC_MESSAGE (4, 3, "New CRPIXi: %.10f, %.10f\n", 
            model->crpix[0], model->crpix[1]) ;

      model_fit_distortion (match, nmatch, model, &chisq) ;
      IF_ABORT_RETURN (1) ;
   }
   MODEL_CHECK (model) ;

   /* Computes FITS keyws ( PCi_j, CDELTi ) in model struct */
   model_fits_keywords (model) ;

   if (gparam.fitmodel == NONE) 
      return 0 ;

   /* Recover original model order */
   model_set_order (model, order) ;

   /* Fit the full plate model distortion function */
   model_fit_distortion (match, nmatch, model, &chisq) ;
   CXC_MESSAGE (0, 2, "Reduced Chi-2 = %12g;\n", chisq) ;

   /* Apply current model to 2nd star list */
   coord_prj2rot (sl2->data, (int)sl2->nstar, model) ;

   model_get_rms (match, nmatch, model) ;
   CXC_MESSAGE (2, 2, "Max displacement: dx: %g, dy: %g\n", model->dx, model->dy) ;
   MODEL_CHECK (model) ;

   return 0 ;
}

void
model_set_order (PlateMod *model, int tran_order)
{
   int i ;

   CPKDEBUG ("model_set_order") ;

   model->nx_coeff = model->ny_coeff = model->set.fncoef[tran_order] ;
   model->order = tran_order ;
   for (i = 0; i < FN_MAXPAR; i++)
      model->mask[i] = (i < model->set.fncoef[tran_order]) ? 1 : 0 ;

   return ;
}

static int
model_get_max_order (PlateMod *model, int nmatch)
{
   int tran_order = gparam.max_ord ;
   /* Checks transformation order against matching stars number */
   while (nmatch < minstar[tran_order]) 
      tran_order-- ;
   if (tran_order != gparam.max_ord)
      cp_warning ("Low number of matching stars [%d]: lowering transf. order", nmatch) ;
   /* Checks transformation order against matching max model order */
   tran_order = MIN (tran_order, model->set.maxord) ;
   CXC_MESSAGE (0, 2, "Fitting full transformation [%d coeffs]... ", 
         model->set.fncoef[tran_order]) ;
   return tran_order ;
}

/*-----------------------------------------------------------------------------
 * Function  : model_set_linear
 * In        : 
 * Out       : 
 * Purpouse  : 
 *             
 *---------------------------------------------------------------------------*/

static void
model_set_linear (PlateMod *model)
{
   model_set_order (model, 1) ;
   model->cdelt[0] = model->cdelt[1] = 1.0 ;
   model->pc[0][0] = model->pc[1][1] = 1.0 ;
   model->pc[1][0] = model->pc[0][1] = 0.0 ;
   return ;
}

static void
model_fits_keywords (PlateMod *model)
{
   double     sg1, sg2 ;

   /*
    *  Computes FITS keyws ( PCi_j, CDELTi ) in model struct 
    */
   CXC_MESSAGE (0, 2, "Computing linear terms FITS keywords\n") ;

   /* Determines correct sign for CDELTi keyws */
   if (ABS (model->xfit_coeff[1]) > ABS (model->xfit_coeff[2])) {
      sg1 = (model->xfit_coeff[1] > 0.0) ?  1.0 : -1.0 ;
      sg2 = (model->yfit_coeff[2] > 0.0) ?  1.0 : -1.0 ;
   }
   else {
      sg1 = (model->xfit_coeff[2] > 0.0) ? -1.0 :  1.0 ;
      sg2 = (model->yfit_coeff[1] > 0.0) ?  1.0 : -1.0 ;
   }
   if (sg2 < 0.0) {
      sg1 *= -1.0 ;
      sg2 *= -1.0 ;
   }

   /* Set CDELTi */
   model->cdelt[0] = sg1 * HYPOT (model->xfit_coeff[1], model->xfit_coeff[2]) ;
   model->cdelt[1] = sg2 * HYPOT (model->yfit_coeff[1], model->yfit_coeff[2]) ;

   /* Set PCi_j matrix */
   model->pc[0][0] = model->xfit_coeff[1] / model->cdelt[0] ;
   model->pc[0][1] = model->xfit_coeff[2] / model->cdelt[0] ;
   model->pc[1][0] = model->yfit_coeff[1] / model->cdelt[1] ;
   model->pc[1][1] = model->yfit_coeff[2] / model->cdelt[1] ;
   model->xfit_coeff[0] = model->yfit_coeff[0] = 0.0 ;
   model->xfit_coeff[1] = model->yfit_coeff[2] = 1.0 ;
   model->xfit_coeff[2] = model->yfit_coeff[1] = 0.0 ;
   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : model_set_lin_key
 * In        : 
 *             
 * Out       : 
 * Purpouse  : Computes CDELTs, PCj_i and CROTA keywords given 1st order model
 *             coefficents
 * Note      : PCi_j normalization differs from that suggested by Greisen
 *             and Calabretta Paper I but is consistent with that of Paper II
 *             section 6.2
 *---------------------------------------------------------------------------*/

void 
model_set_lin_key (PlateMod *pmod)
{
   double c11 = pmod->xfit_coeff[1] ;
   double c12 = pmod->xfit_coeff[2] ;
   double c21 = pmod->yfit_coeff[1] ;
   double c22 = pmod->yfit_coeff[2] ;

   int    inv = (ABS (c11) > ABS (c12)) ?
                ( (c11*c22 > 0.0) ? 0 : 1 ) :
                ( (c12*c21 < 0.0) ? 0 : 1 ) ;

   double r1 = (inv == 0) ? atan2 ( c21, c11) : atan2 (c21, -c11) ;
   double r2 = (inv == 0) ? atan2 (-c12, c22) : atan2 (c12,  c22) ;

   double r = (cos (r1) < 0.0) ? 
              ang_res_sym ((ang_res_pos (r1) + ang_res_pos (r2)) / 2.0) : 
              (r1 + r2) / 2.0 ;

   /* First compute CDELTs */
   if (ABS (c11) > ABS (c12)) {
      pmod->cdelt[0] = c11 / cos (r) ;
      pmod->cdelt[1] = c22 / cos (r) ;
   }
   else {
      pmod->cdelt[0] = -c12 / sin (r) ;
      pmod->cdelt[1] =  c21 / sin (r) ;
   }
   /* Set CDELT2 positive */
   if (pmod->cdelt[1] < 0.0) {
      pmod->cdelt[0] *= -1.0 ;
      pmod->cdelt[1] *= -1.0 ;
      r = ang_res_sym (r - M_PI) ;
   }
   /* Set PCj_i matrix and CROTA2 */
   if (ABS (c11) > ABS (c12)) {
      pmod->pc[0][0] = cos (r) ;
      pmod->pc[0][1] = c12 / pmod->cdelt[0] ;
      pmod->pc[1][0] = c21 / pmod->cdelt[1] ;
      pmod->pc[1][1] = cos (r) ;
   }
   else {
      pmod->pc[0][0] =  c11 / pmod->cdelt[0] ;
      pmod->pc[0][1] = -sin (r) ;
      pmod->pc[1][0] =  sin (r) ;
      pmod->pc[1][1] =  c22 / pmod->cdelt[1] ;
   }
   pmod->crota2 = RADTODEG (r) ;

   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : model_get_rms
 * In        : 
 * Out       : 
 * Purpouse  : Computes x&y rms errors
 * Note      : xyprj fields are set equal to xy residuals
 *             
 *---------------------------------------------------------------------------*/

void 
model_get_rms (StarMatch *match, int nmatch, PlateMod *model)
{
   int    i ;
   double dx, dy, dx2, dy2, dxy, dmatch, dxdev, dydev ;
   double dxsum = 0.0 ;
   double dysum = 0.0 ;
   double dx2sum = 0.0 ;
   double dy2sum = 0.0 ;
   double dxysum = 0.0 ;
/***FIXME: model non serve ???? ****/
   CPKDEBUG ("model_get_rms") ;

   for (i = 0; i < nmatch; i++) {
      dx = match[i].s2->xrot - match[i].s1->xrot ;
      dy = match[i].s2->yrot - match[i].s1->yrot ;
      dx2 = dx * dx ;
      dy2 = dy * dy ;
      dxy = dx2 + dy2 ;
      dxsum += dx ;
      dysum += dy ;
      dx2sum += dx2 ;
      dy2sum += dy2 ;
      dxysum += sqrt (dxy) ;
   }
   dmatch = (double) MAX (nmatch, 1) ;
   dxdev = sqrt (dx2sum / dmatch) ;
   dydev = sqrt (dy2sum / dmatch) ;
   cp_message (0,3,"old X_RMS error = %#.6g\n", dxdev*3600.0) ;
   cp_message (0,3,"old Y_RMS error = %#.6g\n", dydev*3600.0) ;
   cp_message (0,3,"old XY covariance from residuals: %g\n", dxysum / dmatch) ;

   return ;
}

#include <gsl/gsl_statistics.h>
void 
model_stats (StarMatch *match, int nmatch, PlateMod *model, 
             double *xm, double *ym, double *dxdev, double *dydev)
{
   int    i ;
   double *dx, *dy, cov ;

   CPKDEBUG ("model_stats") ;
   CXC_MESSAGE (1, 3,"\n") ;
   CXC_MESSAGE (1, 3,"*** Model Stats ***\n") ;

   dx = (double*) malloc (nmatch*sizeof(double)) ;
   dy = (double*) malloc (nmatch*sizeof(double)) ;
   for (i = 0; i < nmatch; i++) {
      dx[i] = match[i].s2->xrot - match[i].s1->xrot ;
      dy[i] = match[i].s2->yrot - match[i].s1->yrot ;
   }
   *xm = gsl_stats_mean (dx, (size_t)1, (size_t)nmatch) ;
   *ym = gsl_stats_mean (dy, (size_t)1, (size_t)nmatch) ;
   CXC_MESSAGE (0, 3,"XY means: %g, %g\n", *xm, *ym) ;
   *dxdev = sqrt (gsl_stats_variance_m (dx, (size_t)1, (size_t)nmatch, *xm)) ;
   *dydev = sqrt (gsl_stats_variance_m (dy, (size_t)1, (size_t)nmatch, *ym)) ;

   cov = gsl_stats_covariance (dx, (size_t)1, dy, (size_t)1, (size_t)nmatch) ;
   CXC_MESSAGE (0, 3,"XY stdev: %g, %g\n", *dxdev, *dydev) ;
   CXC_MESSAGE (0, 3,"XY covariance from residuals: %g\n", cov) ;

   free (dx) ;
   free (dy) ;
   return ;
}

StarList *
star_list_transform (StarList *sl, PlateMod *model)
{
   StarList *ret = star_list_dup (sl) ;

   coord_rot2pix (ret->data, (int)ret->nstar, model->spheric, model->crval) ;
   ret->spheric = model->spheric ;
   ret->out_sph = model->spheric ;

   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_list_match
 * In        : 
 *             
 * Out       : 
 * Purpouse  : 
 * Note      : Sigma rejection clipping is performed in SearchPairs
 *---------------------------------------------------------------------------*/

StarMatch *
star_list_match (StarList *sl1, StarList *sl2, PlateMod *model, int *nmatch)
{
   StarMatch *ret = (StarMatch *)NULL ;
   int        i ;

   /* Search for matching stars */
   cp_message (1,2,"Searching for matching stars\n");
   cp_message (0,2,"dx: %#g; dy: %#g\n", model->dxdev, model->dydev) ;

   ret = SearchPairs (sl1, sl2, gparam.t3_mag, gparam.sr_3, gparam.nsr_3, 
         gparam.min_tol, gparam.m3_rad, &model->dxdev, &model->dydev, nmatch) ;

   /* Calibrate 2nd star list */
   if (gparam.find == NONE) 
      model->dmag = star_list_calib (ret, *nmatch, sl2) ;

   /* Flag star lists for display mode only */
   if (no_interface == FALSE) {
      for (i = 0; i < (int)sl1->nstar; i++) sl1->data[i]->flag = 1 ;
      for (i = 0; i < (int)sl2->nstar; i++) sl2->data[i]->flag = 1 ;
      for (i = 0; i < *nmatch; i++) {
         ret[i].s1->flag = 0 ;
         ret[i].s2->flag = 0 ;
      }
   }

   return ret ;
}

