/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name   : coinc_rw.c
 * Author      : Paolo Montegriffo
 * Created on  : September, 2001	
 * Language    : ANSI C
 * Description : Get coincidence table structure
 *             
 *---------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <math.h>

#include "fileinfo.h"
#include "cp_messages.h"
#include "catalib.h"
#include "coinc_rw.h"

/*----------------------------------------------------------------------------
 * 							Private types
 *--------------------------------------------------------------------------*/

typedef struct ind_lab {
   char *label ;
   int   index ;
} ind_lab ;

/*----------------------------------------------------------------------------
 * 							Private functions
 *--------------------------------------------------------------------------*/

static int str_cmp (const char **d1, const char **d2)
{
   return ((strlen (*d1) == strlen (*d2)) ?
            strcmp (*d1, *d2) : (int)(strlen (*d1) - strlen (*d2))) ;
}

static int ilab_cmp (const ind_lab **d1, const ind_lab **d2)
{
   return ((strlen ((*d1)->label) == strlen ((*d2)->label)) ?
            strcmp ((*d1)->label, (*d2)->label) :
            (int)(strlen ((*d1)->label) - strlen ((*d2)->label))) ;
}

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

extern  char  **get_column_s    (int     n_file, 
                                 int     n_col, 
                                 int    *len, 
                                 int    *ndec) ;
static  int   **ranking_s       (char ***stmp, 
                                 int     nrec, 
                                 Finfo **f_info) ;
static  char  **getuniq         (char  **array, 
                                 int    *n) ;

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern FileSet *f_set ;

/*-----------------------------------------------------------------------------
 * Function  : read_coinc
 * In        : filename
 * Out       : coinc structure
 * Purpouse  : read all the informations in the coincidence table
 *             
 *---------------------------------------------------------------------------*/

CoincId * 
read_coinc (char *filein)
{
   FILE    *in ;
   char     line[MAX_RECORD_LENGTH] ;
   char   **toks ;
   char  ***id = (char***)NULL ;
   int      n, sz ;
   int      i, j ;
   CoincId *coinc ;
   Finfo  **f_info = f_set->info ;

   /* Checks the existance of input file */
   if (access (filein, F_OK)) {
      cp_error ("File '%s' does not exist: aborting",filein) ;
      return (CoincId*)NULL ;
   }
   /* Allocate memory for returning structure */
   coinc = (CoincId*) calloc ((size_t)1, sizeof(CoincId)) ;

   /* Open Coinc Table */
   in = fopen (filein, "r") ;

   /* Gets file names */
   coinc->files = (char**) calloc ((size_t)(f_set->n), sizeof(char*)) ;
   for (i = 0; i < f_set->n; i++) {
      coinc->files[i] = (char*) calloc (strlen (f_info[i]->filename)+(size_t)1, sizeof(char)) ;
      strcpy (coinc->files[i], f_info[i]->filename) ;
   }
   ((CoincId*)coinc)->nfiles = f_set->n ;

   /* Counts file records */ 
   i = j = 0 ;
   while (fgets (line, MAX_RECORD_LENGTH, in) != NULL) 
      if (line[0] != '#') i++ ;
      else j++ ;
   ((CoincId*)coinc)->nrec = i ;

   /* Skip header */ 
   rewind (in) ;
   for (i = 0; i < j; i++) fgets (line, MAX_RECORD_LENGTH, in) ;

   /* Allocate memory for ID, IDF & NS */ 
   coinc->id  = (int*) calloc ((size_t)coinc->nrec, sizeof(int)) ;
   coinc->ns  = (unsigned int*) calloc ((size_t)coinc->nrec, sizeof(unsigned int)) ;
   coinc->ns2 = (unsigned int*) calloc ((size_t)coinc->nrec, sizeof(unsigned int)) ;

   id = (char***) calloc ((size_t)coinc->nfiles, sizeof(char**)) ;
   for (i = 0; i < coinc->nfiles; i++)
      id[i] = (char**) calloc ((size_t)coinc->nrec, sizeof(char*)) ;

   i=0 ;
   while (fgets (line, MAX_RECORD_LENGTH, in)!=NULL) {
      toks = tokenize (line," ",&n) ;
      if (n != coinc->nfiles+2) {
         destroy_coinc (coinc) ;
         fclose (in) ;
         return (CoincId*)NULL ;
      }
      ((CoincId*)coinc)->id[i] = atoi (toks[0]) ;
      for (j = 0; j < coinc->nfiles; j++) {
         sz = strlen (toks[j+1])+1 ;
         id[j][i] = (char*)calloc((size_t)sz, sizeof(char)) ;
         strcpy (id[j][i],toks[j+1]) ;
      }
      if (strstr(toks[n-1],":") == NULL) {
	 ((CoincId*)coinc)->ns[i] = (unsigned int) atoi (toks[n-1]) ;
      } else {
	 unsigned long int n1, n2 ;
	 sscanf (toks[n-1], "%lu:%lu", &n1, &n2) ;
	 ((CoincId*)coinc)->ns[i]  = n1 ;
	 ((CoincId*)coinc)->ns2[i] = n2 ;
      }

      free_carray (toks, n) ;
      i++ ;
   }
   /* Close the file */
   fclose (in) ;

   /* Indexing input files */
   coinc->rank = ranking_s (id, coinc->nrec, f_info) ;

   if (id != (char***)NULL) {
      for (i = 0; i < coinc->nfiles; i++) 
         free_carray (id[i], coinc->nrec) ;
      free (id) ;
   }

   if (coinc->rank == (int**)NULL) {
      destroy_coinc (coinc) ;
      return (CoincId*)NULL ;
   }
   return coinc ;
}

static int ** 
ranking_s (char ***stmp, int nrec, Finfo **f_info)
{
   int       **rank = (int**)NULL ;
   int         i, j, k, status = 0 ;
   char      **id = (char**)NULL ;
   ind_lab   **c_list, **f_list ;
   
   /* Allocate memory for output array */
   rank = (int **) calloc ((size_t)f_set->n, sizeof(int*)) ;
   for (i = 0; i < f_set->n; i++)
      rank[i] = (int *) calloc ((size_t)nrec, sizeof(int)) ;

   /* Allocate memory for temp array */
   c_list = (ind_lab**) calloc ((size_t)nrec, sizeof(ind_lab*)) ;
   for (j = 0; j < nrec; j++) 
      c_list[j] = (ind_lab*) calloc ((size_t)1, sizeof(ind_lab)) ;

   /* Loop on list files */
   for (i = 0; i < f_set->n; i++) {

      /* Allocate memory for file index table */
      f_list = (ind_lab**) calloc ((size_t)f_info[i]->n_rows, sizeof(ind_lab*)) ;

      /* Read file ids */
      id = get_column_s (i, 0, &j, &j) ;
      for (j = 0; j < f_info[i]->n_rows; j++) {
         f_list[j] = (ind_lab*) calloc ((size_t)1, sizeof(ind_lab)) ;
	 f_list[j]->label = id[j] ;
	 f_list[j]->index = j ;
      }

      /* Checks input IDs against multiple defined objects */
      j = f_info[i]->n_rows ;
      (void) getuniq (id, &j) ;
      if (j != f_info[i]->n_rows) {
         cp_error ("File '%s' contains %d multiple defined objects: aborting",
         f_info[i]->filename, f_info[i]->n_rows-j) ;
         rank = free_iarray (rank, f_set->n) ;
         return (int**)NULL ;
      }

      /* Put input labels in index table */
      for (j = 0; j < nrec; j++) {
         c_list[j]->label = stmp[i][j] ;
	 c_list[j]->index = j ;
      }

      /* Sort index table for id file and corresponding idf */
      qsort ((void*)f_list, (size_t)f_info[i]->n_rows, sizeof(ind_lab*), 
             (int (*)(const void *, const void *))ilab_cmp) ;
      qsort ((void*)c_list, (size_t)nrec, sizeof(ind_lab*), 
             (int (*)(const void *, const void *))ilab_cmp) ;

      /* Compute the ranking table */
      k = 0;
      for (k = j = 0; j < nrec; j++) {
         if (strcmp (c_list[j]->label,"0") != 0) {
            while (k < f_info[i]->n_rows-1 && 
                   strcmp (f_list[k]->label, c_list[j]->label) != 0) 
               k++ ;
            if (strcmp (f_list[k]->label, c_list[j]->label) != 0) {
               cp_error ("Object '%s' missing in file '%s': aborting", 
                     c_list[j]->label, f_info[i]->filename) ;
               status = 1 ;
               rank = free_iarray (rank, f_set->n) ;
               break ;
            }
	    rank[i][c_list[j]->index] = f_list[k]->index+1 ;
	 }
      }

      /* Free f_list array */
      for (j = 0; j < f_info[i]->n_rows; j++) free (f_list[j]) ;
      free (f_list) ;
      free_carray (id, f_info[i]->n_rows) ;
      if (status) break ;
   }

   /* Free c_list array */
   for (j = 0; j < nrec; j++) free (c_list[j]) ;
   free (c_list) ;

   return rank ; 
}

/*-----------------------------------------------------------------------------
 * Function  : build_coinc
 * In        : list of file names
 * Out       : coinc structure
 * Purpouse  : build a coincidence table structure associating objects from
 *             different files by name
 * Note      : the name is read as a string from the 1st column of each file 
 *             
 *---------------------------------------------------------------------------*/

CoincId * 
build_coinc (void)
{
   CoincId   *coinc = (CoincId*)NULL ;
   ind_lab ***f_list = (ind_lab***)NULL ;
   char    ***id   = (char***)NULL ;
   char     **uniq = (char**)NULL ;
   int        i, j, k, nrec ;
   Finfo    **f_info = f_set->info ;
   int        nfiles = f_set->n ;

   /* Alloc mem for temp arrays */
   id = (char***) calloc ((size_t)nfiles, sizeof(char**)) ;
   f_list = (ind_lab***) calloc ((size_t)nfiles, sizeof(ind_lab**)) ;

   /* Accumulate all the IDs into uniq array */
   /* IDs are read from the 1st column of each input file */
   for (k = nrec = i = 0; i < nfiles; i++) {
      id[i] = get_column_s (i, 0, &j, &j) ;
      f_list[i] = (ind_lab**) calloc ((size_t)f_info[i]->n_rows, sizeof(ind_lab*)) ;
      for (j = 0; j < f_info[i]->n_rows; j++) {
         f_list[i][j] = (ind_lab*) calloc ((size_t)1, sizeof(ind_lab)) ;
         f_list[i][j]->label= id[i][j] ;
	 f_list[i][j]->index = j ;
      }
      /* Sort current file IDs */
      qsort ((void*)f_list[i], (size_t)f_info[i]->n_rows, sizeof(ind_lab*), 
             (int (*)(const void *, const void *))ilab_cmp) ;

      /* Checks input IDs against multiple defined objects */
      j = f_info[i]->n_rows ;
      (void) getuniq (id[i], &j) ;
      if (j != f_info[i]->n_rows) {
         cp_error ("File '%s' contains %d multiple defined objects: aborting",
         f_info[i]->filename, f_info[i]->n_rows-j) ;
         destroy_coinc (coinc) ;
         coinc = (CoincId*)NULL ;
         goto exit ;
      }

      /* Append current file IDs */
      nrec += f_info[i]->n_rows ;
      uniq = (char**) realloc (uniq, (size_t)nrec*sizeof(char*)) ;
      if (uniq == (char**)NULL) {
         cp_error ("Cannot alloc mem for uniq vec : aborting") ;
         destroy_coinc (coinc) ;
         coinc = (CoincId*)NULL ;
         goto exit ;
      }
      for (j = 0; j < f_info[i]->n_rows; j++)
         uniq[k++] = id[i][j] ;
   }
   /* Get uniq IDs */
   uniq = getuniq (uniq, &nrec) ;

   /* Allocate memory for returning structure */
   coinc =   (CoincId*)  calloc ((size_t)1, sizeof(CoincId)) ;
   coinc->id  =  (int*)  calloc ((size_t)nrec, sizeof(int)) ;
   coinc->ns  =  (int*)  calloc ((size_t)nrec, sizeof(int)) ;
   coinc->rank = (int**) calloc ((size_t)nfiles, sizeof(int*)) ;
   for (i = 0; i < nfiles; i++)
      coinc->rank[i] = (int *) calloc ((size_t)nrec, sizeof(int)) ;

   /* Indexing input files */
   for (i = 0; i < nfiles; i++) {
      for (k = j = 0; j < f_info[i]->n_rows; j++) {
         while (strcmp (f_list[i][j]->label, uniq[k]) != 0) k++ ;
	 coinc->rank[i][k] = f_list[i][j]->index+1 ;
         coinc->ns[k] += (int) pow ((double)2, (double)i) ;
      }
   }

   /* Assign other values to coinc struct */
   ((CoincId*)coinc)->nfiles = nfiles ;
   ((CoincId*)coinc)->nrec = nrec ;
   for (j = 0; j < nrec; j++)
      ((CoincId*)coinc)->id[j] = j+1 ;

   /* Gets file names */
   coinc->files = (char**) calloc ((size_t)nfiles, sizeof(char*)) ;
   for (i = 0; i < nfiles; i++) {
      coinc->files[i] = (char*) calloc (strlen (f_info[0]->filename)+(size_t)1, sizeof(char)) ;
      strcpy (coinc->files[i], f_info[i]->filename) ;
   }

   exit :

   /* Free memory */
   if (id != (char***)NULL) {
      for (i = 0; i < nfiles; i++) 
         free_carray (id[i], f_info[i]->n_rows) ;
      free (id) ;
   }
   if (f_list != (ind_lab***)NULL) {
      for (i = 0; i < nfiles; i++) {
         if (f_list[i] != (ind_lab**)NULL) {
            for (j = 0; j < f_info[i]->n_rows; j++) 
               if (f_list[i][j] != (ind_lab*)NULL) free (f_list[i][j]) ;
            free (f_list[i]) ;
         }
      }
      free (f_list) ;
   }
   if (uniq != (char**)NULL ) 
      free (uniq) ;

   return coinc ;
}

/*-----------------------------------------------------------------------------
 * Function  : destroy_coinc
 * In        : 
 * Out       : 
 * Purpouse  : 
 *             
 *---------------------------------------------------------------------------*/

void 
destroy_coinc(CoincId *coinc)
{
   if ( coinc != (CoincId*)NULL ) {
      (void)free_carray(coinc->files, coinc->nfiles) ;
      if ( coinc->id != (int*)NULL ) free(coinc->id) ;
      if ( coinc->ns != (int*)NULL ) free(coinc->ns) ;
      (void)free_iarray(coinc->idf, coinc->nfiles) ;
      (void)free_iarray(coinc->rank, coinc->nfiles) ;
      free(coinc) ;
   }
   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : getuniq
 * In        : array of strings
 * Out       : uniq() version of input array
 * Purpouse  : remove duplicates from the input array; output array is sorted
 *             
 *---------------------------------------------------------------------------*/

static char **
getuniq(char ** array, int * n)
{
   int    i, j = *n ;
   char **p1 = array, **p2 = array ;
   qsort((void *)array, (size_t)(*n), sizeof(char*), 
         (int (*)(const void *, const void *))str_cmp) ;
   *n = 1 ;
   for (i=0; i<j-1; i++ ) {
      if (strcmp(*++p2, *p1)!=0) {
         *++p1 = *p2 ;
         (*n)++ ;
      }
   }
   array = (char**)realloc(array, (size_t)(*n)*sizeof(char*)) ;
   return array ;
}

/*---------------------------------------------------------------------------*/
