/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: findxcorr.c
 * Author 	: Paolo Montegriffo
 * Created on	: October, 2000	
 * Language	: ANSI C
 * Description	: 
 * Version      : 2.0 - November 9, 2000
 * Version      : 2.1 - June    26, 2001
 * Version      : 2.1.3 - November  7, 2001 : Run lint on the code
 * Version      : 2.1.4 - January  25, 2002 : n_sr factor as local param
 * Version      : 2.1.5 - March    15, 2002 : sigma rej performed with 
 *                                            top quartile interval
 * Version      : 2.1.9 - September 18, 2002 : n_iter parameter added
 * Version      : 2.1.9 - October   03, 2002 : tol_pos parameter added
 * Version      : 2.1.11 - December 23, 2002 : Output type changed
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <math.h>

#include "cataxcorr.h"
#include "starlist.h"
#include "cp_messages.h"

/*----------------------------------------------------------------------------
 * 							Private Types
 *--------------------------------------------------------------------------*/

typedef struct Pair_tag {
   int    i ;
   int    j ;
   double dr ;
} Pair ;

struct p {
   double d ;
   int    j ;
} ;

/*----------------------------------------------------------------------------
 * 							Private functions
 *--------------------------------------------------------------------------*/

static int sort_p (const struct p *d1, const struct p *d2)
{
   if ( d1->d >= d2->d )
      return 1 ; 
   else
      return -1 ;
}

static int sort_pair (const Pair *d1, const Pair *d2)
{ 
   if ( d1->dr >= d2->dr ) 
      return 1 ; 
   else 
      return -1 ; 
}

static int sort_star_y (const Star **d1, const Star **d2)
{ 
   if ( (*d1)->yrot >= (*d2)->yrot ) 
      return 1 ; 
   else 
      return -1 ;
}

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern  g_pars    gparam ;

/*----------------------------------------------------------------------------
 * 							Function prototype
 *--------------------------------------------------------------------------*/

StarMatch *SearchPairs(StarList *sl1, StarList *sl2, double tol_mag, 
                       double n_sr, int n_iter, double tol_pos, double mix_rad, 
                       double *pdx, double *pdy, int *pnmatch) ;

static double quantile (Pair *match, int nmatch, double f) ;

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#ifndef ABS
#define ABS(a) ((a) < 0 ? (-(a)) : (a))
#endif
#ifndef SQR
#define SQR(x) ((x)*(x))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<=(b)) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b) (((a)>=(b)) ? (a) : (b))
#endif

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
 *
 * Function  : SearchPairs
 * In        : Two star list, tolerance on magnitudes, sigma rejection factor,
 *             minimum tolerance on positions, sigma rejection clipping factor,
 *             max number of SR iterations, estimation of x_rms and y_rms 
 *             errors
 * Out       : Number of matching stars found, index table for matching stars
 *             refined x_rms and y_rms errors
 *
 * Purpouse  : Given the position of a point from the first list, looks for
 *             stars from the second one falling within the tolerance 
 *             given by 5 times input x_rms and y_rms errors;
 *             if more than one star is found, select the closest one
 *
 * Note      : Added a sigma clipping on matching pairs
 *             The algorithm is highly optimized for speedy search
 *             The 2nd star list MUST ALREADY BE CALIBRATED IN MAGNITUDES mcal
 *
 *---------------------------------------------------------------------------*/

StarMatch *
SearchPairs (StarList *sl1, StarList *sl2, double tol_mag, 
             double n_sr, int n_iter, double tol_pos, double mix_rad, 
             double *pdx, double *pdy, int *pnmatch)
{
   StarMatch *ret = (StarMatch*)NULL ;
   double     tol_x, tol_y ;
   double     dxsum = 0.0, dysum = 0.0 ;
   double     dxdev = 0.0, dydev = 0.0 ;
   double     drdev = 0.0, drlim ;
   double     dx = 0.0, dy = 0.0, dxy, dmag, tol ;
   int        i, j, j1, j2, nexp, nexp0 ;
   int        npair = 0, nmatch = 0, nrej, trej = 0, iter = 0 ;
   Pair      *pair = (Pair*)NULL ;
   Pair      *match = (Pair*)NULL ;
   struct p   pb[11] ;
   Star     **s1, **s2, **sd1, **sd2 ;
   int        nstar1 = (int)sl1->nstar ;
   int        nstar2 = (int)sl2->nstar ;

   CPKDEBUG ("SearchPairs") ;

   /* Get tol_radii */
   tol_x = MAX (mix_rad, *pdx * MAX(5.0, n_sr)) ;
   tol_y = MAX (mix_rad, *pdy * MAX(5.0, n_sr)) ;
   /* if (tol_pos > 0.0) */
   if (tol_pos > mix_rad) 
      tol_x = tol_y = tol_pos ;

   *pnmatch = 0 ;
   cp_message(0,2,"Tolerance: x: %#.4g ; y: %#.4g;  tol_pos: %#.4g;\n", tol_x, tol_y, tol_pos) ;

   /* Initialize star flags */
   s1 = sl1->data ;
   s2 = sl2->data ;
   sd1 = star_ptr_dup (sl1->data, sl1->nstar) ;
   sd2 = star_ptr_dup (sl2->data, sl2->nstar) ;
   for (i = 0; i < nstar1; i++) s1[i]->flag = i ;
   for (i = 0; i < nstar2; i++) s2[i]->flag = i ;

   /* Sort star lists for Y-coordinate and magnitudes */
   cp_message(0,4,"Sorting star lists for coordinates\n") ;
   qsort ((void*)sd1, (size_t)nstar1, sizeof(Star*), 
          (int (*)(const void *, const void *))sort_star_y) ;
   qsort ((void*)sd2, (size_t)nstar2, sizeof(Star*), 
          (int (*)(const void *, const void *))sort_star_y) ;

   /* Scan star lists for matching pairs */
   /* Highly optimized for speed search */
   nexp = nexp0 = MAX (nstar1, nstar2) ;
   pair = (Pair*) malloc ((size_t)nexp*sizeof(Pair)) ;
   if (pair == (Pair*)NULL) {
      cp_error("Could not alloc memory for matching pairs") ;
      goto exit ;
   }
   i = j1 = j2 = npair = 0 ;
   while (i < nstar1) {
      for (j = 0; j < 10; j++) pb[j].d = 2.0 ;

      while (j1 < nstar2-1 && sd2[j1]->yrot-sd1[i]->yrot < -tol_y)
	 j1++ ;
      j2 = MAX (j1, j2) ;
      while (j2 < nstar2-1 && sd2[j2]->yrot-sd1[i]->yrot < tol_y)
         j2++ ;

      for (j = j1; j <= j2; j++) {
         dx = sd2[j]->xrot - sd1[i]->xrot ;
         dy = sd2[j]->yrot - sd1[i]->yrot ;
         dxy = SQR (dx/tol_x) + SQR (dy/tol_y) ;
         dmag = ABS (sd2[j]->mcal - sd1[i]->mcal) ;
         if (dxy < 1.0 && dmag < tol_mag) {
            /* Save best 10 matches */
            if (dxy < pb[9].d) {
               pb[9].d = dxy ;
               pb[9].j = j ;
               qsort ((void*)&pb, (size_t)10, sizeof(struct p), 
                      (int (*)(const void *, const void *))sort_p) ;
            }
         }
      }
      j = 0 ;
      while (j < 8 && pb[j].d < 2.0) {
         if (npair >= nexp-10) {
	    /* Reallocate mem for current pair */
            nexp += nexp0 ;
            pair = (Pair*) realloc (pair, (size_t)nexp*sizeof(Pair)) ;
            if (pair == (Pair*)NULL) {
               cp_error("Could not alloc memory for matching pairs");
               goto exit ;
            }
         }
         pair[npair].i = sd1[i]->flag ;
	 pair[npair].j = sd2[pb[j].j]->flag ;
	 pair[npair].dr = pb[j].d ;
         npair++ ;
         j++ ;
      }
      i++ ;
   }

   cp_message (0, 4, "npair %d\n", npair) ;
   if (npair == 0) {
      cp_warning("Could not find any matching stars") ;
      goto exit ;
   }

   /* Sort pairs for increasing dr's */
   qsort ((void*)pair, (size_t)npair, sizeof(Pair), 
          (int (*)(const void *, const void *))sort_pair) ;

   /* Allocate memory for true pairs table */
   match = (Pair*) malloc ((size_t)npair*sizeof(Pair)) ;

   /* Get true pairs */
   for (nmatch = i = 0; i < npair; i++) {
      if (s1[pair[i].i]->flag >= 0 && s2[pair[i].j]->flag >= 0) {
         s1[pair[i].i]->flag = s2[pair[i].j]->flag = -2 ;
         (void) memcpy (&match[nmatch], &pair[i], sizeof(Pair)) ;
         nmatch++ ;
      }
   }

   cp_message (0, 4, "nmatch %d\n", nmatch) ;
   if (nmatch > 2) 
      drdev = quantile (match, nmatch, 0.83) ;
   /* drdev = match[(int)(0.75*nmatch)].dr ; */

   /* Get best matches applying a recursive Sigma clipping */
   cp_message (0, 2, "Applying sigma rejection clipping:\n") ;
   cp_message (0, 4, "drdev=%12g; nmatch %d\n", drdev, nmatch) ;
   tol = (tol_x + tol_y) / 2.0 ;
   nrej = 10 ;
   while (nrej != 0 && iter < n_iter && nmatch > 2) {
      i=0 ;
      /* drdev = match[(int)(0.75*nmatch)].dr ; */
      drdev = quantile (match, nmatch, 0.83) ;
      /***FIXME: ricontrollare la normalizzazione !!! */
      drlim = MAX (n_sr*drdev, SQR (mix_rad/tol)) ;
      /* we don't want drlim let goto zero */
      drlim = MAX (drlim, SQR (0.5*sl1->param->eps/tol)) ;
      while (i < nmatch && match[i].dr <= drlim) 
         i++ ;
      if (i == 0) goto exit ;
      nrej = nmatch-i ;
      nmatch = i ;
      iter++ ;
      trej += nrej ;
   }

   /* Allocate memory for returning index table */
   cp_message (0,6,"Reallocate memory for match matrix\n") ;
   if (!(ret = (StarMatch*) malloc ( (size_t)nmatch * sizeof(StarMatch)))) {
      cp_error ("Could not alloc memory for index table for matching stars") ;
      goto exit ;
   }

   /* Fill returning index table and computes xy_dev */
   dxsum = dysum = dxdev = dydev = 0.0 ;
   for (i = 0; i < nmatch; i++) {
      ret[i].s1 = s1[match[i].i] ;
      ret[i].s2 = s2[match[i].j] ;
      dx = s2[match[i].j]->xrot - s1[match[i].i]->xrot ;
      dy = s2[match[i].j]->yrot - s1[match[i].i]->yrot ;
      dxsum += dx ;
      dysum += dy ;
      dxdev += dx*dx ;
      dydev += dy*dy ;
   }
#ifdef debug
   printf("last dx %#.6g dy %#.6g --  %#.6g\n",dx,dy,sqrt(SQR(dx)+SQR(dy)));
#endif

   /* Compute new standard deviations of dx and dy from the mean */
   dxdev = sqrt ((dxdev-SQR(dxsum)/(double)nmatch) / (double)MAX(nmatch-1,1)) ;
   dydev = sqrt ((dydev-SQR(dysum)/(double)nmatch) / (double)MAX(nmatch-1,1)) ;

   /* Get new x y stdev and nmatch */
   /* if (dxdev < *pdx) *pdx = MAX(dxdev, sl1->param->eps) ;
   if (dydev < *pdy) *pdy = MAX(dydev, sl1->param->eps) ; */

   /* We dont want dxydev let go to zero nor become too big */
   if (tol_pos > 0) {
      *pdx = MAX (MIN (dxdev, tol_pos), sl1->param->eps/4.0) ;
      *pdy = MAX (MIN (dydev, tol_pos), sl1->param->eps/4.0) ;
   } else {
      *pdx = MAX (dxdev, sl1->param->eps/4.0) ;
      *pdy = MAX (dydev, sl1->param->eps/4.0) ;
   }
   
   *pnmatch = nmatch ;
   cp_message (0, 4, "drdev=%12g; dxdev: %12g; dydev: %12g;\n", 
         drdev, dxdev, dydev) ;

   /* Output more info */
   if (nrej == 1) cp_message (0, 2, "Rejected %5d pair\n", trej) ;
   else cp_message (0, 2, "Rejected %5d pairs\n", trej) ;
   cp_message (0, 2, "Matched %5d stars;\n", nmatch) ;

   exit:

   /* Free memory */
   cp_message (0, 6, "Deallocating memory\n") ;
   if (sd1) free (sd1) ;
   if (sd2) free (sd2) ;
   if (pair != (Pair*)NULL) free (pair) ;
   if (match != (Pair*)NULL) free (match) ;
   return ret ;
}

static double
quantile (Pair *match, int nmatch, double f)
{
   int    i = floor(((double)nmatch-1.0)*f) ;
   double r = ((double)nmatch-1.0)*f - (double)i ;
   return (1-r)*match[i].dr + r * match[i+1].dr ;
}
