/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: fitplate.c
 * Author 	: Paolo Montegriffo
 * Created on	: October, 2000	
 * Language	: ANSI C
 * Description	: 
 * Version      : 2.0 - November 9, 2000
 *              : 2.1.4 - February 8, 2002 : fit is made on *prj coords
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <math.h>

#include "cataxcorr.h"
#include "cataxcorr_p.h"
#include "platemodel.h"
#include "cp_messages.h"
#include "fitplate.h"

/*----------------------------------------------------------------------------
 * 							Private Types
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static void   SigRej          (Star **s1, Star **s2, int n, 
                               Match *match, double *dmave) ;
static double get_dmag        (Star **s1, Star **s2, Match *match, int nmatch) ;

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#ifndef MAX
#define MAX(a,b) (((a)>=(b)) ? (a) : (b))
#endif
#ifndef ABS
#define ABS(a) ((a) < 0 ? (-(a)) : (a))
#endif

#if HAVE_GSL
#define fit_module gslfit
#else
#define fit_module qrfit
#endif
#ifndef fit_module
#define fit_module svdfit
#endif


/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern double   IntNull ;     /* Internal value for null magnitudes */
extern boolean  suspend ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
 * Function  : FitPlate
 * In        : Stars to fit, index table 
 *             for matching stars and number of matches
 * Out       : updated plate model structure
 * Purpouse  : fit the matched list and reference stars with a polynomial
 *             to find the plate solution of the field.
 *             A SAO polynomial fit is used to model the distortion
 * Note      : 
 *---------------------------------------------------------------------------*/

int 
FitPlate(Star **s1, Star **s2, Match *match, int nmatch, 
         PlateMod *pmod, double *chisq)
{

   int      i ;
   int      status = 0 ;
   int      ncoeff = 0 ;
   double * x_1 = (double*)NULL ;
   double * y_1 = (double*)NULL ;
   double * x_2 = (double*)NULL ;
   double * y_2 = (double*)NULL ;
   double * x_s = (double*)NULL ;
   double * y_s = (double*)NULL ;
   double   xchisq, ychisq ;
   double   dmave = 0.0 ;

   /* Allocate memory for working arrays */
   if (!(x_1 = (double*) malloc ((size_t)nmatch * sizeof (double))) ||
       !(y_1 = (double*) malloc ((size_t)nmatch * sizeof (double))) ||
       !(x_s = (double*) malloc ((size_t)nmatch * sizeof (double))) ||
       !(y_s = (double*) malloc ((size_t)nmatch * sizeof (double))) ||
       !(x_2 = (double*) malloc ((size_t)nmatch * sizeof (double))) ||
       !(y_2 = (double*) malloc ((size_t)nmatch * sizeof (double))) ){
      cp_error("Could not alloc memory for finding plate solution");
      status = 1 ;
      goto exit ;
   }

   /* Set relative pixel coordinates and individual standard deviations */
printf("\n");
  printf("crval %g %g\n", pmod->crval[0], pmod->crval[1]) ;

   for (i=0; i<6; i++)
      printf("pix %10f %10f prj %10f %10f rot %10f %10f\n",
            s1[i]->xpix,s1[i]->ypix,s1[i]->xprj,s1[i]->yprj,s1[i]->xrot,s1[i]->yrot);
printf("\n");
  printf("crpix %g %g\n", pmod->crpix[0], pmod->crpix[1]) ;
   for (i=0; i<6; i++)
      printf("pix %10f %10f prj %10f %10f rot %10f %10f\n",
            s2[i]->xpix,s2[i]->ypix,s2[i]->xprj,s2[i]->yprj,s2[i]->xrot,s2[i]->yrot);

   for (i=0; i<nmatch; i++) {
      x_1[i] = s1[match[i].i]->xprj ;
      y_1[i] = s1[match[i].i]->yprj ;
      x_s[i] = MAX (pmod->dxdev, 1e-7) ;
      y_s[i] = MAX (pmod->dydev, 1e-7) ;
      x_2[i] = s2[match[i].j]->xprj ;
      y_2[i] = s2[match[i].j]->yprj ;
   }
   pmod->pc[0][0] = pmod->pc[1][1] = 1.0 ;
   pmod->pc[1][0] = pmod->pc[0][1] = 0.0 ;
   pmod->cdelt[0] = pmod->cdelt[1] = 1.0 ;

   /* Fit X and Y model coefficents */
   gui_update () ;
   if (suspend)
      goto exit ;
   (void) fit_module(x_2, y_2, x_1, x_s, nmatch, pmod->xfit_coeff, 
                     pmod->mask, (int)FN_MAXPAR, &xchisq, pmod->set.PlateFunc) ;
   gui_update () ;
   if (suspend)
      goto exit ;
   (void) fit_module(x_2, y_2, y_1, y_s, nmatch, pmod->yfit_coeff, 
                     pmod->mask, (int)FN_MAXPAR, &ychisq, pmod->set.PlateFunc) ;

   /* Get reduced chi-2 */
   for ( ncoeff = i = 0; i<(int)FN_MAXPAR; i++)
      if ( pmod->mask[i] == 1 ) ncoeff++ ;

   *chisq = (xchisq+ychisq)/(double)MAX(nmatch-ncoeff,1) ;

   /* Get relative magnitude zero point */
   dmave = get_dmag(s1, s2, match, nmatch) ;

   /* Apply sigma-clipping to refine dmave */
   SigRej(s1, s2, nmatch, match, &dmave) ;

   /* Fill plate model structure */
   pmod->nx_coeff = MAX(ncoeff, pmod->nx_coeff) ;
   pmod->ny_coeff = MAX(ncoeff, pmod->ny_coeff) ;
   pmod->xchisq = xchisq/(double)MAX(nmatch-ncoeff,1) ;
   pmod->ychisq = ychisq/(double)MAX(nmatch-ncoeff,1) ;
   pmod->dmag = dmave ;
   pmod->nmatch = nmatch ;

   printf("xfit_coeff: %g %g %g\nyfit_coeff: %g %g %g\ndmag: %f\n",
   pmod->xfit_coeff[0],pmod->xfit_coeff[1],pmod->xfit_coeff[2],
   pmod->yfit_coeff[0],pmod->yfit_coeff[1],pmod->yfit_coeff[2],
   pmod->dmag);

   exit:

   /* Free memory */
   if( x_1 != (double*)NULL ) free(x_1) ;
   if( y_1 != (double*)NULL ) free(y_1) ;
   if( x_s != (double*)NULL ) free(x_s) ;
   if( y_s != (double*)NULL ) free(y_s) ;
   if( x_2 != (double*)NULL ) free(x_2) ;
   if( y_2 != (double*)NULL ) free(y_2) ;

   return (status) ;
}

static void SigRej(Star **s1, Star **s2, int n, Match *pair, double *dmave)
{
   double dmsum, dmdev, dm ;
   int    i, nn, iter = 0, nrej = n, nmin = n/2000+1 ;

   for (i=0; i<n; i++) 
      pair[i].f = 0 ;
   cp_message(0,4,"Starting mean: %g\n",*dmave) ;

   while ( nrej > nmin && iter<50 ) {
      /* Get the standard deviation */
      dmsum = dmdev = 0.0 ;
      for ( nn = i = 0; i<n; i++ ) {
         if ( pair[i].f==0 && 
              s1[pair[i].i]->mag != IntNull && 
              s2[pair[i].j]->mag != IntNull ) {
            dm = s1[pair[i].i]->mag - s2[pair[i].j]->mag - *dmave ;
            dmsum += dm ;
            dmdev += dm*dm ;
            nn++ ;
         }
      }
      nn = MAX(nn, 2) ;
      dmdev = sqrt( ( dmdev-dmsum*dmsum/(double)nn ) / (double)(nn-1) ) ;
      /* we don't want to see dmdev goto zero */
      dmdev = MAX(dmdev, 0.05) ;

      /* Flag bad points and get new mean value */
      nrej = 0 ;
      dmsum = 0.0 ;
      for ( nn = i = 0; i<n; i++ ) {
         if ( pair[i].f==0 && 
              s1[pair[i].i]->mag != IntNull && 
              s2[pair[i].j]->mag != IntNull ) {
            dm = s1[pair[i].i]->mag - s2[pair[i].j]->mag ;
            if(  ABS( dm - *dmave ) > 2.0*dmdev ) {
               pair[i].f = 1 ;
               nrej++ ;
            }
            else {
               dmsum += dm ;
               nn++ ;
            }
         }
      }
      *dmave = dmsum / (double) MAX(nn, 1) ;
      cp_message(0,5,"iter %d  nrej %d\n",iter,nrej);
      iter++ ;
   }
   cp_message(0,4,"Outcoming mean: %g\n",*dmave) ;
   return ;
}

static double
get_dmag(Star **s1, Star **s2, Match *match, int nmatch)
{
   double ret = 0.0 ;
   int    i, n ;
   n = 0 ;
   for (i=0; i<nmatch; i++) {
      if ( s1[match[i].i]->mag != IntNull &&
           s2[match[i].j]->mag != IntNull ) {
         ret += s1[match[i].i]->mag - s2[match[i].j]->mag ;
         n++ ;
      }
   }
   ret /= (double) MAX(n, 1) ;
   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : model_fit_distortion
 * In        : Stars to fit, index table for matching stars, 
 *             number of matches and plate model structure
 * Out       : updated plate model structure, reduced chi-square
 * Purpouse  : fit the matched list and reference stars with a polynomial
 *             to find the plate solution of the field.
 *             
 * Note      : Fit is made after PCi_j linear tramsformation and before 
 *             CDELTi scaling
 *---------------------------------------------------------------------------*/

int 
model_fit_distortion (StarMatch *match, int nmatch, 
                      PlateMod *model, double *chisq)
{

   int      i, ncoeff ;
   int      status = 1 ;
   double  *x_1 = (double*)NULL ;
   double  *y_1 = (double*)NULL ;
   double  *x_2 = (double*)NULL ;
   double  *y_2 = (double*)NULL ;
   double  *x_s = (double*)NULL ;
   double  *y_s = (double*)NULL ;
   double   x, y, xchisq, ychisq ;

   if (model->cdelt[0] == 0.0 || model->cdelt[1] == 0.0)
      return 2 ;

   /* Allocate memory for working arrays */
   if (!(x_1 = (double*) malloc ((size_t)nmatch * sizeof (double))) ||
       !(y_1 = (double*) malloc ((size_t)nmatch * sizeof (double))) ||
       !(x_s = (double*) malloc ((size_t)nmatch * sizeof (double))) ||
       !(y_s = (double*) malloc ((size_t)nmatch * sizeof (double))) ||
       !(x_2 = (double*) malloc ((size_t)nmatch * sizeof (double))) ||
       !(y_2 = (double*) malloc ((size_t)nmatch * sizeof (double))) ){
      cp_error ("Could not alloc memory for finding plate solution");
      goto exit ;
   }

   /* Set relative pixel coordinates and individual standard deviations */
   for (i = 0; i < nmatch; i++) {

      /* Scale reference star list to 2nd star list units */
      x_1[i] = match[i].s1->xrot / model->cdelt[0] ;
      y_1[i] = match[i].s1->yrot / model->cdelt[1] ;

      /* Apply PCj_i matrix to 2nd star list */
      x = match[i].s2->xprj ;
      y = match[i].s2->yprj ;
      x_2[i] = model->pc[0][0]*x + model->pc[0][1]*y ;
      y_2[i] = model->pc[1][0]*x + model->pc[1][1]*y ;

      x_s[i] = model->dxdev / model->cdelt[0] ;
      y_s[i] = model->dydev / model->cdelt[1] ;
   }
   IF_ABORT_EXIT ;

   /* Fit X and Y model coefficents */
   (void) fit_module (x_2, y_2, x_1, x_s, nmatch, model->xfit_coeff, model->mask,
         (int)FN_MAXPAR, &xchisq, model->set.PlateFunc) ;

   IF_ABORT_EXIT ;

   (void) fit_module (x_2, y_2, y_1, y_s, nmatch, model->yfit_coeff, model->mask,
         (int)FN_MAXPAR, &ychisq, model->set.PlateFunc) ;

   IF_ABORT_EXIT ;

   /* Get reduced chi-2 */
   for (ncoeff = i = 0; i < (int)FN_MAXPAR; i++)
      if (model->mask[i] == 1) ncoeff++ ;

   /* Fill plate model structure */
   /* ***FIXME***/
   model->nx_coeff = model->ny_coeff = ncoeff ;
   /* model->nx_coeff = MAX(ncoeff, model->nx_coeff) ;
   model->ny_coeff = MAX(ncoeff, model->ny_coeff) ; */
   model->xchisq = xchisq / (double) MAX (nmatch-ncoeff, 1) ;
   model->ychisq = ychisq / (double) MAX (nmatch-ncoeff, 1) ;
   model->nmatch = nmatch ;

   *chisq = (model->xchisq + model->ychisq) / 2.0 ;

   status = 0 ;

   exit:

   /* Free memory */
   if (x_1 != (double*)NULL) free (x_1) ;
   if (y_1 != (double*)NULL) free (y_1) ;
   if (x_s != (double*)NULL) free (x_s) ;
   if (y_s != (double*)NULL) free (y_s) ;
   if (x_2 != (double*)NULL) free (x_2) ;
   if (y_2 != (double*)NULL) free (y_2) ;

   return (status) ;
}

