/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include "cp_messages.h"
#include "fitplate.h"

#include <math.h>
#include <gsl/gsl_multifit.h>

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/


/*-----------------------------------------------------------------------------
 * Function  : gslfit
 * In        : x1[], x2[], y[],   ... fit y[] as function of x1[],x2[]
 *             sig[],             ... the error scale lenght of each data point
 *             ndat,              ... the number of fitting points
 *             fc[],              ... the fitting coefficents
 *             ia[],              ... flag vector: 1 to fit the corresponding
 *                                    coefficent, 2 to subtract off dependences 
 *                                    on known pieces of fitting function
 *             ma,                ... the number of fitting coefficents
 *             funcs,             ... the fitting function
 * Out       : fc[],              ... the fitted coefficents
 *             *chisq             ... the resulting chi-square
 * Purpouse  : Fit a bidimensional function using a chi-2 minimization
 *             Fitting equations are solved using singular value decomposition
 * Note      : 
 *---------------------------------------------------------------------------*/


void 
gslfit(double x1[], double x2[], double yy[], double sig[], int ndata, 
       double fc[], int ia[], int ma, double *chisq,
       void (*funcs)(double, double, double [], int))
{
   int           i, j, k, mfit ;
   double       *afunc, ym ;
   gsl_matrix   *X, *cov ;
   gsl_vector   *y, *w, *c ;
   gsl_multifit_linear_workspace * work ;

   /* Get number of parameters to fit */
   for ( mfit=j=0; j<ma; j++ )
      if ( ia[j] == 1 ) mfit++ ;
   if ( mfit == 0 ) cp_error("lfit: no parameters to be fitted");

   /* Allocate memory for working arrays */
   X = gsl_matrix_alloc ((size_t)ndata, (size_t)mfit) ;
   y = gsl_vector_alloc ((size_t)ndata) ;
   w = gsl_vector_alloc ((size_t)ndata) ;
   c = gsl_vector_alloc ((size_t)mfit) ;
   cov = gsl_matrix_alloc ((size_t)mfit, (size_t)mfit) ;

   afunc = (double*)malloc((size_t)ma*sizeof(double)) ;

   /* Accumulate coefficents for normal equations */
   for ( i=0; i<ndata; i++ ) {
      (*funcs)(x1[i], x2[i], afunc, ma) ;
      ym = yy[i] ;
      /* Subtract off dependences on known pieces of fitting function */
      if ( mfit < ma ) 
         for ( j=0; j<ma; j++ ) 
            if ( ia[j] == 2 ) ym -= fc[j]*afunc[j] ;
      for ( k=j=0; j<ma; j++ ) 
         if ( ia[j] == 1 ) 
            gsl_matrix_set (X, (size_t)i, (size_t)k++, afunc[j]) ;
      gsl_vector_set (y, (size_t)i, ym) ;
      gsl_vector_set (w, (size_t)i, 1.0/(sig[i]*sig[i])) ;
   }

   work = gsl_multifit_linear_alloc ((size_t)ndata, (size_t)mfit) ;
   gsl_multifit_wlinear (X, w, y, c, cov, chisq, work) ;
   gsl_multifit_linear_free (work) ;

   /* Set output array */
   for ( k=j=0; j<ma; j++ ) 
      if ( ia[j] == 1 ) 
         fc[j] = gsl_vector_get (c, (size_t)k++) ;

   /* Free memory */
   gsl_matrix_free (X) ;
   gsl_vector_free (y) ;
   gsl_vector_free (w) ;
   gsl_vector_free (c) ;
   gsl_matrix_free (cov) ;
   free(afunc) ;

   return ;
}
