/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * Module	: logic.c
 * Author	: Paolo Montegriffo, Bologna Astronomical Observatory
 * In 		: character string, logic arithmetic type
 * Out 		: item value
 * Job		: resolves an arithmetic logical expression in polish 
 *		  notation
 * Notice	: This set of routines has been derived from parser.c 
 *		  module written by Nicolas Devillard for ESO Eclipse 
 *		  software: many of the function anyway have been deeply 
 *		  modified by the author.
 * Version      : 2.0 - October 8, 2000
 *                2.1.2 - September 4, 2001
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <ctype.h>
#include <math.h>
#include <regex.h>

#include "catalib.h"
#include "cp_messages.h"
#include "parser.h"
#include "vector_arith.h"

/*----------------------------------------------------------------------------
 * 								Macros
 *--------------------------------------------------------------------------*/

#ifndef min
#define min(a,b) (((a)<=(b)) ? (a) : (b))
#endif
#ifndef max
#define max(a,b) (((a)>=(b)) ? (a) : (b))
#endif

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

item 	         *ParseLogicalExpression(char *expr) ;

static char      *tokenizeLogicalExpression(char *arexp) ;
static int        priority(int op) ;
static itemType   gettokentype(char *token) ;
static item      *applyOperator(item *first, item *second, int operator) ;
static item      *applyUnary( item *first, int op ) ;
static OneVector *l_op_vector(OneVector *vector1, OneVector *vector2, int operation) ;

static int        vector_and( OneVector *c1, OneVector *c2 ) ;
static int        vector_or( OneVector *c1, OneVector *c2 ) ;
static int        vector_not( OneVector *c1 ) ;
static OneVector *load_logical(char *file_number) ;
static OneVector *load_computed_logical(char *expr) ;

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern CoincId  *coinc ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * Function	:	tokenizeLogicalExpression()
 * In 		:	character string containing a standard logical
 * 			expression to tokenize
 * Out 		:	newly allocated character string, containing the
 * 			tokenized standard logical expression
 * Job		:	tokenize a standard logical expression, i.e.
 * 			recognize argument types, and separate all of them by
 * 			blanks. It becomes then easy to cut the expression into
 * 			tokens by looking for blanks as separators.
 * Notice	:	returned string has to be freed by caller
 *--------------------------------------------------------------------------*/

static char *tokenizeLogicalExpression(char *arexp)
{
   int     sz ;
   int     i ;
   char *  ret ;
   char *  cur ;
   char *  truncexp ;
   int     beg ;
   boolean matched ;
   regex_t re_isin, re_numb, re_op, re_par ;

   /*
    *  Let's make use of regexp
    *  we define four possible objects to be recognized:
    *  functio isin, numbers, logical operators, parentheses
    */

   char isin[] = "^(IN[0-9]+(_\\[[0-9]+(([-]|,)[0-9]+)+(,[0-9]+([-][0-9]+)?)*\\])?)$" ;
   char numbers[] = "^([0-9]*)$" ;
   char op[] = "^[^(\\!=)$|^([|]\\{2\\})$|^&\\{2\\}$]$" ;
   char par[] = "^[\\(\\)]$" ;

   /* Compile rule for isin */
   if (regcomp(&re_isin, &isin[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for isin function: aborting") ;
      return (char*)NULL ;
   }
   /* Compile rule for number */
   if (regcomp(&re_numb, &numbers[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for number: aborting") ;
      return (char*)NULL ;
   }
   /* Compile rule for operator */
   if (regcomp(&re_op, &op[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for operator: aborting") ;
      return (char*)NULL ;
   }
   /* Compile rule for parenthese */
   if (regcomp(&re_par, &par[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for parenthese: aborting") ;
      return (char*)NULL ;
   }

   /* Strip first white spaces in expression before starting */
   strip_blanks(arexp) ;

   /* allocate all work strings	*/
   sz = (int)strlen(arexp) ;
   if ((ret = (char*)calloc((size_t)(2*sz+1), sizeof(char))) == (char*)NULL) {
      return(char*)NULL ;
   }
   if ((cur = (char*)calloc((size_t)(sz+1), sizeof(char))) == (char*)NULL) {
      free(ret) ;
      return(char*)NULL ;
   }
   if ((truncexp = (char*)calloc((size_t)(sz+1), sizeof(char))) == (char*)NULL) {
      free(ret) ;
      free(cur) ;
      return(char*)NULL ;
   }

   /*
    * The pattern identification is done in the following way:
    * try to match the entire string. if it matches, store it
    * as an identified token, else remove the rightmost character and
    * try again to match it. 
    */

   beg = 0 ;
   while (beg < sz) {
      (void)memset((char*)truncexp, (int)0, (size_t)sz) ;
      (void)memcpy((char*)truncexp, (char*)arexp+beg, (size_t)(sz-beg+1)) ;
      matched = FALSE ;
      for (i=(int)strlen(truncexp) ; i>0 ; i--) {
         (void)memset((char*)cur, (int)0, (size_t)sz) ;
         (void)strncpy((char*)cur, (char*)truncexp, (size_t)i) ;

	     /* Try to match function isin */
	 if ((regexec(&re_isin, cur, (size_t) 0, NULL, 0) == 0) ||
	     /* Try to match a number */
	     (regexec(&re_numb, cur, (size_t) 0, NULL, 0) == 0) ||
             /* Try to match an operator */
	     (!strcmp(cur,"||")) || (!strcmp(cur,"&&")) || (!strcmp(cur,"!=")) ||
             /* (regexec(&re_op, cur, (size_t) 0, NULL, 0) == 0) || */
             /* Try to match a parenthese */
             (regexec(&re_par, cur, (size_t) 0, NULL, 0) == 0)) {

            /* One match found */
            (void)sprintf(ret, "%s %s", ret, cur) ;
            beg += (int)strlen(cur) ;
            matched = TRUE ;
            break ;
         }
      }
      if (matched == FALSE) {
         /*
	 * The whole expression has been reduced to zero characters
	 * without match found.
         */
         cp_error("unrecognized element in logical expression: %s",truncexp) ;
         exit(-1) ;
      }
   }

   /* free regexp rules	*/
   regfree(&re_isin) ;
   regfree(&re_numb) ;
   regfree(&re_op) ;
   regfree(&re_par) ;

   free(cur) ;
   free(truncexp) ;
   return(ret) ;
}

/*---------------------------------------------------------------------------
 * Function	:	ParseLogicalExpression()
 * In 		:	character string, logic arithmetic type
 * Out 		:	item value
 * Job		:	resolves an arithmetic logical expression in polish 
 *			notation
 * Notice	:	This routine has been derived from ParseExpression()
 *			function of ESO Eclipse software.
 *--------------------------------------------------------------------------*/

item *ParseLogicalExpression( char *in_expr )
{
   char  * expr ;
   char  * token ;
   item  * pushvalue ;
   item  * pop1, * pop2 ;	
   int     tokc ;
   int     count ;
   char    tokv[MAX_OP][NAME_SIZE] ; 
   stack   opStack, polishStack, finalStack ;
   stack * mov ;
   boolean SolvedOp = FALSE ;
   item  * cur ;
   itemType toktype ;

   /*
    * Tokenize expression into single items
    */
   cp_message(0,3,"Tokenize the expression %s\n",in_expr);
   expr = tokenizeLogicalExpression(in_expr) ;
   cp_message(0,3,"Tokenized expression: %s\n",expr);

   /* 
    * Get first all the tokens into an array:
    * this is due to a bug in strtok()
    */
   tokc = 0 ;
   token = strtok(expr, SEPARATOR) ;
   while (token != NULL) {
      (void)strcpy(tokv[tokc], token) ;
      tokc++ ;
      token = strtok(NULL, SEPARATOR) ;
   }

   /* Initialize all stack pointers */
   opStack.next = polishStack.next = finalStack.next = NullStack ;
   opStack.data = polishStack.data = finalStack.data = NullItem ;

   /*
    * the input arithmetic expression is a standard arithmetic logical 
    * expression which has already been parsed to separate
    * arguments by spaces.
    * What is done here is: parse the arithmetic elements one by
    * one and accumulate them on a stack to be solved by the polish
    * stack solver.
    */
   for (count=0 ; count<tokc ; count++) {
      token = tokv[count] ;

      switch(toktype=gettokentype(token)) {

         case number:
	 pushvalue = (item*)malloc(sizeof(item)) ;
	 if (pushvalue == NullItem) {
	    cp_error("cannot allocate item: aborting") ;
	    return(NullItem) ;
	 }
	 pushvalue->t = vector ;
	 pushvalue->v.vector = load_logical(token) ;
	 if (pushvalue->v.vector == NullVector) {
	    cp_error("cannot load %s: aborting", token) ;
	    return(NullItem) ;
	 }
	 push(pushvalue, &polishStack) ;
	 break ;

         case computed_vector:
	 pushvalue = (item*)malloc(sizeof(item)) ;
	 if (pushvalue == NullItem) {
	    cp_error("cannot allocate item: aborting") ;
	    return(NullItem) ;
	 }
	 pushvalue->t = vector ;
	 pushvalue->v.vector = load_computed_logical(token) ;
	 if (pushvalue->v.vector == NullVector) {
	    cp_error("cannot load %s: aborting", token) ;
	    return(NullItem) ;
	 }
	 push(pushvalue, &polishStack) ;
	 break ;

	 case left_bracket:
	 /* Left brackets are just pushed on the operator stack */
	 pushvalue = (item*)malloc(sizeof(item)) ;
	 if (pushvalue == NullItem) {
	    cp_error("cannot allocate item: aborting") ;
	    return(NullItem) ;
	 }
	 pushvalue->t = left_bracket ;
	 (void)sscanf(token, "%c", &(pushvalue->v.op)) ;
	 push(pushvalue, &opStack) ;
	 break ;

	 case right_bracket:
	 /* 
	 * Pop out everything out of opStack and push it on the
	 * polish stack, until a left bracket is found
	 */
	 cur = pop(&opStack) ;
	 while (cur->t != left_bracket) {
	    push(cur, &polishStack) ;
	    cur = pop(&opStack) ;
	    if (cur == NullItem) {
	       cp_error("error in parsing logical expression") ;
	       cp_error("missing parenthese") ;
	       exit(-1) ;
	    }	
	 }
	 free(cur) ;
	 break ;

	 case operator:
	 /*
	  * Logical operators
	  */
	 SolvedOp = FALSE ;
	 while (SolvedOp != TRUE) {
	    cur = topstack(&opStack) ;
	    if (cur == NullItem) {
	       pushvalue = (item*)malloc(sizeof(item)) ;
	       if (pushvalue == NullItem) {
		  cp_error("cannot allocate item: aborting") ;
		  return(NullItem) ;
	       }
	       pushvalue->t = toktype ;
	       (void)sscanf(token, "%c", &(pushvalue->v.op)) ;
	       push(pushvalue, &opStack) ;
	       SolvedOp = TRUE ;
	    } else if (cur->t == left_bracket) {
	       pushvalue = (item*)malloc(sizeof(item)) ;
	       if (pushvalue == NullItem) {
		  cp_error("cannot allocate item: aborting") ;
		  return(NullItem) ;
	       }
	       pushvalue->t = toktype ;
	       (void)sscanf(token, "%c", &(pushvalue->v.op)) ;
	       push(pushvalue, &opStack) ;
	       SolvedOp = TRUE ;
	       free(cur) ;
	    } else if (cur->t == operator || cur->t == unary) {
	       if (priority(cur->v.op) >= priority(token[0])) {
		  push(pop(&opStack), &polishStack) ;
		  free(cur) ;
	       } else {
		  pushvalue = (item*)malloc(sizeof(item)) ;
		  if (pushvalue == NullItem) {
		     cp_error("cannot allocate item: aborting") ;
		     return(NullItem) ;
		  }
		  pushvalue->t = toktype ;
		  (void)sscanf(token, "%c", &(pushvalue->v.op)) ;
		  push(pushvalue, &opStack) ;
		  SolvedOp = TRUE ;
		  free(cur) ;
	       }
	    } else {
	       cp_error("error in operator stack: aborting") ;
	       exit(-1) ;
	    }	
	 }
	 break ;

         case unary:
	 /* 
	 * Push unary operator on the  opStack
	 */
      	 pushvalue = (item*)malloc(sizeof(item)) ;
	 if (pushvalue == NullItem) {
	    cp_error("cannot allocate item: aborting") ;
	    return(NullItem) ;
	 }
	 pushvalue->t = unary ;
	 (void)sscanf(token, "%c", &(pushvalue->v.op)) ;
	 push(pushvalue, &opStack) ;
         break ;

	 case unknown:
	 default:
	 cp_error("in logical expression: exiting") ;
	 exit(-1) ;
      }
   }

   /* Now pop every remaining operator on opStack onto polishStack */
   while (opStack.next != NullStack) push(pop(&opStack), &polishStack) ;	
   
   /* POLISH STACK SOLVING */
   cp_message(0,3,"Polish stack solving\n");
   mov = polishStack.next ;
   while (mov != NullStack) {
      switch(mov->data->t) {

	 /* Simple operands are pushed onto the stack	*/
	 case vector:
	 cp_message(0,4,"Pushing Simple operands onto the stack\n");
	 pushvalue = (item*)malloc(sizeof(item)) ;
	 if (pushvalue == NullItem) {
	    cp_error("cannot allocate item: aborting") ;
	    return(NullItem) ;
	 }
	 pushvalue->t = mov->data->t ;
	 pushvalue->v = mov->data->v ;
	 push(pushvalue, &finalStack) ;
	 break ;

	 /* 
	 * Operator: pop the two operands on top of the stack, apply
	 * the operator to them, and push the result onto the stack
	 */
	 case operator:
	 cp_message(0,4,"Pop two operands on top of the stack, push the result onto the stack\n");
	 pop2 = pop(&finalStack) ;
	 pop1 = pop(&finalStack) ;
	 pushvalue = applyOperator(pop1, pop2, mov->data->v.op) ;
	 free(pop1) ;
	 free(pop2) ;
	 if (pushvalue == NullItem) {
	    cp_error("logical error occurred: aborting") ;
	    return(NullItem) ;
	 }
	 push(pushvalue, &finalStack) ;
	 break ;

	 /* 
	 * Unary operator: pop the operand on top of the stack, apply
	 * the function to it, and push the result onto the stack
	 */
	 case unary:
	 cp_message(0,4,"Pop one operand on top of the stack and invert it\n");
	 pop1 = pop(&finalStack) ;
	 pushvalue = applyUnary(pop1, mov->data->v.op) ;
	 free(pop1) ;
	 if (pushvalue == NullItem) {
	    cp_error("logical error occurred: aborting") ;
	    return(NullItem) ;
	 }
	 push(pushvalue, &finalStack) ;
	 break ;

	 default:
	 cp_error("invalid item in polish logical expression") ;
	 return(NullItem) ;	
	 break ;
      }
      mov = mov->next ;
   }

   /* free polishStack */
   while (polishStack.next != NullStack) {
      pop1 = pop(&polishStack) ;
      free(pop1) ;
   }
   /* free tokenized expression */
   free(expr) ;

   /* The final result is the remaining operand on top of the stack */
   return(pop(&finalStack)) ;
}

/*---------------------------------------------------------------------------
 * Function	:	gettokentype()
 * In 		:	token (character string)
 * Out 		:	symbol: OPERAND or OPERATOR
 * Job		:	find out if a token is a number or an operator
 * Notice	:
 *--------------------------------------------------------------------------*/

static itemType gettokentype( char *token ) 
{
   regex_t re_expr ;
   char    numb[] = "^([0-9]*)$" ;

   if (token == (char*)NULL) {
	   cp_error("internal: cannot get token") ;
	   return unknown ;
   }

   if (token[0] == '(') {
      cp_message(0,4,"matched a left bracket\n");
      return left_bracket ;
   }

   if (token[0] == ')') {
      cp_message(0,4,"matched a right bracket\n");
      return right_bracket ;
   }

   if (token[0] == 'i' || token[0] == 'I') {
      cp_message(0,4,"matched a IsIn function\n");
      return computed_vector ;
   }

   /* try to match a number: */
   if (regcomp(&re_expr, &numb[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for numbers: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      cp_message(0,4,"matched a number\n");
      regfree(&re_expr) ;
      return number;
   }

   /* try to match an operator */
   if (!strcmp(token,"&&") || !strcmp(token,"||")){
      cp_message(0,4,"matched a binary logical operator\n") ;
      return operator ;
   }

   if (!strcmp(token,"!=")) {
      cp_message(0,4,"matched an unary logical operator\n") ;
      return unary ;
   }

   cp_error("unrecognized token: [%s]", token) ;
   return unknown ;
}

/*---------------------------------------------------------------------------
 * Function	:	applyOperator()
 * In 		:	two items and an operator
 * Out 		:	one item
 * Job		:	operates logical between int vectors
 * Notice	:
 *--------------------------------------------------------------------------*/

static item *applyOperator( item *first, item *second, int op )
{
   item      * out ;

   cp_message(0,4,"Applying operator %c\n", op);

   out = (item*)malloc(sizeof(item)) ;
   if (out == NullItem) {
      cp_error("memory allocation failure for operator: aborting") ;
      return(NullItem) ;
   }

   if ((first->t == vector) && (second->t == vector)) {
      /* Two Vectors */
      out->t = vector ;
      if (op!='|' && op!='&') {
	 cp_error("logical operation %c is invalid between vectors", op);
	 free(out) ;
	 return(NullItem) ;
      }
      out->v.vector = l_op_vector(first->v.vector, second->v.vector, op) ;
      destroy_vector(first->v.vector) ;
      destroy_vector(second->v.vector) ;
      if (out->v.vector == NullVector) {
	 cp_error("error during vector arithmetic: aborting") ;
	 free(out) ;
	 return(NullItem) ;
      }
   } else {
      cp_error("type identification error in logical expression: aborting") ;
      free(out) ;
      return(NullItem) ;
   }
   return(out) ;
}

/*---------------------------------------------------------------------------
 * Function	:	applyOperator()
 * In 		:	two items and an operator
 * Out 		:	one item
 * Job		:	operates logical between int vectors
 * Notice	:
 *--------------------------------------------------------------------------*/

static item *applyUnary( item *first, int op )
{
   item      * out ;

   cp_message(0,4,"Applying unary operator %c\n", op);

   if (op!='!') {
      cp_error("logical operation %c is invalid ", op);
      return(NullItem) ;
   }

   out = (item*)malloc(sizeof(item)) ;
   if (out == NullItem) {
      cp_error("memory allocation failure for operator: aborting") ;
      return(NullItem) ;
   }
   out->v.vector = copy_vector(first->v.vector) ;
   if (out == NullItem) {
      cp_error("memory allocation failure for operator: aborting") ;
      return(NullItem) ;
   }
   vector_not(out->v.vector) ;
   destroy_vector(first->v.vector) ;
   if (out->v.vector == NullVector) {
      cp_error("error during vector logical: aborting") ;
      free(out) ;
      return(NullItem) ;
   }

   return(out) ;
}

/*----------------------------------------------------------------------------
 * Function	:	l_op_vector()
 * In 		:	2 vectors, logical operation to perform
 * Out 		:	newly allocated vector
 * Job		:	logical operations between 2 vectors
 * Notice	:	returns NULL in case of error
 *--------------------------------------------------------------------------*/

static OneVector *l_op_vector(OneVector *vector1, OneVector *vector2, int operation) 
{
   OneVector * out_vector ;

   out_vector = copy_vector(vector1) ;
   cp_message(0, 3, "performing vector logical comparison %c\n",operation) ;

   switch(operation) {

      case '&':
      vector_and(out_vector, vector2) ;
      out_vector->width = max(out_vector->width, vector2->width );
      out_vector->ndec = max(out_vector->ndec, vector2->ndec );
      if(vector2->type == 'f') out_vector->type = 'f';
      break ;

      case '|':
      vector_or(out_vector, vector2) ;
      out_vector->width = max(out_vector->width, vector2->width );
      out_vector->ndec = max(out_vector->ndec, vector2->ndec );
      if(vector2->type == 'f') out_vector->type = 'f';
      break ;

      default:
      cp_error("illegal requested operation: aborting logical analysys") ;
      destroy_vector(out_vector) ;
      out_vector = NullVector ;
   }

   return out_vector ;
}

/*----------------------------------------------------------------------------
 * Function	:	vector_and()
 * In 		:	two logicalvectors
 * Out 		:	0 if Ok, -1 if error	
 * Job		:	out vec is 1 if inp vec are 1 & 1; 0 otherwise
 * Notice	:	the first vector is modified: it contains the result of
 * 			logical comparison between the vectors.
 *--------------------------------------------------------------------------*/

static int vector_and( OneVector *c1, OneVector *c2 )
{
   int i ;

   if (c1->size != c2->size) {
      cp_error("vectors have different dimension: cannot add") ;
      return -1 ;
   }
   for (i=0 ; i< c1->size ; i++) {
      if( c1->data[i]==1 && c2->data[i]==1 )
         ;
      else
         c1->data[i] = 0 ;
   }

   return 0 ;
}

/*----------------------------------------------------------------------------
 * Function	:	vector_or()
 * In 		:	two vectors
 * Out 		:	0 if Ok, -1 if error	
 * Job		:	out vec is 1 if almost 1 inp vec is 1
 * Notice	:	the first vector is modified: it contains the result of
 * 			logical comparison between the vectors.
 *--------------------------------------------------------------------------*/

static int vector_or( OneVector *c1, OneVector *c2 )
{
   int i ;

   if (c1->size != c2->size) {
      cp_error("vectors have different dimension: cannot add") ;
      return -1 ;
   }
   for (i=0 ; i< c1->size ; i++) {
      if( c1->data[i]==1 || c2->data[i]==1 )
         c1->data[i] = 1 ;
      else
         c1->data[i] = 0 ;
   }

   return 0 ;
}

/*----------------------------------------------------------------------------
 * Function	:	vector_not()
 * In 		:	one vector
 * Out 		:	0 if Ok, -1 if error	
 * Job		:	out vec is 1 if inp vec is 0 and viceversa
 * Notice	:	the vector is modified: it contains the result of
 * 			logical computation.
 *--------------------------------------------------------------------------*/

static int vector_not( OneVector *c1 )
{
   int i ;

   for (i=0 ; i< c1->size ; i++) {
      if( c1->data[i]==1 )
         c1->data[i] = 0 ;
      else
         c1->data[i] = 1 ;
   }

   return 0 ;
}

/*---------------------------------------------------------------------------
 * Function :  priority()
 * In       :  1 char from: | & !
 * Out      :  int
 * Job      :  returns a priority for a given operation
 *             the bigger the returned integer, the higher the priority
 * Notice   :
 *--------------------------------------------------------------------------*/

static int priority( int op )
{
   int p ;

   switch(op) {
      /* or has the lowest priority, not the higher */
      case '|':
      p = 1 ;
      break ;

      case '&':
      p = 2 ;
      break ;

      case '!':
      p = 3 ;
      break ;

      /* default: lowest possible priority (should not happen) */
      default:
      p = 0 ;
      break ;
   }

   return p ;
}	

/*---------------------------------------------------------------------------
 * Function :   load_logical()
 * In       :   file number to test
 * Out      :   Pointer to newly allocated vector
 * Job      :   
 * Notice   :
 *--------------------------------------------------------------------------*/

static OneVector *load_logical( char *file_number)
{
   OneVector *loaded ;
   int n = (int) atoi(file_number)-1 ;
   int        j ;

   /* Allocate memory for returning vector */
   loaded = new_vector(coinc->nrec);
   if (loaded == NullVector) {
      cp_error("cannot allocate output vector: aborting extraction") ;
      return(NullVector) ;
   }

   /* Fill returning vector */

   for(j=0; j<coinc->nrec; j++) {
      loaded->data[j] = (double)(bin_check_n(coinc->ns[j],n)+bin_check_n(coinc->ns2[j],n-32)) ;
      /*
      cp_error ("%u:%u  n %u : data %d  === %d %d",coinc->ns[j],coinc->ns2[j], n, loaded->data[j] , bin_check_n(coinc->ns[j],n),bin_check_n(coinc->ns2[j],n-32)) ;
      */
      loaded->is_in[j] = 1 ;
   }
   loaded->width = 3 ;
   loaded->ndec  = 0 ;
   loaded->type  = 'd' ;

   /*
   cp_error("3 0 : %d", bin_check_n(3,0));
   cp_error("3 1 : %d", bin_check_n(3,1));
   cp_error("3 2 : %d", bin_check_n(3,2));
   cp_error("2 1 : %d", bin_check_n(2,1));
   cp_error("2 2 : %d", bin_check_n(2,2));
   cp_error("4294967295 1 : %d", bin_check_n(4294967295,1));
   cp_error("4294967295 31 : %d", bin_check_n(4294967295,31));
   cp_error("4294967295 32 : %d", bin_check_n(4294967295,32));
*/
   
   return loaded ;
}

/*---------------------------------------------------------------------------
 * Function :   load_computed_logical()
 * In       :   function isin to test for
 * Out      :   Pointer to newly allocated vector
 * Job      :   
 * Notice   :
 *--------------------------------------------------------------------------*/

static OneVector *load_computed_logical(char *expr)
{
   OneVector *loaded ;
   OneVector *tmp ;
   char      *ptr, cn[4] ;
   int        i, j, n, nn, *scan = (int*)NULL ;

   /* Read min number of measures requested */
   (void)sscanf(expr+2, "%d", &n) ;
   if(n>coinc->nfiles) {
      cp_error("argument of logical operator > number of files: aborting") ;
      return(NullVector) ;
   }

   /* Look for files specification */
   if((ptr=index(expr,'_')) == (char*)NULL) {

      /* All files must be searched */
      nn = coinc->nfiles ;
      scan = (int*) calloc((size_t)nn, sizeof(int)) ;
      for(j=0; j<nn; j++)
         scan[j] = j+1 ;
   }
   else {
      /* Searches in a subsample: get files numbers */
      scan = getNumbers(ptr+2,&nn) ;
      if (scan == (int*)NULL) {
         cp_error("illegal range in expression: aborting") ;
         return(NullVector) ;
      }
      for(i=0; i<nn; i++) {
         if(scan[i]>coinc->nfiles ) {
            cp_error("argument of logical operator > number of files: aborting") ;
            return(NullVector) ;
         }
      }
    }

   /* Allocate memory for returning vector */
   loaded = new_vector(coinc->nrec);
   if (loaded == NullVector) {
      cp_error("cannot allocate output vector: aborting extraction") ;
      free(scan) ;
      return(NullVector) ;
   }

   /* Scan logical vector for each file in list */
   for(j=0; j<nn; j++) {
      (void)sprintf(cn,"%d",scan[j]) ;
      tmp = load_logical(cn) ;
      if (tmp == NullVector) {
	 cp_error("problems during execution: aborting") ;
         free(scan) ;
         destroy_vector(loaded) ;
	 return(NullVector) ;
      }
      vector_add( loaded, tmp ) ;
      destroy_vector(tmp) ;
   }

   /* Fill returning vector */
   for(j=0; j<coinc->nrec; j++) {
      if(loaded->data[j]>=n) {
         loaded->data[j] = 1 ;
         loaded->is_in[j] = 1 ;
      }
      else {
         loaded->data[j] = 0 ;
         loaded->is_in[j] = 0 ;
      }
   }
   loaded->width = 3 ;
   loaded->ndec  = 0 ;
   loaded->type  = 'd' ;

   return loaded ;
}

