/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: parser.c
 * Author 	: Paolo Montegriffo
 * Created on	: July, 2000	
 * Language	: ANSI C
 * Description	: 
 * Version      : 2.0 - October 1, 2000
 *                2.1.2 - September 4, 2001
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <ctype.h>
#include <math.h>
#include <regex.h>
#include <float.h>

#include "parser.h"
#include "catalib.h"
#include "cp_messages.h"
#include "vector_arith.h"

/*----------------------------------------------------------------------------
 * 								Macros
 *--------------------------------------------------------------------------*/

#ifndef min
#define min(a,b) (((a)<=(b)) ? (a) : (b))
#endif
#ifndef max
#define max(a,b) (((a)>=(b)) ? (a) : (b))
#endif

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#define EXPMAX log(DBL_MAX)
#define EXPMIN log(DBL_MIN)

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static int      priority(int op) ;

static itemType gettokentype(char *token) ;
static item   * applyOperator(item *first, item *second, int operator) ;
static item   * applyUnary(item * first, opType op) ;

static OneNumber get_number(char *token) ;
static opType    get_function(char *token) ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
 * Function :  tokenizeExpression()
 * In       :  character string containing a standard arithmetic
 *             expression to tokenize
 * Out      :  newly allocated character string, containing the
 *             tokenized standard arithmetic expression
 * Job      :  tokenize a standard arithmetic expression, i.e.
 *             recognize argument types, and separate all of them by
 *             blanks. It becomes then easy to cut the expression into
 *             tokens by looking for blanks as separators.
 * Notice   :  returned string has to be freed by caller
 * Example  :  tok_exp = tokenizeExpression("2*(@obj-@sky)/ff.fits")
 *             tok_exp contains then: "2 * ( @obj - @sky ) / ff.fits"
 *--------------------------------------------------------------------------*/

char *
tokenizeExpression(char *arexp)
{
   int     sz ;
   int     i ;
   char *  ret ;
   char *  cur ;
   char *  truncexp ;
   int     beg ;
   boolean matched ;
   regex_t re_lst, re_fns, re_fnc, re_col, re_numb, re_op, re_par ;

   /*
    *  Let's make use of regexp
    *  we define seven possible objects to be recognized:
    *  vectors list, special functions, unary operators,  
    *  columns, numbers, operators, parentheses
    */

   char lista[] = "^C[0-9]+_[0-9]+(,C[0-9]+_[0-9]+)+$" ;
   char funct[] = "^(AVE|WAVE|STDEV|STERR|WSTDEV|WSTERR)\\([^)]*\\)$" ;
   char unars[] = "^(LOG|LOG2|LOG10|EXP|SQRT|ABS)$" ;
   char columns[] = "^C[0-9]+_[0-9]+$|^(ID|NS)$" ;
   char numbers[] = "^([0-9]+[.]?[0-9]*|[.][0-9]+)([eE][+-]?[0-9]+)?$" ;
   char op[] = "^[+-/*\\^le]$" ;
   char par[] = "^[\\(\\)]$" ;

   /* Compile rule for vectors list */
   if (regcomp(&re_lst, &lista[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for ...: aborting") ;
      return (char*)NULL ;
   }
   /* Compile rule for special functions */
   if (regcomp(&re_fns, &funct[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for special functions: aborting") ;
      return (char*)NULL ;
   }
   /* Compile rule for unary operators */
   if (regcomp(&re_fnc, &unars[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for unary operators: aborting") ;
      return (char*)NULL ;
   }
   /* Compile rule for columns */
   if (regcomp(&re_col, &columns[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for columns: aborting") ;
      return (char*)NULL ;
   }
   /* Compile rule for number */
   if (regcomp(&re_numb, &numbers[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for number: aborting") ;
      return (char*)NULL ;
   }
   /* Compile rule for operator */
   if (regcomp(&re_op, &op[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for operator: aborting") ;
      return (char*)NULL ;
   }
   /* Compile rule for parenthese */
   if (regcomp(&re_par, &par[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for parenthese: aborting") ;
      return (char*)NULL ;
   }

   /* Strip first white spaces in expression before starting */
   strip_blanks(arexp) ;

   /* allocate all work strings	*/
   sz = (int)strlen(arexp) ;
   if ((ret = (char*)calloc((size_t)(2*sz+1), sizeof(char))) == (char*)NULL) {
      return(char*)NULL ;
   }
   if ((cur = (char*)calloc((size_t)(sz+1), sizeof(char))) == (char*)NULL) {
      free(ret) ;
      return(char*)NULL ;
   }
   if ((truncexp = (char*)calloc((size_t)(sz+1), sizeof(char))) == (char*)NULL) {
      free(ret) ;
      free(cur) ;
      return(char*)NULL ;
   }

   /*
    * The pattern identification is done in the following way:
    * try to match the entire string. if it matches, store it
    * as an identified token, else remove the rightmost character and
    * try again to match it. 
    */

   beg = 0 ;
   while (beg < sz) {
      (void)memset((char*)truncexp, (int)0, (size_t)sz) ;
      (void)memcpy((char*)truncexp, (char*)arexp+beg, (size_t)(sz-beg+1)) ;
      matched = FALSE ;
      for (i=(int)strlen(truncexp) ; i>0 ; i--) {
         (void)memset((char*)cur, (int)0, (size_t)sz) ;
         (void)strncpy((char*)cur, (char*)truncexp, (size_t)i) ;

         if ((regexec(&re_lst, cur, (size_t) 0, NULL, 0) == 0) ||
            /* Try to match a special function */
	    (regexec(&re_fns, cur, (size_t) 0, NULL, 0) == 0) ||
            /* Try to match an unary op */
	    (regexec(&re_fnc, cur, (size_t) 0, NULL, 0) == 0) ||
	    /* Try to match a column */
	    (regexec(&re_col, cur, (size_t) 0, NULL, 0) == 0) ||
	    /* Try to match a number */
	    (regexec(&re_numb, cur, (size_t) 0, NULL, 0) == 0) ||
            /* Try to match an operator */
            (regexec(&re_op, cur, (size_t) 0, NULL, 0) == 0) ||
            /* Try to match a parenthese */
            (regexec(&re_par, cur, (size_t) 0, NULL, 0) == 0)) {

            /* One match found */

            /* 
             * Check for unary operator '-' ( change of sign ) 
             * This happens when '-' comes as 1st character or
             * just after a left bracket
             */
            if( *cur=='-' && ( strlen(ret)==(size_t)0 || 
               ( rindex(ret,'(')!=(char*)NULL && strcmp(rindex(ret,'('),"(")==0 )))
               (void)sprintf(ret, "%s %s", ret, "C_SIGN") ;
            else
               (void)sprintf(ret, "%s %s", ret, cur) ;

            beg += (int)strlen(cur) ;
            matched = TRUE ;
            break ;
         }
      }
      if (matched == FALSE) {
         /*
	 * The whole expression has been reduced to zero characters
	 * without match found.
         */
         cp_error("unrecognized element in arithmetic expression: %s",truncexp) ;
         exit(-1) ;
      }
   }

   /* free regexp rules	*/
   regfree(&re_lst) ;
   regfree(&re_fns) ;
   regfree(&re_fnc) ;
   regfree(&re_col) ;
   regfree(&re_numb) ;
   regfree(&re_op) ;
   regfree(&re_par) ;

   free(cur) ;
   free(truncexp) ;
   return(ret) ;
}

/*---------------------------------------------------------------------------
 * Function	:	ParseExpression()
 * In 		:	character string, arithmetic type
 * Out 		:	item value
 * Job		:	resolves an arithmetic expression in polish notation
 * Notice	:
 *--------------------------------------------------------------------------*/

item *
ParseExpression( char *	expr, int arithm )
{
   char  * token ;
   item  * pushvalue ;
   item  * pop1, * pop2 ;	
   int     tokc ;
   int     count ;
   char    tokv[MAX_OP][NAME_SIZE] ; 
   stack   opStack, polishStack, finalStack ;
   stack * mov ;
   boolean SolvedOp = FALSE ;
   item  * cur ;

   /* 
    * Get first all the tokens into an array:
    * this is due to a bug in strtok()
    */
   tokc = 0 ;
   token = strtok(expr, SEPARATOR) ;
   while (token != NULL) {
      (void)strcpy(tokv[tokc], token) ;
      tokc++ ;
      token = strtok(NULL, SEPARATOR) ;
   }

   /* Initialize all stack pointers */
   opStack.next = polishStack.next = finalStack.next = NullStack ;
   opStack.data = polishStack.data = finalStack.data = NullItem ;

   /* Switch according to expression type: standard or polish	*/

   if (arithm == ARITHM_STANDARD) {
      /*
       * the input arithmetic expression is a standard arithmetic
       * expression which has already been parsed to separate
       * arguments by spaces.
       * What is done here is: parse the arithmetic elements one by
       * one and accumulate them on a stack to be solved by the polish
       * stack solver.
       */
      for (count=0 ; count<tokc ; count++) {
	 token = tokv[count] ;

	 switch(gettokentype(token)) {

            case vector:
	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = vector ;
	    pushvalue->v.vector = load_vector(token) ;
            if(pushvalue->v.vector->type=='s') pushvalue->t = string ;
	    if (pushvalue->v.vector == NullVector) {
	       cp_error("cannot load %s: aborting", token) ;
	       return(NullItem) ;
	    }
	    push(pushvalue, &polishStack) ;
	    break ;

            case vector_list:
	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = vector ;
	    pushvalue->v.vector = load_vector_list(token) ;
            if(pushvalue->v.vector->type=='s') pushvalue->t = string ;
	    if (pushvalue->v.vector == NullVector) {
	       cp_error("cannot load %s: aborting", token) ;
	       return(NullItem) ;
	    }
	    push(pushvalue, &polishStack) ;
	    break ;

            case computed_vector:
	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = vector ;
	    pushvalue->v.vector = load_computed_vector(token) ;
	    if (pushvalue->v.vector == NullVector) {
	       cp_error("cannot load %s: aborting", token) ;
	       return(NullItem) ;
	    }
	    push(pushvalue, &polishStack) ;
	    break ;

	    case number:
	    /* Simple operands are just pushed on the polish stack */
	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = number ;
            pushvalue->v.f = get_number(token) ;
	    push(pushvalue, &polishStack) ;
	    break ;

	    case left_bracket:
	    /* Left brackets are just pushed on the operator stack */
	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = left_bracket ;
	    (void)sscanf(token, "%c", &(pushvalue->v.op)) ;
	    push(pushvalue, &opStack) ;
	    break ;

	    case right_bracket:
	    /* 
	    * Pop out everything out of opStack and push it on the
	    * polish stack, until a left bracket is found
	    */
	    cur = pop(&opStack) ;
	    while (cur->t != left_bracket) {
	       push(cur, &polishStack) ;
	       cur = pop(&opStack) ;
	       if (cur == NullItem) {
		  cp_error("error in parsing arithmetic expression") ;
		  cp_error("missing parenthese") ;
		  exit(-1) ;
	       }	
	    }
	    free(cur) ;
	    break ;

	    case operator:
	    /*
	     * The whole intelligence of converting a standard
	     * arithmetic expression into a polish stack goes here.
	     */
	    SolvedOp = FALSE ;
	    while (SolvedOp != TRUE) {
	       cur = topstack(&opStack) ;
	       if (cur == NullItem) {
		  pushvalue = (item*)malloc(sizeof(item)) ;
		  if (pushvalue == NullItem) {
		     cp_error("cannot allocate item: aborting") ;
		     return(NullItem) ;
		  }
		  pushvalue->t = operator ;
		  (void)sscanf(token, "%c", &(pushvalue->v.op)) ;
		  push(pushvalue, &opStack) ;
		  SolvedOp = TRUE ;
	       } else if (cur->t == left_bracket) {
		  pushvalue = (item*)malloc(sizeof(item)) ;
		  if (pushvalue == NullItem) {
		     cp_error("cannot allocate item: aborting") ;
		     return(NullItem) ;
		  }
		  pushvalue->t = operator ;
		  (void)sscanf(token, "%c", &(pushvalue->v.op)) ;
		  push(pushvalue, &opStack) ;
		  SolvedOp = TRUE ;
		  free(cur) ;
	       } else if (cur->t == operator) {
		  if (priority(cur->v.op) >= priority((char)token[0])) {
		     push(pop(&opStack), &polishStack) ;
		     free(cur) ;
		  } else {
		     pushvalue = (item*)malloc(sizeof(item)) ;
		     if (pushvalue == NullItem) {
			cp_error("cannot allocate item: aborting") ;
			return(NullItem) ;
		     }
		     pushvalue->t = operator ;
		     (void)sscanf(token, "%c", &(pushvalue->v.op)) ;
		     push(pushvalue, &opStack) ;
		     SolvedOp = TRUE ;
		     free(cur) ;
		  }
               } else if (cur->t == unary) {
                  /* Priority of unary is higher than simple operators */
                  push(pop(&opStack), &polishStack) ;
		  free(cur) ;
	       } else {
		  cp_error("error in operator stack: aborting") ;
		  exit(-1) ;
	       }	
	    }
	    break ;

            case unary:
	    /* 
	    * Push unary operator on the  opStack
	    */
      	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = unary ;
            pushvalue->v.unary = get_function(token) ;
	    push(pushvalue, &opStack) ;
            break ;

	    case unknown:
	    default:
	    cp_error("in arithmetic expression: exiting") ;
	    exit(-1) ;
	 }
      }	

      /* Now pop every remaining operator on opStack onto polishStack */
      while (opStack.next != NullStack) push(pop(&opStack), &polishStack) ;	

   } else {

      /* Polish stack processing: push first everything onto polishStack */
      for (count=0 ; count<tokc ; count++) {
	 token = tokv[count] ;
	 switch(gettokentype(token)) {

            case vector:
	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = vector ;
	    pushvalue->v.vector = load_vector(token) ;
            if(pushvalue->v.vector->type=='s') pushvalue->t = string ;
	    if (pushvalue->v.vector == NullVector) {
	       cp_error("cannot load %s: aborting", token) ;
	       return(NullItem) ;
	    }
	    push(pushvalue, &polishStack) ;
	    break ;

            case vector_list:
	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = vector ;
	    pushvalue->v.vector = load_vector_list(token) ;
            if(pushvalue->v.vector->type=='s') pushvalue->t = string ;
	    if (pushvalue->v.vector == NullVector) {
	       cp_error("cannot load %s: aborting", token) ;
	       return(NullItem) ;
	    }
	    push(pushvalue, &polishStack) ;
	    break ;

            case computed_vector:
	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = vector ;
	    pushvalue->v.vector = load_computed_vector(token) ;
	    if (pushvalue->v.vector == NullVector) {
	       cp_error("cannot load %s: aborting", token) ;
	       return(NullItem) ;
	    }
	    push(pushvalue, &polishStack) ;
	    break ;

	    case number:
	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = number ;
            pushvalue->v.f = get_number(token) ;
	    push(pushvalue, &polishStack) ;
	    break ;

	    case operator:
	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = operator ;
	    (void)sscanf(token, "%c", &(pushvalue->v.op)) ;
	    push(pushvalue, &polishStack) ;
	    break ;

            case unary:
      	    pushvalue = (item*)malloc(sizeof(item)) ;
	    if (pushvalue == NullItem) {
	       cp_error("cannot allocate item: aborting") ;
	       return(NullItem) ;
	    }
	    pushvalue->t = unary ;
            pushvalue->v.unary = get_function(token) ;
	    push(pushvalue, &polishStack) ;
            break ;

	    case unknown:
	    default:
	    cp_error("in arithmetic expression: exiting") ;
	    exit(-1) ;
         }
      }
   }
   
   /* POLISH STACK SOLVING */
   mov = polishStack.next ;
   while (mov != NullStack) {
      switch(mov->data->t) {

	 /* Simple operands are pushed onto the stack	*/
	 case vector:
	 case number:
         case string:
	 pushvalue = (item*)malloc(sizeof(item)) ;
	 if (pushvalue == NullItem) {
	    cp_error("cannot allocate item: aborting") ;
	    return(NullItem) ;
	 }
	 pushvalue->t = mov->data->t ;
	 pushvalue->v = mov->data->v ;
	 push(pushvalue, &finalStack) ;
	 break ;

	 /* 
	 * Operator: pop the two operands on top of the stack, apply
	 * the operator to them, and push the result onto the stack
	 */
	 case operator:

	 pop2 = pop(&finalStack) ;
	 pop1 = pop(&finalStack) ;
	 pushvalue = applyOperator(pop1, pop2, mov->data->v.op) ;
	 free(pop1) ;
	 free(pop2) ;
	 if (pushvalue == NullItem) {
	    cp_error("arithmetic error occurred: aborting") ;
	    return(NullItem) ;
	 }
	 push(pushvalue, &finalStack) ;
	 break ;

	 /* 
	 * Unary operator: pop the operand on top of the stack, apply
	 * the function to it, and push the result onto the stack
	 */
	 case unary:

	 pop1 = pop(&finalStack) ;
	 pushvalue = applyUnary(pop1, mov->data->v.unary) ;
	 free(pop1) ;
	 if (pushvalue == NullItem) {
	    cp_error("arithmetic error occurred: aborting") ;
	    return(NullItem) ;
	 }
	 push(pushvalue, &finalStack) ;
	 break ;

	 default:
	 cp_error("invalid item in polish arithmetic expression") ;
	 return(NullItem) ;	
	 break ;
      }
      mov = mov->next ;
   }

   /* free polishStack */
   while (polishStack.next != NullStack) {
      pop1 = pop(&polishStack) ;
      free(pop1) ;
   }	

   /* The final result is the remaining operand on top of the stack */
   return(pop(&finalStack)) ;
}

/*----------------------------------------------------------------------------
 * Function :  push()
 * In       :  value, pointer to beginning of stack
 * Out      :  void
 * Job      :  appends a value on top of the stack
 * Notice   :
 *--------------------------------------------------------------------------*/

void 
push( item * value, stack * sp)
{
   stack * new ;
   stack * mov ;

   if (sp == NullStack) {
      cp_error("unexpected NULL stack: aborting") ;
      exit(-1) ;
   }	
   if ((new = (stack*)malloc(sizeof(stack))) == NullStack) {
      cp_error("error in stack allocation: aborting") ;
      exit(-1) ;
   }

   /* Go to last stack element */
   mov = sp ;
   while (mov->next != NullStack) {
      mov = mov->next ;
   }
   /* initialize element to be appended with relevant information */
   new->next = NullStack ;
   new->data = value ;
   /* append new element at the end of stack */
   mov->next = new ;
}

/*---------------------------------------------------------------------------
 * Function :  pop()
 * In       :  stack pointer
 * Out      :  last pushed stack value
 * Job      :  gets the latest value pushed on stack
 * Notice   :  stack is modified (last element is deleted)
 *--------------------------------------------------------------------------*/

item * 
pop( stack * sp )
{
   stack * mov ;
   item  * value ;

   if (sp == NullStack) {
      cp_error("unexpected NULL stack") ;
      exit(-1) ;
   }

   /* Only one element in stack means the stack is empty */
   if (sp->next == NullStack) {
      cp_error("empty stack: syntax error in expression") ;
      exit(-1) ;
   }	

   /* find the just-before-last element (at least 2 elements in stack) */
   mov = sp ;
   while (mov->next->next != NullStack) {
      mov = mov->next ;
   }	

   /* Fill up the value to return with last element's information */
   value = mov->next->data ;
   /* Delete last element */
   free(mov->next) ;
   mov->next = NullStack ;
   return(value) ;
}

/*---------------------------------------------------------------------------
 * Function :  topstack()
 * In       :  stack pointer
 * Out      :  stack item
 * Job      :  returns a copy of the value on top of the stack
 * Notice   :  does not pop the top value. Stack is not modified
 *--------------------------------------------------------------------------*/

item * 
topstack( stack *sp )
{
   item *  value ;
   stack * mov ;

   if (sp == NullStack) {
      cp_error("unexpected NULL stack: aborting") ;
      exit(-1) ;
   }

   /* Only one element means the stack is empty	*/
   if (sp->next == NullStack) {
      return NullItem ;
   }

   /* go to last element */
   mov = sp ;
   while (mov->next != NullStack) {
      mov = mov->next ;
   }	
   if ((value = (item*)malloc(sizeof(item))) == NullItem) {
      cp_error("cannot allocate stack item: aborting") ;
      exit(-1) ;
   }
   /* copy information into newly allocated stack element */
   value->t = mov->data->t ;
   value->v = mov->data->v ;
   return(value) ;
}	

/*---------------------------------------------------------------------------
 * Function :  gettokentype()
 * In       :  token (character string)
 * Out      :  symbol: OPERAND or OPERATOR
 * Job      :  find out if a token is a number or an operator
 * Notice   :
 *--------------------------------------------------------------------------*/

static itemType
gettokentype( char * token ) 
{
   regex_t re_expr;
   char lista[] = "^C[0-9]*_[0-9]*,(C[0-9]*_[0-9]*,)*C[0-9]*_[0-9]*$" ;
   char funct[] = "^(AVE|WAVE|STDEV|STERR|WSTDEV|WSTERR)\\([^)]*\\)$" ;
   char funct2[] = "^(LOG|LOG2|LOG10|EXP|C_SIGN|SQRT|ABS)$" ;	
   char column[] = "^C[0-9]*_[0-9]*$|^(ID|NS)$" ;
   char numb[] = "^([0-9]+[.]?[0-9]*|[.][0-9]+)([eE][+-]?[0-9]+)?$" ;

   if (token == (char*)NULL) {
	   cp_error("internal: cannot get token") ;
	   return unknown ;
   }

   if (token[0] == '(') {
      cp_message(0,4,"matched a left bracket\n");
      return left_bracket ;
   }

   if (token[0] == ')') {
      cp_message(0,4,"matched a right bracket\n");
      return right_bracket ;
   }	

   /* try to match a vector list: */
   if (regcomp(&re_expr, &lista[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for vector-list: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      cp_message(0,4,"matched a vector list\n");
      regfree(&re_expr) ;
      return vector_list;
   }
   /* try to match a computed vector: */
   if (regcomp(&re_expr, &funct[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for computed vector: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      cp_message(0,4,"matched a computed vector\n");
      regfree(&re_expr) ;
      return computed_vector;
   }
   /* try to match a vector name: */
   if (regcomp(&re_expr, &column[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for vectors: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      cp_message(0,4,"matched a single vector\n");
      regfree(&re_expr) ;
      return vector;
   }
   /* try to match a number: */
   if (regcomp(&re_expr, &numb[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for numbers: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      cp_message(0,4,"matched a number\n");
      regfree(&re_expr) ;
      return number ;
   }
   /* try to match a unary function: */
   if (regcomp(&re_expr, &funct2[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for unary operators: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      cp_message(0,4,"matched a function\n");
      regfree(&re_expr) ;
      return unary ;
   }

   /* no number, no image. It should be an operator */
   if ((token[0] == '+') || 
       (token[0] == '-') ||
       (token[0] == '*') ||
       (token[0] == '/') ||
       (token[0] == '^') ||
       (token[0] == 'e') ||
       (token[0] == 'l')) {
      cp_message(0,4,"matched an operator\n");
      return operator ;
   }

   cp_error("unrecognized token: [%s]", token) ;
   return unknown ;
}

/*---------------------------------------------------------------------------
 * Function :  priority()
 * In 	    :  1 char from: + - * / l ^
 * Out 	    :  int
 * Job	    :  returns a priority for a given operation
 *             the bigger the returned integer, the higher the priority
 * Notice   :
 *--------------------------------------------------------------------------*/

static int
priority( int op )
{
   int p ;

   switch(op) {
      /* addition and subtration have the same priority */
      case '+':
      case '-':
      p = 1 ;
      break ;

      /* multiplication and division have the same priority */
      case '*':
      case '/':
      p = 2 ;
      break ;

      /* logarithm	*/
      case 'l':
      p = 3 ;
      break ;

      /* exponentiation	*/
      case '^':
      case 'e':
      p = 4 ;
      break ;

      /* default: lowest possible priority (should not happen) */
      default:
      p=0 ;
      break ;
   }

   return p ;
}	

/*---------------------------------------------------------------------------
 * Function :  applyOperator()
 * In       :  two items and an operator
 * Out      :  one item
 * Job      :  operates arithmetic between vectors and numbers
 * Notice   :
 *--------------------------------------------------------------------------*/

static item * 
applyOperator( item * first, item * second, int op )
{
   item      * out ;
   OneVector * tmpcube ;

   out = (item*)malloc(sizeof(item)) ;
   if (out == NullItem) {
      cp_error("memory allocation failure for operator: aborting") ;
      return(NullItem) ;
   }

   if ((first->t == number) && (second->t == number)) {

      /* Two numbers */
      out->t = number ;
      if(first->v.f.type=='d'&&second->v.f.type=='d')
         out->v.f.type = 'd' ;
      else
         out->v.f.type = 'f' ;

      switch(op) {
	 case '+':
	    out->v.f.val = first->v.f.val + second->v.f.val ;
            out->v.f.ndig = max(first->v.f.ndig,second->v.f.ndig) ;
            out->v.f.ndec = max(first->v.f.ndec,second->v.f.ndec) ;
	    break ;

	 case '-':
	    out->v.f.val = first->v.f.val - second->v.f.val ;
            out->v.f.ndig = max(first->v.f.ndig,second->v.f.ndig) ;
            out->v.f.ndec = max(first->v.f.ndec,second->v.f.ndec) ;
	 break ;

	 case '*':
	    out->v.f.val = first->v.f.val * second->v.f.val ;
            out->v.f.ndig = first->v.f.ndig + second->v.f.ndig - 1 ;
            out->v.f.ndec = first->v.f.ndec + second->v.f.ndec ;
	    break ;

	 case '/':
	    if (fabs((double)second->v.f.val) <= 1e-40) {
	       cp_error("error: division by zero requested: aborting") ;
	       free(out) ;
	       return(NullItem) ;
	    }
	    out->v.f.val = first->v.f.val / second->v.f.val ;
            out->v.f.type = 'g' ;
            out->v.f.ndig = first->v.f.ndig + second->v.f.ndig ;
            out->v.f.ndec = max(first->v.f.ndec,second->v.f.ndec) ;
	    break ;

	 case 'l':
	    if ((double)first->v.f.val <= 1e-40 ||
	        (double)second->v.f.val <= 1e-40) {
	       cp_error("logarithm requested on negative or zero value") ;
	       free(out) ;
	       return(NullItem) ;
	    }
	    out->v.f.val = log((double)first->v.f.val) / log((double)second->v.f.val) ;
            out->v.f.type = 'f' ;
            out->v.f.ndig = first->v.f.ndig ;
            out->v.f.ndec = first->v.f.ndec ;
	    break ;

	 case '^':
	     out->v.f.val = pow((double)first->v.f.val, (double)second->v.f.val);
             out->v.f.ndig = first->v.f.ndig ;
             out->v.f.ndec = max(first->v.f.ndec,second->v.f.ndec) ;
	     break ;

	 default:
	    cp_error("unrecognized operator: %c", op) ;
	    free(out) ;
	    return(NullItem) ;
      }

   } else if ((first->t == vector) && (second->t == vector)) {

      /* Two Vectors */
      out->t = vector ;
      if (op!='+' && op!='-' && op!='*' && op!='/') {
	 cp_error("operation %c is invalid between vectors", op);
	 free(out) ;
	 return(NullItem) ;
      }
      out->v.vector = op_vector(first->v.vector, second->v.vector, op) ;
      destroy_vector(first->v.vector) ;
      destroy_vector(second->v.vector) ;
      if (out->v.vector == NullVector) {
	 cp_error("error during vector arithmetic: aborting") ;
	 free(out) ;
	 return(NullItem) ;
      }

   } else if ((first->t == vector) && (second->t == number)) {

      /* A vector and a number */
      out->t = vector ;
      if (op!='+' && op!='-' && op!='*' && op!='/' && op!='^' && op!='l' && op!='e') {
	 cp_error("unrecognized operation: %c", op) ;
	 free(out) ;
	 return(NullItem) ;
      }
      out->v.vector = cst_op_vector(first->v.vector, second->v.f, op) ;
      destroy_vector(first->v.vector) ;
      if (out->v.vector == NullVector) {
	 cp_error("error during vector arithmetic: aborting") ;
	 free(out) ;
	 return(NullItem) ;
      }

   } else if ((first->t == number) && (second->t == vector)) {

      /* a number and a vector */
      OneNumber tmp = {-1.0,4,1,'d'} ;
      out->t = vector ;
      switch(op) {
	 case '+':
	 case '*':
	 case 'l':
	 out->v.vector = cst_op_vector(second->v.vector, first->v.f, op) ;
	 break ;

         /* Exponentiation: 'vec e num' means 'pow(num,vec)' */
         case '^':
         out->v.vector = cst_op_vector(second->v.vector, first->v.f, 'e') ;
         break ;
         case 'e':
         out->v.vector = cst_op_vector(second->v.vector, first->v.f, '^') ;
         break ;

	 /* Subtraction and division need two passes */
	 case '-':
	 tmpcube = cst_op_vector(second->v.vector, tmp, '*') ;
	 out->v.vector = cst_op_vector(tmpcube, first->v.f, '+') ;
	 destroy_vector(tmpcube) ;
	 break ;

	 case '/':
	 tmpcube = cst_op_vector(second->v.vector, tmp, '^') ;
	 out->v.vector = cst_op_vector(tmpcube, first->v.f, '*') ;
	 break ;

	 default:
	 cp_error("unrecognized operation: %c", op) ;
	 free(out) ;
	 return NullItem ;
      }
      destroy_vector(second->v.vector) ;
      if (out->v.vector == NullVector) {
	 cp_error("error during vector arithmetic: aborting") ;
	 free(out) ;
	 return(NullItem) ;
      }
   } else {
      cp_error("invalid item in polish arithmetic expression") ;
      cp_error("type identification error: aborting") ;
      free(out) ;
      return(NullItem) ;
   }
   return(out) ;
}

/*---------------------------------------------------------------------------
 * Function :  applyUnary()
 * In       :  one item and an operator
 * Out      :  one item
 * Job      :  apply arithmetic unary operators on vectors and numbers
 * Notice   :
 *--------------------------------------------------------------------------*/

static item * 
applyUnary( item * first, opType op )
{
   item      * out ;

   out = (item*)malloc(sizeof(item)) ;
   if (out == NullItem) {
      cp_error("memory allocation failure for operator: aborting") ;
      return(NullItem) ;
   }

   if (first->t == number) {

      /* One number */
      out->t = number ;
      out->v.f.type = first->v.f.type ;
      out->v.f.ndig = first->v.f.ndig ;
      out->v.f.ndec = first->v.f.ndec ;

      switch(op) {

         case cng_sign:
	    out->v.f.val = -1.0 * (double)first->v.f.val ;
	    break ;

	 case op_log:
	    if ((double)first->v.f.val <= 1e-40) {
	       cp_error("logarithm requested on negative or zero value") ;
	       free(out) ;
	       return(NullItem) ;
	    }
	    out->v.f.val = log((double)first->v.f.val) ;
            out->v.f.type = 'g' ;
	    break ;

	 case op_log2:
	    if ((double)first->v.f.val <= 1e-40) {
	       cp_error("logarithm requested on negative or zero value") ;
	       free(out) ;
	       return(NullItem) ;
	    }
	    out->v.f.val = log((double)first->v.f.val) / log(2.0) ;
            out->v.f.type = 'g' ;
            out->v.f.ndig += 3 ;
	    break ;

	 case op_log10:
	    if ((double)first->v.f.val <= 1e-40) {
	       cp_error("logarithm requested on negative or zero value") ;
	       free(out) ;
	       return(NullItem) ;
	    }
	    out->v.f.val = log10((double)first->v.f.val) ;
            out->v.f.type = 'g' ;
            out->v.f.ndig += 3 ;
	    break ;

	 case op_exp:
            if ( (double)first->v.f.val < EXPMIN || (double)first->v.f.val > EXPMAX ) {
	       cp_error("exponential requested on exceptional argument") ;
	       free(out) ;
	       return(NullItem) ;
	    }
	    out->v.f.val = exp((double)first->v.f.val) ;
            out->v.f.type = 'g' ;
            out->v.f.ndig += 3 ;
	    break ;

	 case op_sqrt:
            if ((double)first->v.f.val <= -1e-40) {
               cp_error("square root requested on negative value") ;
               free(out) ;
               return(NullItem) ;
            }
            out->v.f.val = sqrt((double)first->v.f.val) ;
            out->v.f.type = 'g' ;
	    break ;

	 case op_abs:
	    out->v.f.val = (first->v.f.val>=0.0) ? (double)first->v.f.val : -1.0 * (double)first->v.f.val ;
	    break ;

	 default:
	    cp_error("unrecognized operator: %c", op) ;
	    free(out) ;
	    return(NullItem) ;
      }
   } else if (first->t == vector) {

      /* One Vector */
      out->t = vector ;
      out->v.vector = fn_op_vector(first->v.vector, op) ;
      destroy_vector(first->v.vector) ;
      if (out->v.vector == NullVector) {
	 cp_error("error during vector arithmetic: aborting") ;
	 free(out) ;
	 return(NullItem) ;
      }

   } else {
      cp_error("type identification error: aborting") ;
      free(out) ;
      return(NullItem) ;
   }
   return(out) ;
}

/*---------------------------------------------------------------------------
 * Function :  get_number()
 * In       :  a string
 * Out      :  number value, type, digits and decimals number
 * Job      :	
 * Notice   :
 *--------------------------------------------------------------------------*/

static OneNumber 
get_number(char *token)
{
   OneNumber ret ;
   double    read_f ;
   char     *ptr1, *ptr2 ;

   (void)sscanf(token, "%lg", &read_f) ;
   ret.val = (double)read_f ;

   /* Look for '.' to guess data type and number of decimals */
   if( (ptr1 = index(token,'.')) == (char*)NULL ) {
      ret.type = 'd' ;
      ret.ndec = 0 ;
   }
   else {
      if( (ptr2 = index(token,'E')) != (char*)NULL ) {
         ret.type = 'E' ;
         ret.ndec = (int)(strlen(ptr1)-strlen(ptr2))-1 ;
      }
      else {
         if( (ptr2 = index(token,'e')) != (char*)NULL ) {
            ret.type = 'e' ;
            ret.ndec = (int)(strlen(ptr1)-strlen(ptr2))-1 ;
         }
         else {
            ret.type = 'f' ;
            ret.ndec = (int)strlen(ptr1)-1 ;
         }
      }
   }
   /* Get number of digits */
   ret.ndig = (int)strlen(token) ;

   return ret ;
}

/*---------------------------------------------------------------------------
 * Function :  get_function()
 * In       :  a string
 * Out      :  function type
 * Job      :	
 * Notice   :
 *--------------------------------------------------------------------------*/

static opType 
get_function(char *token)
{
   regex_t re_expr;
   char f1[]  = "^(LOG)$" ;
   char f2[]  = "^(LOG2)$" ;
   char f3[]  = "^(LOG10)$" ;
   char f4[]  = "^(EXP)$" ;
   char f5[]  = "^(C_SIGN)$" ;
   char f6[]  = "^(SQRT)$" ;
   char f7[]  = "^(ABS)$" ;
   char f8[]  = "^(AVE)$" ;
   char f9[]  = "^(WAVE)$" ;
   char f10[] = "^(STDEV)$" ;
   char f11[] = "^(STERR)$" ;
   char f12[] = "^(WSTDEV)$" ;
   char f13[] = "^(WSTERR)$" ;
   
   if (token == (char*)NULL) {
	   cp_error("internal: cannot get token") ;
	   return unknown ;
   }

   /* try to match a function */
   if (regcomp(&re_expr, &f1[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return op_log;
   }
   if (regcomp(&re_expr, &f2[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return op_log2;
   }
   if (regcomp(&re_expr, &f3[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return op_log10;
   }
   if (regcomp(&re_expr, &f4[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return op_exp;
   }
   if (regcomp(&re_expr, &f5[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return cng_sign ;
   }
   if (regcomp(&re_expr, &f6[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return op_sqrt ;
   }
   if (regcomp(&re_expr, &f7[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return op_abs ;
   }
   if (regcomp(&re_expr, &f8[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return fn_ave ;
   }
   if (regcomp(&re_expr, &f9[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return fn_wave ;
   }
   if (regcomp(&re_expr, &f10[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return fn_stdev ;
   }
   if (regcomp(&re_expr, &f11[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return fn_sterr ;
   }
   if (regcomp(&re_expr, &f12[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return fn_wstdev ;
   }
   if (regcomp(&re_expr, &f13[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error("cannot compile rule for function: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      regfree(&re_expr) ;
      return fn_wsterr ;
   }
   cp_error("unrecognized token: [%s]", token) ;
   return unknown ;
}


