/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: parser.h
 * Author 	: Paolo Montegriffo
 * Created on	: July, 2000	
 * Language	: ANSI C
 * Description	: This set of definitions has been derived from ccube.c
 *                module written by Nicolas Devillard for ESO Eclipse software.
 * Version      : 2.0 - October 1, 2000
 *                2.1.2 - September 4, 2001
 *--------------------------------------------------------------------------*/

#ifndef PARSER_H
#define PARSER_H 1

#include "common.h"

BEGIN_C_DECLS

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

/*
 * 2 types of arithmetic expressions are supported
 */
#define ARITHM_STANDARD		0
#define ARITHM_POLISH		1

/* The separator is defined to break an expression into tokens	*/
#define SEPARATOR		" "
#define MAX_OP			100 /* maximum nb of arguments in an expression	*/

#define NAME_SIZE               512

#define OPT_STANDARD		1001
#define OPT_POLISH		1002

/*----------------------------------------------------------------------------
 * 								New types
 *--------------------------------------------------------------------------*/

/*
 * Following is a list of labels to identify the operator type
 */

typedef enum opType_tag {
   /*
    * Unary operators
    */
   cng_sign,      /* change of sign */
   op_log,        /* Log_e */
   op_log2,       /* Log_2 */
   op_log10,      /* Log_10 */
   op_exp,        /* Exponential */
   op_sqrt,       /* Square root */
   op_abs,        /* Absolute value */
   /*
    * Binary operators   -   not used
    */
   op_add,
   op_sub,
   op_mul,
   op_div,
   op_pot,
   /*
    * Multi-arguments operators   -   not used
    */
   fn_ave,
   fn_wave,
   fn_stdev,
   fn_sterr,
   fn_wstdev,
   fn_wsterr
} opType ;

/*
 * OneVector and OneNumber definitions
 */

typedef struct OneVector_tag
{
   double * data ;           /* Pointers to vector zones */
   char  ** lab ;            /* Pointers to vector zones */
   int    * is_in;           /* Pointers to associated flag zones */
   int      size;            /* Vector size */
   int      width;           /* Field width */
   int      ndec;            /* Number of decimals */
   char     type;            /* Vector type */
} OneVector ;

typedef struct OneNumber_tag
{
   double   val ;            /* Number value */
   int      ndig;            /* Number of digits */
   int      ndec;            /* Number of decimals */
   char     type;            /* Number type */
} OneNumber ;

typedef union itemValue_tag {
   OneVector * vector ;
   OneNumber f ;
   char	    op ;
   opType   unary ;
} itemValue ;

/*
 * Following is a list of labels to identify the object type
 */

typedef enum itemType_tag { 
   vector,
   vector_list,
   computed_vector,
   number,
   string,
   /* 
    * left and right brackets are no operators, they are used to
    * convert standard arithmetic expressions into PRN ones.
    */
   left_bracket,
   right_bracket, 
   /* a right bracket cannot be pushed on the stack with
    * the current algorithm. However, the definition is
    * included, just in case...
    */
   operator,
   /* funtions are included but not used (replaced by computed_vector type)*/
   function,
   unary,
   unknown
} itemType ;

/*
 * An item is a struct containing a label for its type, and a value
 */

typedef struct item_tag {
   itemValue v ;
   itemType  t ;
} item ;

/*
 * This is one stack element: it contains an item, and a pointer
 * to the next item.
 */

typedef struct stack_tag {
	item * data ;
	struct stack_tag * next ;
} stack ;

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#define NullVector  (OneVector*)NULL
#define NullString  (OneString*)NULL
#define NullStack   (stack*)NULL
#define NullItem    (item*)NULL

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

char   * tokenizeExpression(char *arexp) ;
item   * ParseExpression( char *expr, int arithm) ;
void     push(item *value, stack *sp) ;
item   * pop(stack *sp) ;
item   * topstack(stack *sp) ;

END_C_DECLS

#endif   /* !PARSER_H */
/*--------------------------------------------------------------------------*/
