/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <math.h>
#include "cp_messages.h"
#include "fitplate.h"
#include "matrix2.h"

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/


/*-----------------------------------------------------------------------------
 * Function  : qrfit
 * In        : x1[], x2[], y[],   ... fit y[] as function of x1[],x2[]
 *             sig[],             ... the error scale lenght of each data point
 *             ndat,              ... the number of fitting points
 *             fc[],              ... the fitting coefficents
 *             ia[],              ... flag vector: 1 to fit the corresponding
 *                                    coefficent, 2 to subtract off dependences 
 *                                    on known pieces of fitting function
 *             ma,                ... the number of fitting coefficents
 *             funcs,             ... the fitting function
 * Out       : fc[],              ... the fitted coefficents
 *             *chisq             ... the resulting chi-square
 * Purpouse  : Fit a bidimensional function using a chi-2 minimization
 *             Fitting equations are solved using singular value decomposition
 * Note      : Derived from Numerical Recipies
 *             Originally written for VIRMOS DATA REDUCTION SOFTWARE
 *---------------------------------------------------------------------------*/


void qrfit(double x1[], double x2[], double y[], double sig[], int ndata, 
           double fc[], int ia[], int ma, double *chisq,
	   void (*funcs)(double, double, double [], int))
{
   MAT    *A, *QR ;
   VEC    *b, *x, *diag, *afunc ;
   int     i, j, k, mfit ;
   double  ym, tmp, sum ;

   /* Get number of parameters to fit */
   for ( mfit=j=0; j<ma; j++ )
      if ( ia[j] == 1 ) mfit++ ;
   if ( mfit == 0 ) cp_error("lfit: no parameters to be fitted");

   /* Allocate memory for working arrays */
   A = m_get(ndata, mfit) ;
   b = v_get(ndata) ;
   afunc = v_get(ma) ;

   /* Accumulate coefficents for normal equations */
   for ( i=0; i<ndata; i++ ) {
      (*funcs)(x1[i], x2[i], afunc->ve, ma) ;
      ym = y[i] ;

      /* Subtract off dependences on known pieces of fitting function */
      if ( mfit < ma ) 
         for ( j=0; j<ma; j++ ) 
            if ( ia[j] == 2 ) ym -= fc[j]*afunc->ve[j] ;
      tmp = 1.0/sig[i] ;
      for ( k=j=0; j<ma; j++ ) 
         if ( ia[j] == 1 ) A->me[i][k++] = afunc->ve[j]*tmp ;
      b->ve[i] = ym*tmp ;
   }
   diag = v_get(A->m) ;

   /* QR is to be the QR factorisation of A */
   QR = m_copy(A, MNULL) ;
   QRfactor(QR, diag) ;
   /* solve for x */
   x = QRsolve(QR, diag, b, VNULL) ;
   /* ... and work out norm of errors... */
   /* printf("||A*x-b|| = %g\n", v_norm2(v_sub(mv_mlt(A,x,VNULL),b,VNULL))); */

   /* Set output array */

   for ( k=j=0; j<ma; j++ ) 
      if ( ia[j] == 1 ) fc[j] = x->ve[k++] ;

   /* Evaluate the chi-square of the fit */
   *chisq = 0.0 ;
   for ( i=0; i<ndata; i++ ) {
      (*funcs)(x1[i], x2[i], afunc->ve, ma) ;
      for ( sum=0.0,j=0; j<ma; j++ ) sum += fc[j]*afunc->ve[j] ;
      *chisq += (tmp=(y[i]-sum)/sig[i],tmp*tmp) ;
   }
   /* Here call svdvar if need covariance matrix */

   /* Free memory */
   M_FREE(A) ;
   M_FREE(QR) ;
   V_FREE(b) ;
   V_FREE(diag) ;
   V_FREE(afunc) ;
   V_FREE(x) ;
}
