/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: vector_arith.c
 * Author 	: Paolo Montegriffo
 * Created on	: July, 2000	
 * Language	: ANSI C
 * Description	: 
 * Version      : 2.0 - October 3, 2000
 *                2.1.2 - September 4, 2001
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <ctype.h>
#include <math.h>
#include <regex.h>
#include <float.h>

#include "cp_messages.h"
#include "fileinfo.h"
#include "parser.h"
#include "vector_arith.h"
#include "catalib.h"

/*----------------------------------------------------------------------------
 * 								Macros
 *--------------------------------------------------------------------------*/

#ifndef min
#define min(a,b) (((a)<=(b)) ? (a) : (b))
#endif
#ifndef max
#define max(a,b) (((a)>=(b)) ? (a) : (b))
#endif

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#define AVE   1
#define WAVE  2
#define STDEV 3
#define STERR 4
#define WSTDEV 5
#define WSTERR 6

#define UNDEF    0
#define IS_COINC 1
#define IS_LIST  2
#define ACTIVE   7
#define DISPLAY  8
#define INFO     9

#define MAXLENSTR 20

#define EXPMAX log(DBL_MAX)
#define EXPMIN log(DBL_MIN)

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static double * get_column(int n_file, int n_col, int * len, int * ndec) ;
char  **get_column_s(int n_file, int n_col, int * len, int * ndec) ;
static char  ** GetErrorName(char ** list, int n_list) ;
static itemType checklist( char * token ) ;

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern CoincId  *coinc ;
extern FileSet  *f_set ;
extern int       N_NullVal ;      /* number of elements in Null values list */
extern double    NullValues[] ;   /* Input Null values */
extern int       max_iter ;       /* maximum number of sigma rejection iterations */
extern double    nsig_rej ;       /* rejection limit in stdev units */
extern int       n_min ;          /* minimum number of measures for a valid result */
extern int       sigma ;          /* 0 to turn-off sigma rejection */

double    OutNull = 0.0 ;        /* Output Null value */

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * Function	:	op_vector()
 * In 		:	2 vectors, operation to perform
 * Out 		:	newly allocated vector
 * Job		:	4 operations between 2 vectors
 * Notice	:	returns NULL in case of error
 *--------------------------------------------------------------------------*/

OneVector *
op_vector(OneVector * vector1, OneVector * vector2, int operation) 
{
   OneVector * out_vector ;

   if (vector1==NullVector || vector2==NullVector) return NullVector ;
   out_vector = copy_vector(vector1) ;
   cp_message(0, 2, "performing vector arithmetic\n") ;

   switch(operation) {

      case '+':
      vector_add(out_vector, vector2) ;
      break ;

      case '-':
      vector_sub(out_vector, vector2) ;
      break ;

      case '*':
      vector_mul(out_vector, vector2) ;
      break ;

      case '/':
      vector_div(out_vector, vector2) ;
      break ;

      default:
      cp_error("illegal requested operation: aborting vector arithmetic") ;
      destroy_vector(out_vector) ;
      out_vector = NullVector ;
      return out_vector ;
   }
   out_vector->width = max(out_vector->width, vector2->width ) ;
   out_vector->ndec = max(out_vector->ndec, vector2->ndec ) ;
   if(vector2->type != 'd') out_vector->type = vector2->type ;

   return out_vector ;
}

OneVector *
fn_op_vector(OneVector * vector_in, opType operation) 
{
   OneVector * vector_out ;
   int      i ;

   if (vector_in==NullVector) return NullVector ;
   vector_out = new_vector(vector_in->size) ; ;
   for (i=0 ; i < vector_out->size ; i++)
      vector_out->is_in[i] = vector_in->is_in[i] ;
   vector_out->width = vector_in->width ;
   vector_out->ndec = vector_in->ndec ;
   vector_out->type = vector_in->type;

   cp_message(0, 2, "performing vector arithmetic\n") ;
   switch(operation) {

      case cng_sign:
         for (i=0 ; i<vector_out->size ; i++)
            vector_out->data[i] = -1.0 * (double)vector_in->data[i] ;
         break ;

      case op_log:
         for (i=0 ; i<vector_out->size ; i++) {
            if((double)vector_in->data[i] > -1e-40)
               vector_out->data[i] = (double)log((double)vector_in->data[i]) ;
            else {
               vector_out->data[i] = OutNull ;
               vector_out->is_in[i] = 0 ;
            }
         }
         vector_out->width += 3 ;
         vector_out->ndec += 3 ;
         vector_out->type = 'f';
         break ;

      case op_log2:
         for (i=0 ; i<vector_out->size ; i++) {
            if((double)vector_in->data[i] > -1e-40)
               vector_out->data[i] = (double)log((double)vector_in->data[i]) / log(2.0) ;
            else {
               vector_out->data[i] = OutNull ;
               vector_out->is_in[i] = 0 ;
            }
         }
         vector_out->width += 3 ;
         vector_out->ndec += 3 ;
         vector_out->type = 'f';
         break ;

      case op_log10:
         for (i=0 ; i<vector_out->size ; i++) {
            if((double)vector_in->data[i] > -1e-40)
               vector_out->data[i] = (double)log10((double)vector_in->data[i]) ;
            else {
               vector_out->data[i] = OutNull ;
               vector_out->is_in[i] = 0 ;
            }
         }
         vector_out->width += 3 ;
         vector_out->ndec += 3 ;
         vector_out->type = 'f';
         break ;

      case op_exp:
         for (i=0 ; i<vector_out->size ; i++) {
            if( (double)vector_in->data[i]>EXPMIN &&
                (double)vector_in->data[i]<EXPMAX )
	       vector_out->data[i] = (double)exp((double)vector_in->data[i]) ;
            else {
               vector_out->data[i] = OutNull ;
               vector_out->is_in[i] = 0 ;
            }
         }
         vector_out->width += 3 ;
         vector_out->ndec += 3 ;
         vector_out->type = 'g';
         break ;

      case op_sqrt:
         for (i=0 ; i<vector_out->size ; i++) {
            if((double)vector_in->data[i] > -1e-40)
	       vector_out->data[i] = sqrt((double)vector_in->data[i]) ;
            else {
               vector_out->data[i] = OutNull ;
               vector_out->is_in[i] = 0 ;
            }
         }
         vector_out->width += 3 ;
         vector_out->ndec += 3 ;
         vector_out->type = 'f';
         break ;

      case op_abs:
         for (i=0 ; i<vector_out->size ; i++) {
            if((double)vector_in->data[i] >= 0.0)
	       vector_out->data[i] = (double)vector_in->data[i] ;
            else
               vector_out->data[i] = -1.0 * (double)vector_in->data[i] ;
         }
         break ;

      default:
         cp_error("illegal requested operation: aborting vector arithmetic") ;
         destroy_vector(vector_out) ;
         vector_out = NullVector ;
   }

   return vector_out ;
}

/*----------------------------------------------------------------------------
 * Function	:	cst_op_vector()
 * In 		:	1 vector, 1 constant, operation to perform
 * Out 		:	1 vector
 * Job		:	4 operations between a vector and a constant
 * Notice	:	possible operations are:
 *				Addition	'+'
 *				Subtraction 	'-'
 *				Multiplication	'*'
 *				Division	'/'
 *				Logarithm	'l'
 *				Power		'^'
 *				Exponentiation	'e'

********** mettere controllo su valori negativi per log e divisioni per zero...****

 *--------------------------------------------------------------------------*/

OneVector *
cst_op_vector( OneVector * vector_in, OneNumber constant, int operation)
{
   OneVector * vector_out ;
   int         i ;
   double      invlog ;
   double      dexp ;
   double      invconst ;

   /* Error handling   */
   if (vector_in == NullVector) {
      cp_error("NULL input vector: cannot compute") ;
      return(NullVector) ;
   }

   if ((fabs((double)constant.val) < 1e-40) && (operation == '/')){
      cp_error("division by zero requested in vector/constant operation") ;
      return(NullVector) ;
   }

    /* Allocations  */
   vector_out = new_vector(vector_in->size) ;
   if (vector_out == NullVector) {
      cp_error("cannot allocate output vector for constant .op. vector\n") ;
      return(NullVector) ;
   }
   for (i=0 ; i < vector_out->size ; i++)
      vector_out->is_in[i] = vector_in->is_in[i] ;

   /* Set outvec type accordingly to input vector and constant type */
   vector_out->width = max(vector_in->width, constant.ndig) ;
   vector_out->ndec  = max(vector_in->ndec, constant.ndec)  ;
   vector_out->type  = vector_in->type  ;
   if(vector_in->type=='d'&&constant.type!='d') vector_out->type = constant.type ;

   switch(operation)
   {
      case '+':
         for (i=0 ; i<vector_out->size ; i++)
	    vector_out->data[i] = (double)((double)vector_in->data[i] + constant.val) ;
         break ;

      case '-':
         for (i=0 ; i<vector_out->size ; i++)
	    vector_out->data[i] = (double)((double)vector_in->data[i] - constant.val) ;
         break ;

      case '*':
         for (i=0 ; i<vector_out->size ; i++)
	    vector_out->data[i] = (double)((double)vector_in->data[i] * constant.val) ;
         break ;

      case '/':
         /* Multiplications are faster than divisions !	*/
         invconst = (double)1.0 / constant.val ;
         for (i=0 ; i<vector_out->size ; i++)
	    vector_out->data[i] = (double)((double)vector_in->data[i] * invconst) ;
         vector_out->type = 'g' ;
         vector_out->ndec  = max(vector_out->ndec,6) ;
         break ;

      case 'l':
         invlog = (double)(1.0 / log((double)constant.val)) ;
         for (i=0 ; i<vector_out->size ; i++) {
            if((double)vector_in->data[i] > 1e-40)
               vector_out->data[i] = (double)log((double)vector_in->data[i])*invlog ;
            else {
               vector_out->data[i] = OutNull ;
               vector_out->is_in[i] = 0 ;
            }
         }
         vector_out->width += 3 ;
         vector_out->ndec += 3 ;
         vector_out->type = 'f';
         break ;
      break ;

      case '^':
         dexp = (double)constant.val ;
         for (i=0 ; i<vector_out->size ; i++) {
            if((double)vector_in->data[i] > 1e-40 ||
               (double)vector_in->data[i] < -1e-40)
	       vector_out->data[i] = (double)pow((double)vector_in->data[i], dexp) ;
            else {
               if(dexp!=1.0)
                  vector_out->data[i] = 0.0 ;
               else {
                  vector_out->data[i] = OutNull ;
                  vector_out->is_in[i] = 0 ;
               }
            }
         }
         vector_out->width += 3 ;
         vector_out->ndec += 3 ;
         vector_out->type = 'g';
         break ;

      case 'e':
      dexp = (double)constant.val ;
      for (i=0 ; i<vector_out->size ; i++)
	 vector_out->data[i] = (double)pow(dexp, (double)vector_in->data[i]) ;
      break ;

      default:
      cp_error("unrecognized requested operation : aborting") ;
      break ;
   }
 
   return(vector_out) ;
}

/*----------------------------------------------------------------------------
 * Function	:	vector_add()
 * In 		:	two vectors
 * Out 		:	0 if Ok, -1 if error, -2 if cannot do it in fast mode	
 * Job		:	add a vector to another
 * Notice	:	the first vector is modified: it contains the result of
 * 				addition.
 *--------------------------------------------------------------------------*/

int
vector_add( OneVector * c1, OneVector * c2 )
{
   int i ;

   if (c1->size != c2->size) {
      cp_error("vectors have different dimension: cannot add") ;
      return -1 ;
   }
   for (i=0 ; i< c1->size ; i++) {
      c1->data[i] += c2->data[i] ;
      c1->is_in[i] = min(c1->is_in[i], c2->is_in[i]) ;
   }

   return 0 ;
}

/*----------------------------------------------------------------------------
 * Function	:	vector_sub()
 * In 		:	two cubes
 * Out 		:	0 if Ok, -1 if error, -2 if cannot do it in fast mode	
 * Job		:	subtract one cube from another
 * Notice	:	the first cube is modified: it contains the result of
 * 				subtraction.
 *--------------------------------------------------------------------------*/

int
vector_sub( OneVector * c1, OneVector * c2 )
{
   int i ;

   if (c1->size != c2->size) {
      cp_error("vectors have different dimension: cannot add") ;
      return -1 ;
   }
   for (i=0 ; i< c1->size ; i++) {
      c1->data[i] -= c2->data[i] ;
      c1->is_in[i] = min(c1->is_in[i], c2->is_in[i]) ;
   }

   return 0 ;
}

/*----------------------------------------------------------------------------
 * Function	:	vector_mul()
 * In 		:	two vectors
 * Out 		:	0 if Ok, -1 if error, -2 if cannot do it in fast mode	
 * Job		:	multiply 2 vectors	
 * Notice	:	the first vector is modified: it contains the result of
 * 				multiplication.
 *--------------------------------------------------------------------------*/

int
vector_mul( OneVector * c1, OneVector * c2 )
{
   int   i ;

   if (c1->size != c2->size) {
      cp_error("vectors have different dimension: cannot multiply") ;
      return -1 ;
   }
   for (i=0 ; i< c1->size ; i++) {
      c1->data[i] *= c2->data[i] ;
      c1->is_in[i] = min(c1->is_in[i], c2->is_in[i]) ;
   }

   return 0 ;
}

/*----------------------------------------------------------------------------
 * Function	:	vector_div()
 * In 		:	two vectors
 * Out 		:	0 if Ok, -1 if error, -2 if cannot do it in fast mode	
 * Job		:	divide 2 vectors	
 * Notice	:	the first vector is modified: it contains the result of
 * 				division.
 *--------------------------------------------------------------------------*/

int
vector_div( OneVector * c1, OneVector * c2 )
{
   int i ;

   if (c1->size != c2->size) {
      cp_error ("vectors have different dimension: cannot multiply") ;
      return -1 ;
   }
   for (i = 0 ; i < c1->size ; i++) {
      if (fabs ((double)c2->data[i]) < 1e-10) {
         c1->data[i] = OutNull ;
         c1->is_in[i] = 0 ;
      }
      else
         c1->data[i] /= c2->data[i] ;
      c1->is_in[i] = min (c1->is_in[i], c2->is_in[i]) ;
   }

   return 0 ;
}

/*---------------------------------------------------------------------------
 * Function :   new_vector()
 * In       :   vector size
 * Out      :   New allocated structure
 * Job      :   Allocates an vector structure and fills it with
 *              relevant information. Also allocates data buffer
 * Notice   :   Object constructor
 *--------------------------------------------------------------------------*/

OneVector *
new_vector (int size)
{
   OneVector *ret_vector ;

   /* error handling: test entries */

   /* if ((size<1) || (size>MAX_COLUMN_NUMBER)) { */
   if (size < 1) {
      cp_error ("requested size: [%d]", size) ;
      cp_error ("cannot create such size: aborting vector creation") ;
      return (NullVector) ;
   }

   /* Try to allocate for a new structure  */
   ret_vector = (OneVector*) calloc ((size_t)1, sizeof(OneVector)) ;

   /* Fill structure with initialization information   */
   ((OneVector*)ret_vector)->size = (int)size ;
   /* Allocate memory for data buffer */
   ((OneVector*)ret_vector)->data = (double*) calloc ((size_t)size, sizeof(double)) ;
   ((OneVector*)ret_vector)->is_in = (int*) calloc ((size_t)size, sizeof(int)) ;
   return (ret_vector) ;
}

OneVector *
new_vector_s (int size, int type)
{
   OneVector *ret_vector ;

   /* error handling: test entries */

   /* if ((size<1) || (size>MAX_COLUMN_NUMBER)) { */
   if (size < 1) {
      cp_error ("requested size: [%d]", size) ;
      cp_error ("cannot create such size: aborting vector creation") ;
      return (NullVector) ;
   }
   if (type != 's' && type != 'f' && type != 'd' && type != 'g') {
      cp_error ("illegal requested type: [%c]: aborting",(char)type) ;
      return (NullVector) ;
   }

   /* Try to allocate for a new structure  */
   ret_vector = (OneVector*) calloc ((size_t)1, sizeof(OneVector)) ;

   /* Fill structure with initialization information   */
   ((OneVector*)ret_vector)->size = (int)size ;
   ((OneVector*)ret_vector)->type = (char)type ;
   /* Allocate memory for data buffer */
   if (type != 's') {
      ((OneVector*)ret_vector)->data = (double*) calloc ((size_t)size, sizeof(double)) ;
      ((OneVector*)ret_vector)->lab = (char**)NULL ;
   } else {
      int i ;
      ((OneVector*)ret_vector)->data = (double*)NULL ;
      ((OneVector*)ret_vector)->lab = (char**) calloc ((size_t)size, sizeof(char*)) ;
      for (i = 0; i < size; i++) ((OneVector*)ret_vector)->lab[i] = (char*)NULL ;
   }
   ((OneVector*)ret_vector)->is_in = (int*) calloc ((size_t)size, sizeof(int)) ;
   return (ret_vector) ;
}

/*---------------------------------------------------------------------------
 * Function :   destroy_vector()
 * In       :   Pointer to allocated structure
 * Out      :   void
 * Job      :   free memory associated to an vector structure
 * Notice   :   Object destructor
 *--------------------------------------------------------------------------*/

void
destroy_vector (OneVector * to_destroy) 
{
   int i ;
   if (to_destroy != NullVector) {
      /* Try to free data and flag area */
      if (to_destroy->data != (double*)NULL)
         free(to_destroy->data) ;
      to_destroy->data = (double*)NULL ;
      if (to_destroy->lab != (char**)NULL) {
         for (i = 0 ; i < to_destroy->size ; i++) 
            if (to_destroy->lab[i] != (char*)NULL) free (to_destroy->lab[i]) ;
         free (to_destroy->lab) ;
      }
      if (to_destroy->is_in != (int*)NULL)
         free (to_destroy->is_in) ;
      to_destroy->is_in = (int*)NULL ;

      /* free structure itself    */
      free (to_destroy) ;
      to_destroy = NullVector ;
   } else {
      cp_warning ("destruction requested on NULL vector") ;
   }
}

/*---------------------------------------------------------------------------
 * Function :   copy_vector()
 * In       :   1 vector
 * Out      :  	1 newly allocated vector 
 * Job      :   copy an vector structure
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector *
copy_vector (OneVector * src_vec)
{
   OneVector * dest_vec ;
   int i ;

   /* Error handling: test entries */
   if (src_vec == NullVector) {
      cp_error ("cannot copy NULL vector: aborting vector copy") ;
      return (NullVector) ;
   }

   /* Allocate new structure and fill up information fields */
   dest_vec = new_vector_s(src_vec->size,src_vec->type) ;
   if (dest_vec == NullVector) {
      cp_error ("cannot allocate output vector: aborting copy\n") ;
      return (NullVector) ;
   }

   /* Copy data and flag area */
   if (src_vec->type != 's') 
      (void)memcpy ((void*)dest_vec->data, (void*)src_vec->data,
        	    (size_t)src_vec->size * sizeof(double)) ;
   else {
      for (i = 0 ; i < src_vec->size ; i++) {
         dest_vec->lab[i] = (char*) calloc (strlen (src_vec->lab[i])+(size_t)1, sizeof(char)) ;
         (void) strcpy (dest_vec->lab[i], src_vec->lab[i]) ;
      }
   }
   (void) memcpy ((void*)dest_vec->is_in, (void*)src_vec->is_in,
          	  (size_t)src_vec->size * sizeof(int)) ;
   dest_vec->width = src_vec->width ;
   dest_vec->ndec  = src_vec->ndec ;
   dest_vec->type  = src_vec->type ;

   /* done */
   return (dest_vec) ;
}

/*---------------------------------------------------------------------------
 * Function :   load_vector()
 * In       :   vector name
 * Out      :   Pointer to newly allocated vector
 * Job      :   Loads a vector from a file ordering elements with a ranking
 *		table build with the coincidence table.
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector * 
load_vector (char *vec_name)
{
   OneVector  *loaded = NullVector ;
   Finfo     **f_info = f_set->info ;
   int         n_col, n_file ;
   int         i, j ;
   int         width ;
   int         ndec ;

   if (strcmp (vec_name,"ID") == 0 || strcmp (vec_name,"NS") == 0) {
      char stmp[32] ;
      int  len = 0 ;
      /* Allocate memory for returning vector */
      loaded = new_vector_s (coinc->nrec,'f') ;
      if (loaded == NullVector) {
         cp_error ("cannot allocate output vector: aborting extraction") ;
         return (NullVector) ;
      }
      /* Fill returning vector */
      for (j = 0; j < coinc->nrec; j++) {
	 if (strcmp (vec_name,"ID")==0) {
            ((OneVector*)loaded)->data[j] = coinc->id[j] ;
	 } else {
      	    ((OneVector*)loaded)->data[j] = coinc->ns[j] ;
	 }
	 ((OneVector*)loaded)->is_in[j] = 1 ;
	 sprintf (stmp,"%d", (int)((OneVector*)loaded)->data[j]);
	 len = max (len, (int)strlen (stmp)+1) ;
      }
      ((OneVector*)loaded)->width = len ;
      ((OneVector*)loaded)->ndec  = 0 ;
      ((OneVector*)loaded)->type  = 'd' ;
   }
   else {
      char type ;
      /* Get file and column number from vec_name */
      sscanf (vec_name, "C%d_%d", &n_col, &n_file);
      n_col -= 1;
      n_file -= 1;
      if (n_file >= coinc->nfiles) {
         cp_error ("illegal file number in %s: aborting",vec_name) ;
	 return (NullVector) ;
      }
      if (n_col >= f_info[n_file]->n_cols) {
         cp_error ("illegal column number in %s: aborting",vec_name) ;
	 return (NullVector) ;
      }
      /* Allocate memory for returning vector */
      type = f_info[n_file]->fmt[n_col].type ;
      loaded = new_vector_s (coinc->nrec, type) ;
      if (loaded == NullVector) {
         cp_error ("cannot allocate output vector: aborting extraction") ;
         return (NullVector) ;
      }

      if (type != 's') {
         double *tmp = (double*)NULL ;
         /* Get column from file */
         tmp = get_column(n_file, n_col, &width, &ndec);

         /* Fill returning vector */
         ((OneVector*)loaded)->width = width ;
         ((OneVector*)loaded)->ndec  = ndec ;
         ((OneVector*)loaded)->type  = f_info[n_file]->fmt[n_col].type ;
         for (j = 0; j < coinc->nrec; j++) {
	    if (coinc->rank[n_file][j]) {
               ((OneVector*)loaded)->data[j] = tmp[coinc->rank[n_file][j]-1] ;
	       ((OneVector*)loaded)->is_in[j] = 1 ;
	       /* Checks for Null values */
	       for (i = 0; i < N_NullVal; i++) {	    
	          if (loaded->data[j] == NullValues[i]) 
                     ((OneVector*)loaded)->is_in[j] = 0 ;
	       }
	    }
	    else {
               ((OneVector*)loaded)->data[j] = 0.0 ;
	       ((OneVector*)loaded)->is_in[j] = 0 ;
	    }
         }
         free (tmp);
      } else {
         char **tmp = (char**)NULL ;
         /* Get column from file */
         tmp = get_column_s(n_file, n_col, &width, &ndec);

         /* Fill returning vector */
         ((OneVector*)loaded)->width = width ;
         ((OneVector*)loaded)->ndec  = ndec ;
         ((OneVector*)loaded)->type  = f_info[n_file]->fmt[n_col].type ;
         for (j = 0; j < coinc->nrec; j++) {
	    if (coinc->rank[n_file][j]) {
               size_t sz = strlen(tmp[coinc->rank[n_file][j]-1])+(size_t)1 ;
               loaded->lab[j] = (char*)calloc(sz, sizeof(char)) ;
               (void)strcpy(loaded->lab[j],tmp[coinc->rank[n_file][j]-1]) ;
	       ((OneVector*)loaded)->is_in[j] = 1 ;
	    }
	    else {
	       ((OneVector*)loaded)->is_in[j] = 0 ;
	    }
         }
         for (j = 0; j < f_info[n_file]->n_rows; j++) 
            free (tmp[j]) ;
         free (tmp) ;
      }
   }

   return loaded ;
}

/*---------------------------------------------------------------------------
 * Function :   get_column()
 * In       :   file and column numbers
 * Out      :   Pointer to newly allocated vector, lenght and number of decimals
 * Job      :   Loads a vector from a file given an external FileInfo structure.
 * Notice   :
 *--------------------------------------------------------------------------*/

static double * 
get_column (int n_file, int n_col, int * len, int * ndec)
{
   double * ret ;
   Finfo  **f_info = f_set->info ;
   int      j, n, nn ;
   char  ** toks = (char**)NULL ;
   char     line[MAX_RECORD_LENGTH] ;
   FILE   * in;

   /* Allocate memory for returning vector */
   ret = (double*) calloc ((size_t)f_info[n_file]->n_rows, sizeof(double));

   /* Open the file */
   in = fopen (f_info[n_file]->filename, "r") ;

   /* Skip the header */
   for (j = 0; j < f_info[n_file]->n_head; j++) 
      (void) fgets (line, MAX_RECORD_LENGTH, in);

   /* Read file column */
   *len = *ndec = 0 ;
   j = 0 ;
   while (j < f_info[n_file]->n_rows && fgets (line, MAX_RECORD_LENGTH, in) != NULL) {
      toks = tokenize (line, " ", &n) ;
      if (n == f_info[n_file]->n_cols) {
         ret[j] = atof (toks[n_col]) ;
         *len = max(*len, (int) strlen (toks[n_col])) ;
         nn = (int)((long int) rindex (toks[n_col],'.') - (long int)toks[n_col]) ;
         nn = (rindex (toks[n_col], '.') == (char*)NULL) ? 0 : (int) strlen (toks[n_col])-nn-1 ;
         *ndec = max (*ndec, nn) ;
         j++ ;
      }
      (void)free_carray(toks, n) ;
   }
   (void)fclose(in) ;

   return ret ;
}

char **
get_column_s(int n_file, int n_col, int * len, int * ndec)
{
   char  ** ret ;
   Finfo  **f_info = f_set->info ;
   int      j, n ;
   char  ** toks = (char**)NULL ;
   char     line[MAX_RECORD_LENGTH] ;
   FILE   * in;

   /* Allocate memory for returning vector */
   ret = (char**) calloc ((size_t)f_info[n_file]->n_rows, sizeof(char*)) ;

   /* Open the file */
   in = fopen (f_info[n_file]->filename, "r") ;

   /* Skip the header */
   for (j = 0; j < f_info[n_file]->n_head; j++) 
      (void) fgets (line, MAX_RECORD_LENGTH, in) ;

   /* Read file column */
   *len = *ndec = 0;
   j = 0 ;
   while (j < f_info[n_file]->n_rows && fgets (line, MAX_RECORD_LENGTH, in) != NULL) {
      toks = tokenize (line," ",&n) ;
      if (n == f_info[n_file]->n_cols) {
         ret[j] = (char*) calloc (strlen (toks[n_col])+(size_t)1, sizeof(char)) ;
         (void) strcpy (ret[j], toks[n_col]) ;
         *len = max (*len, (int) strlen (toks[n_col])) ;
         *ndec = 0 ;
         j++ ;
      }
      (void) free_carray (toks, n) ;
   }
   (void) fclose (in) ;

   return ret ;
}

/*---------------------------------------------------------------------------
 * Function :   load_vector_list()
 * In       :   comma separated vector list
 * Out      :   Pointer to newly allocated vector
 * Job      :   Loads a list of vectors from files and contract it into one
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector * 
load_vector_list (char *vec_list)
{
   OneVector *loaded = NullVector ;
   Finfo    **f_info = f_set->info ;
   int       n_vecs = 0, n_col, n_file ;
   int     * nfile = (int*)NULL ;
   int       i, j, k;
   double ** tmp = (double**)NULL ;
   char   ***stmp = (char***)NULL ;
   char   ** vec_name ;
   int       width, width_o = 0 ;
   int       ndec, ndec_o = 0 ;
   char      type_o = ' ' ;
   char      type ;

   /* Get vector names */
   vec_name = tokenize (vec_list, ",", &n_vecs) ;
   
   /* Check input list and get returning type */
   for (i = 0; i < n_vecs; i++) {
      /* Get file and column number from current vec_name */
      sscanf (vec_name[i], "C%d_%d", &n_col, &n_file) ;
      n_col -= 1 ;
      n_file -= 1 ;
      if (n_file >= coinc->nfiles) {
         cp_error ("illegal file number in list %s: aborting", vec_list) ;
	 goto exit ;
      }
      if (n_col >= f_info[n_file]->n_cols) {
         cp_error ("illegal column number in list %s: aborting", vec_list) ;
	 goto exit ;
      }
      type = f_info[n_file]->fmt[n_col].type ;
      if (type == 's') type_o = 's' ;
   }
   if (type_o == ' ') type_o = 'f' ;

   /* Allocate memory for returning vector */
   loaded = new_vector_s (coinc->nrec, type_o);
   if (loaded == NullVector) {
      cp_error ("cannot allocate output vector: aborting extraction") ;
      return (NullVector) ;
   }

   /* Allocate memory for vectors array */
   tmp = (double**) calloc((size_t)n_vecs, sizeof(double*)) ;
   stmp = (char***) calloc((size_t)n_vecs, sizeof(char**)) ;
   nfile = (int*) calloc((size_t)n_vecs, sizeof(int)) ;

   /* Load vectors from file */
   for(i=0; i<n_vecs; i++ ) {
      /* Get file and column number from current vec_name */
      sscanf(vec_name[i],"C%d_%d",&n_col,&n_file) ;
      n_col -= 1 ;
      n_file -= 1 ;
      nfile[i] = n_file ;
      if(type_o!='s')
         tmp[i] = get_column(n_file, n_col, &width, &ndec) ;
      else
         stmp[i] = get_column_s(n_file, n_col, &width, &ndec) ;
      width_o = max(width_o, width) ;
      ndec_o = max(ndec_o, ndec) ;
   }

   /* Fill returning vector */
   ((OneVector*)loaded)->width = width_o ;
   ((OneVector*)loaded)->ndec  = ndec_o ;

   if(type_o!='s') {
      for(j=0; j<coinc->nrec; j++) {
         i = 0 ;
         loaded->data[j] = 0.0 ;
         loaded->is_in[j] = 0 ;
         while ( loaded->is_in[j] == 0 && i<n_vecs ) {
            n_file = nfile[i] ;
	    if(coinc->rank[n_file][j]) {
               loaded->data[j] = tmp[i][coinc->rank[n_file][j]-1] ;
	       loaded->is_in[j] = 1 ;
	       /* Checks for Null values */
	       for(k=0; k<N_NullVal; k++) {
	          if(loaded->data[j] == NullValues[k]) loaded->is_in[j] = 0 ;
               }
	    }
	    i++ ;
         }
      }
   } else {
      for(j=0; j<coinc->nrec; j++) {
         i = 0 ;
         /* strcpy(loaded->lab[j],"") ; */
         loaded->is_in[j] = 0 ;
         while ( loaded->is_in[j] == 0 && i<n_vecs ) {
            n_file = nfile[i] ;
	    if(coinc->rank[n_file][j]) {
               size_t sz = strlen(stmp[i][coinc->rank[n_file][j]-1])+(size_t)1 ;
               loaded->lab[j] = (char*) calloc(sz, sizeof(char)) ;
               (void)strcpy(loaded->lab[j],stmp[i][coinc->rank[n_file][j]-1]) ;
	       loaded->is_in[j] = 1 ;
	    }
	    i++ ;
         }

      }
   }

   exit:
   /* Free memory */
   if (tmp != (double**)NULL) {
      for (i = 0; i < n_vecs; i++ ) free (tmp[i]) ;
      free (tmp) ;
   }
   if (stmp != (char***)NULL) {
      for (i = 0; i < n_vecs; i++) {
         if (stmp[i] != (char**)NULL) {
            n_file = nfile[i] ;
            for (j = 0; j < f_info[n_file]->n_rows; j++) 
               if (stmp[i][j] != (char*)NULL) free (stmp[i][j]) ;
            free (stmp[i]) ;
         }
      }
      free (stmp) ;
   }
   if (nfile != (int*)NULL) free (nfile) ;
   (void) free_carray (vec_name, n_vecs) ;

   return loaded ;
}

/*---------------------------------------------------------------------------
 * Function :   load_computed_vector()
 * In       :   comma separated vector list
 * Out      :   Pointer to newly allocated vector
 * Job      :   Loads a list of vectors from files and 
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector *
load_computed_vector( char *expr )
{
   OneVector *loaded = NullVector ;
   int       i, n_vecs = 0, n_errs, n_class ;
   char   ** vec_name = (char **)NULL ;
   char   ** err_name = (char **)NULL ;  
   char      vec_list[MAX_RECORD_LENGTH] = "\0" ;
   char   ** vec_classes ; 
   char      opr[10] = "\0" ;
   int       len ;
   char    * tok_exp ;
   char      tokv[MAX_OP][NAME_SIZE] ;
   char    * token ;
   int       tokc ;
   int       weight = 0 ;

   /* Get operator to apply and vector list */
   len = (int)((long int)rindex(expr,')')-(long int)index(expr,'('))-1 ;
   (void)strncpy(vec_list, (const char *)index(expr,'(')+1, (size_t)len) ;
   len = (int)((long int)index(expr,'(')-(long int)expr) ;
   (void)strncpy(opr, expr, (size_t)len) ;
   if(strncmp(opr,"W",(size_t)1)==0||strncmp(opr,"w",(size_t)1)==0) weight = 1 ;
   
   /* Separes vectors from errors list */
   vec_classes = tokenize(vec_list,":",&n_class) ;
   if(n_class==2&&weight==0) {
      cp_error("cannot specify an error list with operator %s: aborting",opr) ;
      goto exit ;
   }
   if(n_class>2) {
      cp_error("cannot specify more than 2 vectors list with operator %s: aborting",opr) ;
      goto exit ;
   }
   
   /* Check the expression syntax */
   for(i=0; i<n_class; i++) {
      tok_exp = tokenizeExpression(vec_classes[i]) ;
      tokc = 0 ;
      token = strtok(tok_exp, SEPARATOR) ;
      while (token != NULL) {
	 (void)strcpy(tokv[tokc], token) ;
	 tokc++ ;
	 token = strtok(NULL, SEPARATOR) ;
      }
      token = tokv[0] ;
      if(tokc>1||checklist(token)!=vector_list) {
	 cp_error("operator %s applys only to a vectors list: aborting",opr) ;
         free(tok_exp) ;
	 goto exit ;
      }
      free(tok_exp) ;
   }

   /* Get vector names */
   vec_name = tokenize(vec_classes[0],",",&n_vecs) ;
   if(weight) {
      if(n_class==2) {
	 err_name  = tokenize(vec_classes[1],",",&n_errs) ;
	 if( n_vecs != n_errs ) {
	    cp_error("vectors and errors lists have different number of elements: aborting") ;
	    goto exit ;
	 }
      }
      else
         err_name = GetErrorName(vec_name, n_vecs) ;
      n_errs = n_vecs ;
   }

   /*
    * AverageVector: computes the arithmetic mean, stdev or sterr vector
    * WeightAverageVector: computes the weighted mean, stdev or sterr vector
    */
   if( strcmp(opr,"AVE")==0 || strcmp(opr,"ave") == 0 )
      loaded = AverageVector(vec_name, n_vecs, AVE) ;
   else if( strcmp(opr,"STDEV")==0 || strcmp(opr,"stdev") == 0 )
      loaded = AverageVector(vec_name, n_vecs, STDEV) ;
   else if( strcmp(opr,"STERR")==0 || strcmp(opr,"sterr") == 0 )
      loaded = AverageVector(vec_name, n_vecs, STERR) ;
   else if( strcmp(opr,"WAVE")==0 || strcmp(opr,"wave") == 0 )
      loaded = WeightAverageVector(vec_name, err_name, n_vecs, WAVE) ;
   else if( strcmp(opr,"WSTDEV")==0 || strcmp(opr,"wstdev") == 0 )
      loaded = WeightAverageVector(vec_name, err_name, n_vecs, WSTDEV) ;
   else if( strcmp(opr,"WSTERR")==0 || strcmp(opr,"wsterr") == 0 )
      loaded = WeightAverageVector(vec_name, err_name, n_vecs, WSTERR) ;

   exit:

   /* Free memory */
   (void)free_carray(vec_classes, n_class) ;
   (void)free_carray(vec_name, n_vecs) ;
   if(weight) (void)free_carray(err_name, n_errs) ;
   return loaded ;

}

/*---------------------------------------------------------------------------
 * Function :   AverageVector()
 * In       :   array of vector names, no. of vectors, ave/sig code
 * Out      :   Pointer to newly allocated vector
 * Job      :   Loads a list of vectors from files and computes the mean, 
 *              the standard deviation or the standard error vector
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector * 
AverageVector(char ** vec_name, int n_vecs, int opr)
{
   OneVector **vecs ;
   OneVector *loaded = NullVector ;
   OneVector *average = NullVector ;
   OneVector *sum = NullVector ;
   OneVector *stdev = NullVector ;
   int       i, j, imax ;
   double    tsum, tmax, tstdev, n_1 ;
   int       n_iter = 0, n_rej ;

   /* Allocate memory and load vectors from files */
   vecs = (OneVector **)calloc((size_t)n_vecs, sizeof(OneVector *)) ;
   if (vecs == (OneVector **)NULL) {
      cp_error("cannot allocate output vector: aborting extraction") ;
      goto exit ;
   }
   for(i=0; i<n_vecs; i++ ) {
      vecs[i] = load_vector(vec_name[i]) ;
      if (vecs[i] == NullVector) {
	 cp_error("cannot allocate output vector: aborting extraction") ;
	 goto exit ;
      }
   }

   /* Starts loop for sigma-rejection */
   do
   {
      /* Computes the mean vector */
      average = new_vector(coinc->nrec);
      average->width = vecs[0]->width+1 ;
      average->ndec  = vecs[0]->ndec+1 ;
      average->type  = vecs[0]->type ;
      for(i=0; i<n_vecs; i++ ) {
	 for (j=0 ; j<average->size ; j++) {
            if(vecs[i]->is_in[j]) {
               average->data[j] += vecs[i]->data[j] ;
               average->is_in[j] += vecs[i]->is_in[j] ;
	    }
	 }
      }
      for (j=0 ; j<average->size ; j++) {
	 if (fabs((double)average->is_in[j]) < 1e-10)
            average->data[j] = 0.0 ;
	 else
            average->data[j] /= (double)average->is_in[j] ;
      }
      /*
       *  Computes the stdev vector.
       *  Here we use the two pass formula as in Num.Recip., p.607
       */
      sum = new_vector(coinc->nrec);
      stdev = new_vector(coinc->nrec);
      if ( stdev == NullVector || sum== NullVector ) {
	 cp_error("cannot allocate tmp vectors: aborting execution") ;
	 goto exit ;
      }
      stdev->width = average->width ;
      stdev->ndec  = average->ndec ;
      stdev->type  = average->type ;
      for(i=0; i<n_vecs; i++ ) {
	 for (j=0 ; j<average->size ; j++) {
            if (vecs[i]->is_in[j]) {
	       tsum = vecs[i]->data[j]-average->data[j] ;
	       sum->data[j] += tsum ;
               stdev->data[j] += tsum*tsum ;
               stdev->is_in[j]++ ;
	    }
	 }
      }
      for (j=0 ; j<average->size ; j++) {
	 if (stdev->is_in[j]) {
  	    n_1 = (double)max(stdev->is_in[j]-1, 1) ;
  	    tsum = (double)pow(sum->data[j],2.0)/(double)stdev->is_in[j] ;
  	    stdev->data[j] = (stdev->data[j]-tsum)/n_1 ;
  	    stdev->data[j] = sqrt(stdev->data[j]) ;
	 }
	 else
            stdev->data[j] = 0.0 ;
      }
      /*
       * On first pass we check the minimum number of measures 
       * for a valid result
       */
      if(n_iter==0) {
         for (j=0 ; j<average->size ; j++) {
            if(average->is_in[j]<n_min) 
	       for(i=0; i<n_vecs; i++ ) vecs[i]->is_in[j] = 0 ;
         }
      }
      /*
       * Sigma rejection
       */
      n_iter++ ;
      n_rej = 0;
      if( sigma && n_iter<=max_iter ) {
	 for (j=0 ; j<average->size ; j++) {
            tmax = -1.0 ;
	    imax = -1 ;
            for(i=0; i<n_vecs; i++ ) {
               if (vecs[i]->is_in[j]) {
	          tstdev = max(stdev->data[j],0.5*(double)pow(10.0,(double)(-stdev->ndec))) ;
		  tsum = (vecs[i]->data[j]-average->data[j])/tstdev ;
		  if(tsum>tmax) {
	             tmax = tsum ;
		     imax = i ;
		  }
	       }
	    }
	    if(tmax > nsig_rej) {
	       vecs[imax]->is_in[j] = 0 ;
	       n_rej++ ;
	       /*inserire logfile? */
	    }	    
	 }
	 cp_message(0,2,"Iteration %d: rejected %d measures\n",n_iter,n_rej);
	 destroy_vector(sum) ;
	 destroy_vector(average) ;
	 destroy_vector(stdev) ;
      }
   } while ( sigma && n_iter<=max_iter ) ;

   /* Copy the result to output vector */
   if( opr == AVE )
      loaded = copy_vector(average) ;
   else if( opr == STDEV )
      loaded = copy_vector(stdev) ;
   else {
      for (j=0 ; j<average->size ; j++)
	 if (stdev->is_in[j])
  	    stdev->data[j] = (stdev->data[j])/sqrt((double)stdev->is_in[j]) ;
      loaded = copy_vector(stdev) ;
   }

   exit :

   /* Free memory */
   destroy_vector(sum) ;
   destroy_vector(average) ;
   destroy_vector(stdev) ;
   for(i=0; i<n_vecs; i++ ) destroy_vector(vecs[i]) ;
   free(vecs) ;

   return loaded ;
}

/*---------------------------------------------------------------------------
 * Function :   WeightAverageVector()
 * In       :   array of vector and error names, no. of vectors, ave/sig code
 * Out      :   Pointer to newly allocated vector
 * Job      :   Loads a list of vectors and errors from files and computes 
 *              the error-weighted mean, standard deviation or standard error
 *              vector
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector * 
WeightAverageVector(char ** vec_name, char ** err_name, int n_vecs, int opr)
{
   OneVector **vecs ;
   OneVector **errs ;
   OneVector *loaded = NullVector ;
   OneVector *average = NullVector ;
   OneVector *stdev = NullVector ;
   int       i, j, imax ;
   double    tsum, tmax, tstdev, n_1 ;
   double    ww, wt, wtot ;
   int       n_iter = 0 ;

   /* Allocate memory and load vectors from files */
   vecs = (OneVector **)calloc((size_t)n_vecs, sizeof(OneVector *)) ;
   errs = (OneVector **)calloc((size_t)n_vecs, sizeof(OneVector *)) ;
   if (vecs == (OneVector **)NULL || errs == (OneVector **)NULL) {
      cp_error("cannot allocate output vector: aborting extraction") ;
      goto exit ;
   }
   for(i=0; i<n_vecs; i++ ) {
      vecs[i] = load_vector(vec_name[i]) ;
      errs[i] = load_vector(err_name[i]) ;
      if (vecs[i] == NullVector || errs[i] == NullVector ) {
	 cp_error("cannot allocate output vector: aborting extraction") ;
	 goto exit ;
      }
   }

   /* Starts loop for sigma-rejection */
   do
   {
      /* Computes the mean vector */
      average = new_vector(coinc->nrec);
      average->width = vecs[0]->width+1 ;
      average->ndec  = vecs[0]->ndec+1 ;
      average->type  = vecs[0]->type ;
      for (j=0 ; j<average->size ; j++) {
         wtot = 0.0 ;
         for(i=0; i<n_vecs; i++ ) {
	    if(!errs[i]->is_in[j]) errs[i]->data[j] = 0.0 ;
            if(vecs[i]->is_in[j]) {
	       ww = max(errs[i]->data[j],0.5*(double)pow(10.0,(double)(-errs[i]->ndec))) ;
               wt = 1.0 / (double)pow(ww,(double)2) ;
               average->data[j] += wt*vecs[i]->data[j] ;
               average->is_in[j] += vecs[i]->is_in[j] ;
	       wtot += wt ;
	    }
	 }
	 if (fabs((double)average->is_in[j]) < 1e-32)
            average->data[j] = 0.0 ;
	 else
            average->data[j] /= wtot ;
      }
      /*
       *  Computes the stdev vector.
       */
      stdev = new_vector(coinc->nrec);
      if ( stdev == NullVector ) {
	 cp_error("cannot allocate tmp vectors: aborting execution") ;
	 goto exit ;
      }
      stdev->width = average->width ;
      stdev->ndec  = average->ndec ;
      stdev->type  = average->type ;
      for (j=0 ; j<average->size ; j++) {
         wtot = 0.0 ;
         for(i=0; i<n_vecs; i++ ) {
            if (vecs[i]->is_in[j]) {
	       ww = max(errs[i]->data[j],0.5*(double)pow(10.0,(double)(-errs[i]->ndec))) ;
               wt = 1.0 / (double)pow(ww,(double)2) ;
	       tsum = vecs[i]->data[j]-average->data[j] ;
               stdev->data[j] += wt*tsum*tsum ;
               stdev->is_in[j]++ ;
	       wtot += wt ;
	    }
	 }
         if (stdev->is_in[j]) {
	    n_1 = (double)max(stdev->is_in[j]-1, 1) ;
	    stdev->data[j] = sqrt(stdev->data[j]/(n_1*wtot)) ;
	    /* multiply by sqrt(n) */
	    stdev->data[j] = stdev->data[j] * sqrt((double)stdev->is_in[j]) ;
	 }
	 else
	    stdev->data[j] = 0.0 ;
      }
      /*
       * On first pass we check the minimum number of measures 
       * for a valid result
       */
      if(n_iter==0) {
         for (j=0 ; j<average->size ; j++) {
            if(average->is_in[j]<n_min) 
	       for(i=0; i<n_vecs; i++ ) vecs[i]->is_in[j] = 0 ;
         }
      }
      /*
       * Sigma rejection
       */
      n_iter++ ;
      if( sigma && n_iter<=max_iter ) {
	 for (j=0 ; j<average->size ; j++) {
            tmax = -1.0 ;
	    imax = -1 ;
            for(i=0; i<n_vecs; i++ ) {
               if (vecs[i]->is_in[j]) {
	          tstdev = max(stdev->data[j],0.5*(double)pow(10.0,(double)(-stdev->ndec))) ;
		  tsum = (vecs[i]->data[j]-average->data[j])/tstdev ;
		  if(tsum>tmax) {
	             tmax = tsum ;
		     imax = i ;
		  }
	       }
	    }
	    if(tmax > nsig_rej) vecs[imax]->is_in[j] = 0 ;
	    /*inserire logfile? */
	 }
	 destroy_vector(average) ;
	 destroy_vector(stdev) ;
      }
   } while ( sigma && n_iter<=max_iter ) ;

   /* Copy the result to output vector */
   if( opr == WAVE )
      loaded = copy_vector(average) ;
   else if( opr == WSTDEV )
      loaded = copy_vector(stdev) ;
   else {
      for (j=0 ; j<average->size ; j++)
	 if (stdev->is_in[j])
  	    stdev->data[j] = (stdev->data[j])/sqrt((double)stdev->is_in[j]) ;
      loaded = copy_vector(stdev) ;
   }

   exit :

   /* Free memory */
   destroy_vector(average) ;
   destroy_vector(stdev) ;
   for(i=0; i<n_vecs; i++ ) destroy_vector(vecs[i]) ;
   for(i=0; i<n_vecs; i++ ) destroy_vector(errs[i]) ;
   free(vecs) ;
   free(errs) ;

   return loaded ;
}

static char ** 
GetErrorName(char ** list, int n_list)
{
   int i, n_col, n_file ;
   char    outvec[10] = "\0" ;
   char ** outlist ;
   outlist = (char**) calloc((size_t)n_list, sizeof(char*)) ;
   for(i=0; i<n_list; i++) {
      sscanf(list[i], "C%d_%d", &n_col, &n_file);
      sprintf(outvec, "C%d_%d", ++n_col, n_file);
      outlist[i] = (char *) calloc(strlen(outvec)+(size_t)1, sizeof(char)) ;
      (void)strcpy(outlist[i], outvec) ;
      cp_message(0,4,"Vector name: %s; error vector associated: %s\n",list[i],outlist[i]) ;
   }
   return outlist ;
}

/*---------------------------------------------------------------------------
 * Function :  checklist()
 * In       :  token (character string)
 * Out      :  symbol
 * Job      :  find out if a token is a vector_list
 * Notice   :
 *--------------------------------------------------------------------------*/

static itemType 
checklist( char * token ) 
{
   regex_t re_expr;
   char lista[] = "^C[0-9]*_[0-9]*,(C[0-9]*_[0-9]*,)*C[0-9]*_[0-9]*$" ;

   if (token == (char*)NULL) {
	   cp_error("internal: cannot get token") ;
	   return unknown ;
   }

   /* try to match a vector list: */
   if (regcomp(&re_expr, &lista[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error("cannot compile rule for vector-list: aborting") ;
      return unknown ;
   }
   if (regexec(&re_expr, token, (size_t) 0, NULL, 0) == 0) {
      cp_message(0,4,"matched a vector list\n");
      regfree(&re_expr) ;
      return vector_list;
   }
   cp_error("unrecognized token: [%s]", token) ;
   return unknown ;
}

