/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: vector_arith.h
 * Author 	: Paolo Montegriffo
 * Created on	: July, 2000	
 * Language	: ANSI C
 * Description	: 
 * Version      : 2.0 - October 3, 2000
 *                2.1.2 - September 4, 2001
 *--------------------------------------------------------------------------*/

#ifndef VECTOR_ARITHMETIC_H
#define VECTOR_ARITHMETIC_H 1

#include "common.h"

BEGIN_C_DECLS

/*----------------------------------------------------------------------------
 *						Function ANSI C prototypes
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * Function	:	add_vector()
 * In 		:	2 vectors
 * Out 		:	1 newly allocated vector
 * Job		:	add two vectors
 * Notice	:	vectors must share same dimensions
 *--------------------------------------------------------------------------*/

int vector_add(OneVector * vector1, OneVector * vector2) ; 

/*----------------------------------------------------------------------------
 * Function	:	sub_vector()
 * In 		:	2 vectors
 * Out 		:	1 newly allocated vector
 * Job		:	subtract two vectors
 * Notice	:	vectors must share same dimensions
 *--------------------------------------------------------------------------*/

int vector_sub(OneVector *vector1, OneVector *vector2) ;

/*----------------------------------------------------------------------------
 * Function	:	mul_vector()
 * In 		:	2 vectors
 * Out 		:	1 newly allocated vector
 * Job		:	multiply two vectors
 * Notice	:	vectors must share same dimensions
 *--------------------------------------------------------------------------*/

int vector_mul(OneVector *vector1, OneVector *vector2) ;
 
/*----------------------------------------------------------------------------
 * Function	:	div_vector()
 * In 		:	2 vectors
 * Out 		:	1 newly allocated vector
 * Job		:	divide two vectors
 * Notice	:	vectors must share same dimensions
 *--------------------------------------------------------------------------*/

int vector_div(OneVector *vector1, OneVector *vector2) ;

/*----------------------------------------------------------------------------
 * Function	:	add_img_local()
 * In 		:	2 vectors
 * Out 		:	int 0 if ok, -1 otherwise
 * Job		:	add two vectors, store the result in the first vector
 * Notice	:	optimized (i.e. mostly unreadable)	
 *--------------------------------------------------------------------------*/

int add_img_local(OneVector * im1, OneVector * im2) ;

/*----------------------------------------------------------------------------
 * Function	:	sub_img_local()
 * In 		:	2 vectors
 * Out 		:	int 0 if ok, -1 otherwise
 * Job		:	subtract two vectors, store the result in the first vector
 * Notice	:	optimized (i.e. mostly unreadable)	
 *--------------------------------------------------------------------------*/

int sub_img_local(OneVector * im1, OneVector * im2) ;

/*----------------------------------------------------------------------------
 * Function	:	mul_img_local()
 * In 		:	2 vectors
 * Out 		:	int 0 if ok, -1 otherwise
 * Job		:	multiply two vectors, store the result in the first vector
 * Notice	:	optimized (i.e. mostly unreadable)	
 *--------------------------------------------------------------------------*/

int mul_img_local(OneVector * im1, OneVector * im2) ;

/*----------------------------------------------------------------------------
 * Function	:	op_vector()
 * In 		:	2 vectors, operation to perform
 * Out 		:	newly allocated vector
 * Job		:	4 operations between 2 vectors
 * Notice	:	returns NULL in case of error
 *--------------------------------------------------------------------------*/

OneVector * op_vector( OneVector * vector1, OneVector * vector2, int operation) ;

OneVector * fn_op_vector( OneVector * vector1, opType operation) ;

/*----------------------------------------------------------------------------
 * Function	:	cst_op_vector()
 * In 		:	1 vector, 1 constant, operation to perform
 * Out 		:	1 newly allocated vector
 * Job		:	arithmetic operations between an vector and a constant
 * Notice	:	possible operations are:
    Addition        '+'
    Subtraction     '-'
    Multiplication  '*'
    Division        '/'
    Logarithm       'l'
    Power           '^'
    Exponentiation  'e'
 *--------------------------------------------------------------------------*/

OneVector * cst_op_vector(OneVector * vector_in, OneNumber constant, int operation) ;

/*----------------------------------------------------------------------------
 * Function	:	abs_vector()
 * In 		:	1 vector
 * Out 		:	1 newly allocated vector
 * Job		:	for each pixel, out = abs(in)
 * Notice	:
 *--------------------------------------------------------------------------*/

OneVector * abs_vector(OneVector * vector_in) ;

/*---------------------------------------------------------------------------
   Function :   invert_vector()  
   In       :   1 vector
   Out      :   int 0 if Ok, -1 otherwise
   Job      :   invert all pixels, i.e. pix = -pix
   Notice   :   operates in place - no new memory allocated
 ---------------------------------------------------------------------------*/

int invert_vector(OneVector * in) ;

/*---------------------------------------------------------------------------
 * Function :   new_vector()
 * In       :   vector size (x and y)
 * Out      :   New allocated structure
 * Job      :   Allocates an vector structure and fills it with
 *              relevant information. Also allocates data buffer
 * Notice   :   Object constructor
 *--------------------------------------------------------------------------*/

OneVector * new_vector(int size_x) ;
OneVector * new_vector_s(int size_x, int type) ;

/*---------------------------------------------------------------------------
 * Function :   destroy_vector()
 * In       :   Pointer to allocated structure
 * Out      :   void
 * Job      :   free memory associated to an vector structure
 * Notice   :   Object destructor
 *--------------------------------------------------------------------------*/

void destroy_vector(OneVector * to_destroy) ; 

/*---------------------------------------------------------------------------
 * Function :   copy_vector()
 * In       :   1 vector
 * Out      :  	1 newly allocated vector 
 * Job      :   copy an vector structure
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector * copy_vector(OneVector * src_vec) ;

#ifndef VECTOR_LOAD_H
#define VECTOR_LOAD_H 1

/*---------------------------------------------------------------------------
 * Function :   load_vector()
 * In       :   vector name
 * Out      :   Pointer to newly allocated vector
 * Job      :   Loads a vector from a file ordering elements with a ranking
 *		table build with the coincidence table.
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector * load_vector(char * vec_name) ;

/*---------------------------------------------------------------------------
 * Function :   load_vector_list()
 * In       :   comma separated vector list
 * Out      :   Pointer to newly allocated vector
 * Job      :   Loads a list of vectors from files and contract it into one
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector * load_vector_list(char * vec_string) ;

/*---------------------------------------------------------------------------
 * Function :   load_computed_vector()
 * In       :   comma separated vector list
 * Out      :   Pointer to newly allocated vector
 * Job      :   Loads a list of vectors from files and 
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector * load_computed_vector(char * expr) ;

/*---------------------------------------------------------------------------
 * Function :   AverageVector()
 * In       :   array of vector names, no. of vectors, ave/sig code
 * Out      :   Pointer to newly allocated vector
 * Job      :   Loads a list of vectors from files and computes the mean, 
 *              the standard deviation or the standard error vector
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector * AverageVector(char ** vec_name, int n_vecs, int opr) ;

/*---------------------------------------------------------------------------
 * Function :   WeightAverageVector()
 * In       :   array of vector and error names, no. of vectors, ave/sig code
 * Out      :   Pointer to newly allocated vector
 * Job      :   Loads a list of vectors and errors from files and computes 
 *              the error-weighted mean, standard deviation or standard error
 *              vector
 * Notice   :
 *--------------------------------------------------------------------------*/

OneVector * WeightAverageVector(char ** vec_name, char ** err_name, int n_vecs, int opr) ;

#endif   /* !VECTOR_LOAD_H */

END_C_DECLS

#endif   /* !VECTOR_ARITHMETIC_H */
/*--------------------------------------------------------------------------*/
